// {{{ copyright

/********************************************************************
 *
 * $Id: CommandDistributor.jpp,v 1.8 2000/11/14 11:29:28 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.command;

// {{{ imports

import java.awt.Component;


import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;


import javax.swing.JOptionPane;

import de.qfs.lib.gui.Message;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.ExceptionHandler;

// }}}

// {{{ doc

/**
 * This class is responsible for the distribution of high level commands among
 * possible command handlers. <p>
 *
 * Commands will either be invoked directly, if they implement the
 * {@link Invokable Invokable} interface, or passed through a series
 * of {@link CommandHandler CommandHandlers} until one calls the Command's
 * {@link Command#setHandled setHandled} method.<p>
 *
 * Command handlers may be registered for all types of commands, for commands
 * originating from a certain source or for commands whose source is of a
 * certain class. The usefulness of this mechanism has yet to be proved. <p>
 *
 * There are several reasons for having a high level command distribution
 * system in addition to the GUI's event loop:
 * <ul>
 * <li>
 * It enables error handling through exceptions which can be caught at
 * the top level for command execution and turned into a message for the
 * user.
 * </li>
 * <li>
 * It decouples GUI elements through which commands are initiated from the
 * possible executors of the commands.
 * </li>
 * <li>
 * It is a good place to add support for undo/redo.
 * </li>
 * </ul>
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.8 $
 */

// }}}
public class CommandDistributor
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (CommandDistributor.class);

    /**
     * The default CommandDistributor.
     */
    private static CommandDistributor defaultDistributor
        = new CommandDistributor ();

    /**
     * Registered exception handlers.
     */
    private List exHandlers = new ArrayList ();

    /**
     * Generic handlers.
     */
    private Set handlers = new HashSet ();

    /**
     * Handlers for commands from special sources.
     */
    private Map objectHandlers = new HashMap ();

    /**
     * Handlers for commands with sources of special classes.
     */
    private Map classHandlers = new HashMap ();

    /**
     * ID counter for CommandThreads.
     */
    private static int threadCount;

    // }}}

    // {{{ constructor

    /**
     * Create a new CommandDistributor.
     */
    public CommandDistributor ()
    {
    }


    // }}}

    //----------------------------------------------------------------------
    // The default distributor
    //----------------------------------------------------------------------
    // {{{ getDefaultDistributor

    /**
     * Get the default CommandDistributor.
     *
     * @return  The default CommandDistributor.
     */
    public static CommandDistributor getDefaultDistributor()
    {
        return defaultDistributor;
    }

    // }}}
    // {{{ setDefaultDistributor

    /**
     * Set the default CommandDistributor.
     *
     * @param   distributor     The CommandDistributor to set.
     */
    public static void setDefaultDistributor(CommandDistributor distributor)
    {
        defaultDistributor = distributor;
    }

    // }}}

    //----------------------------------------------------------------------
    // Assigning commands
    //----------------------------------------------------------------------
    // {{{ assignCommand

    /**
     * Either invoke a command directly or assign it to the registered
     * CommandHandlers. CommandExceptions thrown during the execution
     * of the command will be caught and handled as follows:
     * <dl>
     * <dt>{@link MessageCommandException MessageCommandException}</dt>
     * <dd>The resources of the exception will be used to create a message
     * dialog via {@link de.qfs.lib.gui.Message Message}.</dd>
     * <dt>{@link DialogCommandException DialogCommandException}</dt>
     * <dd>The Dialog for the exception will be displayed.</dd>
     * <dt>{@link CancelCommandException CancelCommandException}</dt>
     * <dd>The exception will be silently ignored.</dd>
     * <dt>Other CommandExceptions</dt>
     * <dd>An error will be logged via the {@link de.qfs.lib.log.Log Log}
     * mechanism </dd>
     * </dl>
     *
     * To handle RuntimeExceptions or Errors thrown by a Command, {@link
     * de.qfs.lib.util.ExceptionHandler ExceptionHandlers} can be registered
     * with {@link #addExceptionHandler addExceptionHandler}.
     *
     * @param   command The command to assign.
     *
     * @return  Whatever is returned from the execution of the command.
     */
    public Object assignCommand(Command command)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "assignCommand(Command)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "command: " + command);
        }
        try {
            return assignCommandSilently(command);
        } catch (RuntimeCommandException ex) {
            if (ex.getMessageResource() != null) {
                Object parent = ex.getParent();
                if (parent == null) {
                    parent = command.getSource();
                }
                Message.showMessage((parent instanceof Component)
                                    ? (Component) parent : null,
                                    ex.getMessageResource(),
                                    ex.getMessageParameters());
            } else if (ex.getNested() != null) {
                ExceptionHandler[] hdlr;
                Exception nested = ex.getNested();
                synchronized (this) {
                    hdlr = (ExceptionHandler[])
                        exHandlers.toArray(new ExceptionHandler[0]);
                }
                for (int i = 0; i < hdlr.length; i++) {
                    if (hdlr[i].handleException(nested)) {
                        return null;
                    }
                }
                if (nested instanceof RuntimeException) {
                    throw (RuntimeException) nested;
                }
                // Still not handled
                throw ex;
            }
        } catch (MessageCommandException ex) {
            Message.showMessage(ex.getParent(), ex.getMessageResource(),
                                ex.getMessageParameters());
        } catch (DialogCommandException ex) {
            ex.getDialog().show();
        } catch (CancelCommandException ex) {
            if (logger.level >= Log.MSG) {
                logger.log(Log.MSG, "assignCommand(Command)", ex);
            }
        } catch (UnhandledCommandException ex) {
            if (logger.level >= Log.WRN) {
                logger.log(Log.WRN, "assignCommand(Command)", ex);
            }
        } catch (CommandException ex) {
            if (logger.level >= Log.ERR) {
                logger.log("assignCommand(Command)", ex);
            }
        } catch (RuntimeException ex) {
            ExceptionHandler[] hdlr;
            synchronized (this) {
                hdlr = (ExceptionHandler[])
                    exHandlers.toArray(new ExceptionHandler[0]);
            }
            for (int i = 0; i < hdlr.length; i++) {
                if (hdlr[i].handleException(ex)) {
                    return null;
                }
            }
            throw ex;
        } catch (Error ex) {
            ExceptionHandler[] hdlr;
            synchronized (this) {
                hdlr = (ExceptionHandler[])
                    exHandlers.toArray(new ExceptionHandler[0]);
            }
            for (int i = 0; i < hdlr.length; i++) {
                if (hdlr[i].handleException(ex)) {
                    return null;
                }
            }
            throw ex;
        }
        return null;
    }

    // }}}
    // {{{ assignCommandSilently

    /**
     * Either invoke a command directly or assign it to the registered
     * CommandHandlers. This methos is similar to {@link #assignCommand
     * assignCommand}, except that no Exception handling takes place, neither
     * for CommandExceptions, nor for RuntimeExceptions.
     *
     * @param   command The command to assign.
     *
     * @return  Whatever is returned from the execution of the command.
     *
     * @throws  CommandException        If something goes wrong during the
     *                          execution of the command.
     */
    public Object assignCommandSilently(Command command)
         throws CommandException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "assignCommandSilently(Command)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "command: " + command);
        }
        // Try to invoke the command directly
        if (command instanceof Invokable) {
            return ((Invokable) command).invoke();
        }

        Object ret;
        CommandHandler[] hdl = null;

        // Try out handlers registered for the source of the command
        synchronized (this) {
            Object oHandlers = objectHandlers.get(command.getSource());
            if (oHandlers != null) {
                hdl = (CommandHandler[])
                    ((Collection)oHandlers).toArray (new CommandHandler [0]);
            }
        }
        if (hdl != null) {
            for (int i = 0; i < hdl.length; i++) {
                ret = hdl[i].handleCommand(command);
                if (command.isHandled()) {
                    return ret;
                }
            }
        }

        // Try out handlers registered for the class of the source
        // of the command
        synchronized (this) {
            Object cHandlers =
                classHandlers.get(command.getSource().getClass());
            if (cHandlers != null) {
                hdl = (CommandHandler[])
                    ((Collection)cHandlers).toArray (new CommandHandler [0]);
            }
        }
        if (hdl != null) {
            for (int i = 0; i < hdl.length; i++) {
                ret = hdl[i].handleCommand(command);
                if (command.isHandled()) {
                    return ret;
                }
            }
        }

        // Finally try the general handlers
        hdl = (CommandHandler[]) handlers.toArray (new CommandHandler [0]);
        for (int i = 0; i < hdl.length; i++) {
            ret = hdl[i].handleCommand(command);
            if (command.isHandled()) {
                return ret;
            }
        }

        // We failed
        if (command.mustBeHandled()) {
            throw new UnhandledCommandException(command);
        }
        return null;
    }

    // }}}
    // {{{ assignAsyncCommand

    /**
     * Start a background thread and assign a Command in it. Exception
     * handling is similar to {@link #assignCommand assignCommand}.
     *
     * @param   command The Command to run asynchronously.
     */
    public void assignAsyncCommand(AsyncCommand command)
    {
        new CommandThread (command).start();
    }

    // }}}

    //----------------------------------------------------------------------
    // The CommandHandlers
    //----------------------------------------------------------------------
    // {{{ addCommandHandler(CommandHandler)

    /**
     * Register a general CommandHandler.
     *
     * @param   handler The CommandHandler to register.
     */
    public synchronized void addCommandHandler(CommandHandler handler)
    {
        handlers.add(handler);
    }

    // }}}
    // {{{ addCommandHandler(CommandHandler, Object)

    /**
     * Register a CommandHandler for Commands originating from a certain
     * source.
     *
     * @param   handler The CommandHandler to register.
     * @param   source  The source of the commands.
     */
    public synchronized void addCommandHandler(CommandHandler handler,
                                               Object source)
    {
        Collection oHandlers = (Collection) objectHandlers.get(source);
        if (oHandlers == null) {
            oHandlers = new HashSet ();
            objectHandlers.put(source, oHandlers);
        }
        oHandlers.add(handler);
    }

    // }}}
    // {{{ addCommandHandler(CommandHandler, Class)

    /**
     * Register a CommandHandler for commands whose source is of a certain
     * class.
     *
     * @param   handler The CommandHandler to register.
     * @param   sourceClass     The Class to look for.
     */
    public synchronized void addCommandHandler(CommandHandler handler,
                                               Class sourceClass)
    {
        Collection cHandlers = (Collection) classHandlers.get(sourceClass);
        if (cHandlers == null) {
            cHandlers = new HashSet ();
            classHandlers.put(sourceClass, cHandlers);
        }
        cHandlers.add(handler);
    }

    // }}}
    // {{{ removeCommandHandler(CommandHandler)

    /**
     * Remove a general CommandHandler.
     *
     * @param   handler The CommandHandler to remove.
     */
    public synchronized void removeCommandHandler(CommandHandler handler)
    {
        handlers.remove(handler);
    }

    // }}}
    // {{{ removeCommandHandler(CommandHandler, Object)

    /**
     * Remove a source specific CommandHandler.
     * @param   handler The CommandHandler to remove.
     * @param   source  The source of the commands.
     */
    public synchronized void removeCommandHandler(CommandHandler handler,
                                                  Object source)
    {
        if (objectHandlers.containsKey(source)) {
            ((Collection)objectHandlers.get(source)).remove(handler);
        }
    }

    // }}}
    // {{{ removeCommandHandler(CommandHandler, Class)

    /**
     * Remove a source class specific CommandHandler.

     * @param   handler The CommandHandler to remove.
     * @param   sourceClass     The source's Class.
     */
    public synchronized void removeCommandHandler(CommandHandler handler,
                                                  Class sourceClass)
    {
        if (classHandlers.containsKey(sourceClass)) {
            ((Collection)classHandlers.get(sourceClass)).remove(handler);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Exception handlers
    //----------------------------------------------------------------------
    // {{{ addExceptionHandler

    /**
     * Add an ExceptionHandler to handle RuntimeExceptions. Handlers added
     * later will be called first.
     *
     * @param   handler The ExceptionHandler to add.
     */
    public synchronized void addExceptionHandler(ExceptionHandler handler)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "addExceptionHandler(ExceptionHandler)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "handler: " + handler);
        }
        exHandlers.remove(handler);
        exHandlers.add(0, handler);
    }

    // }}}
    // {{{ removeExceptionHandler

    /**
     * Remove an ExceptionHandler.
     *
     * @param   handler The ExceptionHandler to remove.
     */
    public synchronized void removeExceptionHandler(ExceptionHandler handler)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "removeExceptionHandler(ExceptionHandler)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "handler: " + handler);
        }
        exHandlers.remove(handler);
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ getParent

    private Component getParent (Command command)
    {
        return (command.getSource() instanceof Component)
            ? (Component) command.getSource()
            : null;
    }

    // }}}

    // {{{ class CommandThread

    /**
     * A thread class that runs Commands in the background.
     */
    private class CommandThread extends Thread
    {
        /**
         * The command to run.
         */
        private Command command;

        /**
         * Create a new CommandThread.
         *
         * @param       command The command to run.
         */
        public CommandThread (Command command)
        {
            setName("CommandThread-" + threadCount++);
            this.command = command;
        }

        /**
         * Run the command.
         */
        public void run()
        {
            assignCommand(command);
        }
    }

    // }}}
}
