// {{{ copyright

/********************************************************************
 *
 * $Id: Configuration.jpp,v 1.7 2001/02/04 12:19:42 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.config;

// {{{ imports

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.ObjectStreamClass;
import java.io.PrintWriter;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Vector;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

// {{{ doc

/**
 * The Configuration class provides the means for an application
 * to save and restore some of its state, e.g. GUI relevant stuff like
 * window positions, or options that can be customized interactively by the
 * user. <p>
 *
 * Objects which implement the {@link Configurable Configurable} interface,
 * can be registered with the Configuration. They will automatically be
 * queried for their state, when the Configuration is saved and have their
 * state restored, when the Configuration is restored. <p>
 *
 * The Configuration also takes care of restored state for objects that are
 * not registered at the time of the restore. It will call the {@link
 * Configurable#setConfigurableState Configurable.setConfigurableState} method
 * on those objects, once they get registered and will even save their state
 * back so it doesn't get lost. Thus e.g. a Frame can be constructed and
 * configured at runtime only when needed and its state won't get lost during
 * an invocation of the application in which the user never opens it. <p>
 *
 * It is possible to register more than one Configurable under the same name.
 * This makes sence if you have multiple instances of something sharing the
 * same kind of state. The state that will be saved with the Configuration
 * is that of the most recently registered Configurable.<p>
 *
 * The Configuration follows the Singleton pattern, i.e. there should always
 * be only one Configuration instance per application, which can be obtained
 * via the static {@link #instance instance} method. There is however a
 * protected constructor to allow saving or restoring parts of the
 * Configuration in different files through a derived class. <p>
 *
 * Changes in the state of registered Configurables can be tracked by
 * registering a {@link ConfigurationListener ConfigurationListener} with
 * {@link #addConfigurationListener addConfigurationListener}. However it is
 * up to the Configurables themselves to notify the Configuration of a state
 * change. <p>
 *
 * The Configuration is saved in a human readable and editable format. For
 * each Configurable a section is added that starts with the Configurable's
 * name in brackets, e.g. <code>[SomeConfigurableName]</code>, followed by an
 * optional comment and the state of the Configurable in the standard property
 * file format. Comments start with '#' and can be set for the whole file via
 * {@link #setComment(String) setComment} or for a Configurable via {@link
 * #setComment(String,String) setComment}. When the Configuration is restored
 * and from a file then saved again, all comments and the order of the
 * Configurables will be preserved, but the order of the properties inside the
 * section of a Configurable is unpredictable. <p>
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.7 $
 */

// }}}
public class Configuration
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (Configuration.class);

    /**
     * The Singleton instance.
     */
    private static Configuration theInstance = new Configuration ();

    /**
     * Line separator (system specific).
     */
    private final static String LINESEP = System.getProperty("line.separator");

    /**
     * A mapping from names to Configurables.
     */
    protected Hashtable configurables = new Hashtable ();

    /**
     * A mapping from Configurables to names.
     */
    protected Hashtable nameMap = new Hashtable ();

    /**
     * The current state of all registered Configurables and of objects that
     * have been restored but not registered, indexed by their name.
     */
    protected Hashtable states = new Hashtable ();

    /**
     * A vector holding all registered names in the sequence they were read
     * or registered.
     */
    protected Vector names = new Vector ();

    /**
     * The header comment of the config file.
     */
    private String comment;

    /**
     * The section comments of the config file.
     */
    private Hashtable comments = new Hashtable ();

    /**
     * The ConfigurationListeners
     */
    private Vector listeners = new Vector ();

    /**
     * Threads in this Vector will not cause stateChanged events.
     */
    private Vector silenced = new Vector ();

    // }}}

    // {{{ constructor

    /**
     * The constructor is protected to prevent the creation of more than one
     * Configuration object.
     *
     * @see #instance
     */
    protected Configuration ()
    {
    }

    // }}}

    // {{{ singleton

    /**
     * Get the single Configuration instance.
     *
     * @return  The instance.
     */
    public static Configuration instance()
    {
        return theInstance;
    }

    // }}}

    //----------------------------------------------------------------------
    // Registration
    //----------------------------------------------------------------------
    // {{{ register

    /**
     * Register a Configurable. <p>
     *
     * If the state of the Configurable has already been restored, also set
     * its state.
     *
     * @param   configurable    The Configurable.
     */
    public void register(Configurable configurable)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "register(Configurable)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "configurable: " + configurable);
        }

        registerAs(configurable, configurable.getConfigurableName());
    }

    /**
     * Register a Configurable under a (possibly) different name.  If the
     * state of the Configurable has already been restored, also set its
     * state.
     *
     * @param   configurable    The Configurable.
     * @param   name            The name to register under.
     */
    public void registerAs(Configurable configurable, String name)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "registerAs(Configurable,String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "configurable: " + configurable + ", " +
                       "name: " + name);
        }

        Properties state;
        boolean mustSet = false;

        synchronized (this) {
            Object old = configurables.get(name);
            Object newObject;

            if (old != null) {
                if (old instanceof Vector) {
                    // There's already some Configurables there,
                    // just add the new one
                    newObject = old;
                    state = ((Configurable)((Vector)old).elementAt(0))
                        .getConfigurableState();
                } else {
                    // There's exactly one Configurable there, make a new
                    // Vector with the new and the old one.
                    newObject = new Vector();
                    ((Vector)newObject).insertElementAt(old, 0);
                    state = ((Configurable)old).getConfigurableState();
                }
                ((Vector)newObject).insertElementAt(configurable, 0);
                mustSet = true;
            } else {
                // Just add the new Configurable
                newObject = configurable;
                state = (Properties) states.get(name);
                if (state == null) {
                    state = configurable.getConfigurableState();
                    states.put(name, state);
                } else {
                    mustSet = true;
                }
            }
            configurables.put(name, newObject);
            nameMap.put(configurable, name);
            if (! names.contains(name)) {
                names.addElement(name);
            }
            configurable.registrationChanged(this, name, true, false);
        }
        // Setting the state can lead to deadlocks, so it's not synchronized
        if (state != null && mustSet) {
            try {
                silenced.addElement(Thread.currentThread());
                configurable.setConfigurableState(state);
            } finally {
                silenced.removeElement(Thread.currentThread());
            }
        }
    }

    // }}}
    // {{{ unregister

    /**
     * Remove a Configurable from the Configuration.
     *
     * @param   configurable    The Configurable.
     * @param   keepValues      If true, keep the current state of the
     *                          Configurable, otherwise forget it.
     */
    public void unregister(Configurable configurable,
                           boolean keepValues)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "unregister(Configurable,boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "configurable: " + configurable + ", " +
                       "keepValues: " + keepValues);
        }

        unregisterAs(configurable, configurable.getConfigurableName(),
                     keepValues);
    }

    /**
     * Remove a Configurable under a (possibly) different name from
     * the Configuration.
     *
     * @param   configurable    The Configurable.
     * @param   name            The name to use.
     * @param   keepValues      If true, keep the current state of the
     *                          Configurable, otherwise forget it.
     */
    public synchronized void unregisterAs(Configurable configurable,
                                          String name,
                                          boolean keepValues)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "unregisterAs(Configurable,String,boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "configurable: " + configurable + ", " +
                       "name: " + name + ", " +
                       "keepValues: " + keepValues);
        }

        Object old = configurables.get(name);
        if (old == null) {
            // oops, nothing there
            if (logger.level >= Log.ERR) {
                logger.log(Log.ERR,
                           "unregisterAs(Configurable,String,boolean)",
                           "Trying to unregister unknown Configurable " +
                           name);
            }
        } else if (old instanceof Vector) {
            ((Vector)old).removeElement(configurable);
            if (((Vector)old).size() == 0) {
                // This was the last one, maybe keep its state
                if (keepValues) {
                    Properties state = configurable.getConfigurableState();
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG,
                                   "unregisterAs(Configurable,String,boolean)",
                                   "Keeping state: " + state);
                    }
                    states.put(name, state);
                } else {
                    names.removeElement(name);
                }
                configurables.remove(name);
            }
        } else if (old != configurable) {
            // oops, unknown Configurable
            if (logger.level >= Log.ERR) {
                logger.log(Log.ERR,
                           "unregisterAs(Configurable,String,boolean)",
                           "Trying to unregister different Configurable " +
                           "under name " + name);
            }
        } else {
            if (keepValues) {
                Properties state = configurable.getConfigurableState();
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG,
                               "unregisterAs(Configurable,String,boolean)",
                               "Keeping state: " + state);
                }
                states.put(name, state);
            } else {
                names.removeElement(name);
            }
            configurables.remove(name);
        }
        nameMap.remove(configurable);
        configurable.registrationChanged(this, name, false, keepValues);
    }

    // }}}
    // {{{ unregisterAll

    /**
     * Remove all Configurables of the same name from the Configuration.
     *
     * @param   name            The name of the Configurables.
     * @param   keepValues      If true, keep the current state of the
     *                          most recently registered Configurable,
     *                          otherwise forget it.
     */
    public void unregisterAll(String name, boolean keepValues)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "unregisterAll(String,boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "keepValues: " + keepValues);
        }

        Object old = configurables.get(name);
        if (old == null) {
            return;
        }

        if (old instanceof Vector) {
            for (int i = 0; i < ((Vector) old).size(); i++) {
                Configurable configurable =
                    (Configurable) ((Vector) old).elementAt(i);
                nameMap.remove(configurable);
                configurable.registrationChanged(this, name, false,
                                                 keepValues);
            }
            old = ((Vector) old).elementAt(((Vector) old).size() - 1);
        } else {
            nameMap.remove(old);
            ((Configurable) old).registrationChanged(this, name, false,
                                                     keepValues);
        }

        configurables.remove(name);

        if (keepValues) {
            Properties state = ((Configurable) old).getConfigurableState();
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "unregisterAll(String,boolean)",
                           "Keeping state: " + state);
            }
            states.put(name, state);
        }
    }

    // }}}
    // {{{ getRegistrationName

    /**
     * Get the last name a Configurable was registered under. This may be
     * different from the Configurable's name.
     *
     * @param   configurable    The Configurable whose name to get.
     *
     * @return  The name the Configurable is registered under or null, if it
     *          is not registered.
     */
    public String getRegistrationName(Configurable configurable)
    {
        return (String) nameMap.get(configurable);
    }

    // }}}

    //----------------------------------------------------------------------
    // Saving and restoring
    //----------------------------------------------------------------------
    // {{{ save

    /**
     * Save the complete Configuration to an OutputStream.
     *
     * @param   out     The stream to save to.
     */
    public void save(OutputStream out)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "save(OutputStream)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "out: " + out);
        }

        PrintWriter pw = new PrintWriter (new OutputStreamWriter (out));
        if (comment != null) {
            printComment(pw, comment);
        }

        // iterate through names and save states
        for (int i = 0; i < names.size(); i++) {
            String name = (String) names.elementAt(i);

            Object conf = configurables.get(name);
            Properties state = null;

            if (conf instanceof Vector) {
                state = ((Configurable) ((Vector) conf).elementAt(0))
                    .getConfigurableState();
            } else if (conf instanceof Configurable) {
                state = ((Configurable) conf).getConfigurableState();
            } else {
                state = (Properties) states.get(name);
            }
            if (state == null) {
                logger.log(Log.WRN, "save(OutputStream)",
                           "No state for " + name);
                continue;
            }
            if (i > 0) {
                pw.println();
            }
            pw.print("[");
            pw.print(name);
            pw.println("]");

            String cmt = (String) comments.get(name);
            if (cmt != null) {
                printComment(pw, cmt);
            }
            pw.print(writeProps(state));
            pw.flush();
        }
    }

    // }}}
    // {{{ restore

    /**
     * Restore the Configuration from an InputStream.
     *
     * @param   in      The stream to restore from.
     *
     * @throws  java.io.IOException     If the data from the stream is not
     *                                  in the format generated by {@link
     *                                  #save save}.
     */
    public void restore(InputStream in)
        throws IOException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "restore(InputStream)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "in: " + in);
        }

        BufferedReader br = new BufferedReader (new InputStreamReader (in));

        StringBuffer sb = new StringBuffer ();

        String line = fetchLine(br);
        line = fetchComment(line, br, sb);
        comment = sb.toString();
        try {
            silenced.addElement(Thread.currentThread());
            while (line != null) {
                if (line.startsWith("[") && line.endsWith("]")) {
                    String name = line.substring(1, line.length() - 1);
                    sb = new StringBuffer ();
                    line = fetchComment(fetchLine(br), br, sb);
                    String cmt = sb.toString();
                    if (cmt.length() > 0) {
                        comments.put(name, cmt);
                    }
                    sb = new StringBuffer ();
                    line = fetchContent(line, br, sb);
                    Properties state = readProps(sb.toString());
                    Object ob = configurables.get(name);
                    if (state != null) {
                        states.put(name, state);
                    }
                    if (ob instanceof Vector) {
                        for (int i = 0; i < ((Vector)ob).size(); i++) {
                            ((Configurable)((Vector)ob).elementAt(i))
                                .setConfigurableState(state);
                        }
                    } else if (ob != null) {
                        ((Configurable)ob).setConfigurableState(state);
                    }
                    if (! names.contains(name)) {
                        names.addElement(name);
                    }
                } else {
                    throw new IOException
                        ("Malformed Configuration InputStream");
                }
            }
        } finally {
            silenced.removeElement(Thread.currentThread());
        }
    }

    // }}}
    // {{{ setComment

    /**
     * Set the comment for the header of the config file.
     *
     * @param   comment The comment to set.
     */
    public void setComment(String comment)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setComment(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "comment: " + comment);
        }
        this.comment = comment;
    }

    /**
     * Set a comment for one Configurable name.
     *
     * @param   name    The Configurable name.
     * @param   comment The comment to set.
     */
    public void setComment(String name, String comment)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setComment(String,String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "comment: " + comment);
        }
        comments.put(name, comment);
    }

    // }}}
    // {{{ getComment

    /**
     * Get the comment read from the config file.
     */
    public String getComment()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getComment()", "");
        }
        return comment;
    }

    /**
     * Get the comment for one Configurable name.
     *
     * @param   name    The Configurable name.
     *
     * @return  The comment or null.
     */
    public String getComment(String name)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setComment(String,String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "comment: " + comment);
        }
        return (String) comments.get(name);
    }

    // }}}

    //----------------------------------------------------------------------
    // ConfigurationListeners
    //----------------------------------------------------------------------
    // {{{ addConfigurationListener

    /**
     * Add a ConfigurationListener to the Configuration.
     *
     * @param   listener        The ConfigurationListener to add.
     */
    public void addConfigurationListener(ConfigurationListener listener)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "addConfigurationListener(ConfigurationListener)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "listener: " + listener);
        }
        listeners.addElement(listener);
    }

    // }}}
    // {{{ removeConfigurationListener

    /**
     * Remove a ConfigurationListener from the Configuration.
     *
     * @param   listener        The ConfigurationListener to remove.
     */
    public void removeConfigurationListener(ConfigurationListener listener)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "removeConfigurationListener(ConfigurationListener)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "listener: " + listener);
        }
        listeners.removeElement(listener);
    }

    // }}}
    // {{{ stateChanged

    /**
     * Notify the Configuration of a state change in one of its registered
     * Configurables. It is up to the Configurable whether it notifies the
     * Configuration of changes or not. <p>
     *
     * If any ConfigurationListeners are registered, the Configuration will
     * compare the Configurables new state to its old state and notify the
     * listeners if anything has changed.
     *
     * @param   configurable    The Configurable whose state changed.
     */
    public void stateChanged(Configurable configurable)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "stateChanged(Configurable)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "configurable: " + configurable);
        }
        if (silenced.contains(Thread.currentThread())) {
            // ignore
            return;
        }
        String name = (String) nameMap.get(configurable);
        if (name == null) {
            logger.log(Log.WRN, "stateChanged(Configurable)",
                       "unregistered Configurable: " + configurable);
            return;
        }
        Properties newState = configurable.getConfigurableState();
        Properties oldState = (Properties) states.get(name);
        states.put(name, newState);
        if (listeners.size() == 0) {
            // nobody interested
            return;
        }

        if (oldState != null) {
            boolean equal = (newState.size() == oldState.size());
            // check if anything has changed
            if (equal) {
                for (Enumeration keys = oldState.keys();
                     keys.hasMoreElements(); ) {
                    String key = (String) keys.nextElement();
                    if (! oldState.get(key).equals(newState.get(key))) {
                        equal = false;
                        break;
                    }
                }
            }
            if (equal) {
                // No changes
                return;
            }
        }

        ConfigurationEvent event = new ConfigurationEvent
            (configurable, name, newState);
        Vector tmp = (Vector) listeners.clone();
        for (int i = 0; i < tmp.size(); i++) {
            ((ConfigurationListener) tmp.elementAt(i)).stateChanged(event);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // String/Properties conversion
    //----------------------------------------------------------------------
    // {{{ writeProps

    /**
     * Write one set of Properties to a String.
     *
     * @param   state   The Properties to write.
     *
     * @return  The String containing the Properties.
     */
    private String writeProps(Properties state)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "writeProps(Properties)", "");
        }
        ByteArrayOutputStream out = new ByteArrayOutputStream ();

        try {
            state.store(out, "");
        } catch (IOException ex) {
            // shouldn't happen, ignore
        }

        String ret = out.toString();
        while (ret.length() > 0 &&
               ret.charAt(0) == '#'
               && ret.indexOf(LINESEP) > 0) {
            ret = ret.substring(ret.indexOf(LINESEP) + LINESEP.length());
        }
        return ret;
    }

    // }}}
    // {{{ readProps

    /**
     * Read a set of Properties from a String.
     *
     * @param   input   The String to read from.
     *
     * @return  The Properties read.
     */
    private Properties readProps(String input)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "readProps(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "input: " + input);
        }
        Properties state = new Properties ();
        try {
            state.load(new ByteArrayInputStream(input.getBytes()));
        } catch (IOException ex) {
            logger.log("fetch(String)", ex);
        }
        return state;
    }

    // }}}

    //----------------------------------------------------------------------
    // Writing
    //----------------------------------------------------------------------
    // {{{ printComment

    /**
     * Write a comment to a stream.
     *
     * @param   out     The stream to print to.
     * @param   comment The comment to print.
     */
    private void printComment(PrintWriter out, String comment)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "printComment(PrintStream,String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "comment: " + comment);
        }

        StringTokenizer st = new StringTokenizer (comment, LINESEP, true);
        boolean needNewLine = false;
        while (st.hasMoreTokens()) {
            String line = st.nextToken();
            char c = line.charAt(0);
            if (c == '\n' || c == '\r') {
                needNewLine = false;
            } else {
                needNewLine = true;
                if (! isComment(line)) {
                    out.print("# ");
                }
            }
            out.print(line);
        }
        if (needNewLine) {
            out.println();
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Reading
    //----------------------------------------------------------------------
    // {{{ isComment

    /**
     * Check if a line is a valid comment line i.e. it is either whitespace
     * only, or the first non-whitespace character is a '#'.
     *
     * @param   line    The line to check.
     *
     * @return  True, if the line is a valid comment.
     */
    private boolean isComment(String line)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "isComment(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "line: " + line);
        }

        int i;
        for (i = 0; i < line.length(); i++) {
            char c = line.charAt(i);
            if (c == '#') {
                return true;
            }
            if (! Character.isWhitespace(c)) {
                return false;
            }
        }
        return true;
    }

    // }}}
    // {{{ fetchLine

    /**
     * Get the next line from a BufferedReader, suppressing IOExceptions.
     *
     * @param   br      The reader to query.
     *
     * @return  The line read.
     */
    private String fetchLine(BufferedReader br)
    {
        try {
            String line = br.readLine();
            return line;
        } catch (IOException ex) {
            logger.log("fetchComment(String,BufferedReader,StringBuffer)",
                       ex);
            return null;
        }
    }


    // }}}
    // {{{ fetchComment

    /**
     * Read lines from a BufferedReader, collecting comments into a
     * StringBuffer.
     *
     * @param   line    A prefetched line form the reader.
     * @param   br      The Reader
     * @param   sb      The StringBuffer that will hold the comment
     *
     * @return  The first non-comment line
     */
    private String fetchComment(String line, BufferedReader br,
                                StringBuffer sb)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "fetchComment(String,BufferedReader,StringBuffer)", "");
        }
        while (line != null) {
            if (isComment(line)) {
                sb.append(line);
                sb.append(LINESEP);
            } else {
                break;
            }
            line = fetchLine(br);
        }
        return line;
    }

    // }}}
    // {{{ fetchContent

    /**
     * Read lines from a BufferedReader, collecting them into a StringBuffer.
     * Will stop at the next section heading (a line starting with '[').
     * Comments are skipped.
     *
     * @param   line    A prefetched line form the reader.
     * @param   br      The Reader
     * @param   sb      The StringBuffer that will hold the content
     *
     * @return  The first non-content line
     */
    private String fetchContent(String line, BufferedReader br,
                                StringBuffer sb)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "fetchContent(String,BufferedReader,StringBuffer)", "");
        }
        while (line != null) {
            if (line.startsWith("[")) {
               break;
            } else if (! isComment(line)) {
                sb.append(line);
                sb.append(LINESEP);
            }
            line = fetchLine(br);
        }
        return line;
    }

    // }}}
}

