// {{{ copyright

/********************************************************************
 *
 * $Id: DialogConfigWrapper.java,v 1.10 2001/01/25 17:06:59 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.config;

// {{{ imports

import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;

import java.io.Serializable;

import java.util.Properties;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * A ConfigWrapper for {@link java.awt.Dialog Dialog} objects. The state saved
 * is the dialog's location and size.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.10 $
 */
public class DialogConfigWrapper
    extends ConfigWrapper
    implements ComponentListener
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (DialogConfigWrapper.class);

    /**
     * The wrapped Dialog.
     */
    protected Dialog dialog;

    /**
     * Helper for registration callbacks.
     */
    protected RegistrationHelper regHelper = new RegistrationHelper ();

    // }}}

    // {{{ constructor

    /**
     * Create a new DialogConfigWrapper with the default name "dialog".
     */
    public DialogConfigWrapper()
    {
        setConfigurableName("dialog");
    }

    // }}}

    // {{{ setClient

    /**
     * Set the client object for the ConfigWrapper.
     *
     * @param   client  The client object to set.
     */
    protected void setClient(Object client)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setClient(Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "client: " + client);
        }
        dialog = (Dialog)client;
    }

    // }}}

    //----------------------------------------------------------------------
    // Configurable methods
    //----------------------------------------------------------------------
    // {{{ getConfigurableState

    /**
     * Get the Configurable's state.
     */
    public Properties getConfigurableState()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getConfigurableState()", "");
        }
        Properties state = new Properties ();
        Point pos = dialog.getLocation();
        Dimension dim = dialog.getSize();
        state.put("x", "" + pos.x);
        state.put("y", "" + pos.y);
        state.put("width", "" + dim.width);
        state.put("height", "" + dim.height);
        return state;
    }

    // }}}
    // {{{ setConfigurableState

    /**
     * Set the Configurable's state.
     */
    public void setConfigurableState(Properties state)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setConfigurableState(Properties)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "state: " + state);
        }

        String x = state.getProperty("x");
        String y = state.getProperty("y");
        String width = state.getProperty("width");
        String height = state.getProperty("height");
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "setConfigurableState(Properties)",
                       "x: " + x + ", y: " + y +
                       ", width: " + width + ", height: " + height);
        }
        if (x == null || y == null || width == null || height == null) {
            if (logger.level >= Log.WRN) {
                logger.log(Log.WRN, "setConfigurableState(Properties)",
                           "Invalid state: " +
                           "x: " + x + ", y: " + y +
                           ", width: " + width + ", height: " + height);
            }
            return;
        }
        try {
            restore(new Dimension (Integer.parseInt(width),
                                   Integer.parseInt(height)),
                    new Point (Integer.parseInt(x),
                               Integer.parseInt(y)));
        } catch (NumberFormatException ex) {
            logger.log(Log.WRN, "setConfigurableState(Properties)", ex);
        }
    }

    // }}}
    // {{{ registrationChanged

    /**
     * This callback is used to notify the Configurable whenever it is
     * registered with or unregistered from the Configuration, so it can
     * decide whether to notify the Configuration when its state changes.
     *
     * @param   config  The affected Configuration.
     * @param   name    The registration name of the Configurable which may
     *                  be different from the Configurable's name.
     * @param   registered True if the Configurable was just registered,
     *                  false if it was unregistered.
     * @param   valuesKept Whether the Configurable's values were kept, if
     *                  it was just unregistered.
     */
    public void registrationChanged(Configuration config, String name,
                                    boolean registered, boolean valuesKept)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "registrationChanged(Configuration,String," +
                       "boolean,boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "registered: " + registered);
        }
        int ret = regHelper.registrationChanged(config, registered);

        if (dialog != null) {
            if (ret == 1) {
                dialog.addComponentListener(this);
            } else if (ret == -1) {
                dialog.removeComponentListener(this);
            }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // ComponentListener interface
    //----------------------------------------------------------------------
    // {{{ componentMoved

    public void componentMoved(ComponentEvent e)
    {
        regHelper.stateChanged(this);
    }

    // }}}
    // {{{ componentResized

    public void componentResized(ComponentEvent e)
    {
        regHelper.stateChanged(this);
    }

    // }}}
    // {{{ componentHidden

    public void componentHidden(ComponentEvent e)
    {
    }

    // }}}
    // {{{ componentShown

    public void componentShown(ComponentEvent e)
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ restore

    protected void restore (final Dimension size, final Point location)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "restore(Dimension,Point)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "size: " + size + ", " +
                       "location: " + location);
        }

        dialog.setLocation (location);
        dialog.pack();
        dialog.setSize(size);
        dialog.validate();
    }

    // }}}
}
