// {{{ copyright

/********************************************************************
 *
 * $Id: RelativeDialogConfigWrapper.java,v 1.2 2001/02/04 12:19:43 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.config;

// {{{ imports

import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Window;
import java.awt.event.ComponentEvent;

import java.util.Properties;

import javax.swing.JDialog;

import de.qfs.lib.gui.SwingUtil;
import de.qfs.lib.gui.EventQueueMark;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * Special configuration wrapper that stores the Dialog location relative
 * to some other window instead of as absolute screen values.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.2 $
 * @since       0.98.0
 */
public class RelativeDialogConfigWrapper
    extends DialogConfigWrapper
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger (RelativeDialogConfigWrapper.class);

    /**
     * The master window to position relative to.
     */
    protected Window master;

    /**
     * The stored relative location of the dialog. Must be stored since the
     * Dialog's location may have been overridden to constrain it to the
     * screen.
     */
    protected Point location;

    /**
     * The reference for the Dialog's location, one of "Absolute", "Center",
     * "Northwest", "Northeast", "Southwest" and "Southeast".
     */
    protected String reference;

    /**
     * Mark to distinguish user events from artificial ones.
     */
    protected EventQueueMark mark;

    /**
     * Whether to constrain the dialog to the screen
     */
    protected boolean constrain;

    /**
     * Whether to store the dialog's size as well as its location.
     */
    protected boolean storeSize;

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ RelativeDialogConfigWrapper

    /**
     * Create a new RelativeDialogConfigWrapper.
     */
    public RelativeDialogConfigWrapper (Dialog dialog, String name,
                                        boolean constrain, boolean storeSize)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "RelativeDialogConfigWrapper(Dialog,String,boolean," +
                       "boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "dialog: " + dialog + ", " +
                       "name: " + name + ", " +
                       "constrain: " + constrain + ", " +
                       "storeSize: " + storeSize);
        }
        this.dialog = dialog;
        this.constrain = constrain;
        this.storeSize = storeSize;
        setConfigurableName(name);
        getLocation();
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ setMaster

    /**
     * Set the master window for the dialog.
     *
     * @param   window  The new master.
     */
    public void setMaster(Window window)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setMaster(Window)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "window: " + window);
        }
        if (this.master != null) {
            this.master.removeComponentListener(this);
        }
        master = window;
        if (master != null) {
            master.addComponentListener(this);
        }
        if (master == null || "Absolute".equalsIgnoreCase(reference)) {
            getLocation();
            regHelper.stateChanged(this);
        } else {
            setLocation(dialog.getSize());
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Configurable methods
    //----------------------------------------------------------------------
    // {{{ getConfigurableState

    /**
     * Get the Configurable's state.
     */
    public Properties getConfigurableState()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getConfigurableState()", "");
        }
        Properties state = new Properties ();
        state.put("x", "" + location.x);
        state.put("y", "" + location.y);
        state.put("reference", reference);
        if (storeSize) {
            Dimension dim = dialog.getSize();
            state.put("width", "" + dim.width);
            state.put("height", "" + dim.height);
        }
        return state;
    }

    // }}}
    // {{{ setConfigurableState

    /**
     * Set the Configurable's state.
     */
    public void setConfigurableState(Properties state)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setConfigurableState(Properties)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "state: " + state);
        }

        String x = state.getProperty("x");
        if (x != null) {
            try {
                location.x = Integer.parseInt(x);
            } catch (NumberFormatException ex) {
                if (logger.level >= Log.WRN) {
                    logger.log(Log.WRN,
                               "setConfigurableState(Properties)", ex);
                }
            }
        }
        String y = state.getProperty("y");
        if (y != null) {
            try {
                location.y = Integer.parseInt(y);
            } catch (NumberFormatException ex) {
                if (logger.level >= Log.WRN) {
                    logger.log(Log.WRN,
                               "setConfigurableState(Properties)", ex);
                }
            }
        }

        if (state.containsKey("reference")) {
            reference = state.getProperty("reference");
        }

        Dimension size = dialog.getSize();
        if (storeSize) {
            String width = state.getProperty("width");
            if (width != null) {
                try {
                    size.width = Integer.parseInt(width);
                } catch (NumberFormatException ex) {
                    if (logger.level >= Log.WRN) {
                        logger.log(Log.WRN,
                                   "setConfigurableState(Properties)", ex);
                    }
                }
            }

            String height = state.getProperty("height");
            if (height != null) {
                try {
                    size.height = Integer.parseInt(height);
                } catch (NumberFormatException ex) {
                    if (logger.level >= Log.WRN) {
                        logger.log(Log.WRN,
                                   "setConfigurableState(Properties)", ex);
                    }
                }
            }
        }
        setLocation(size);
    }

    // }}}

    //----------------------------------------------------------------------
    // ComponentListener interface
    //----------------------------------------------------------------------
    // {{{ componentMoved

    /**
     * Notify the RelativeDialogConfigWrapper that its dialog has moved.
     *
     * @param   e       Details about the event.
     */
    public void componentMoved(ComponentEvent e)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "componentMoved(ComponentEvent)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "e: " + e);
        }
        if (e.getSource() == master) {
            getLocation();
            return;
        } else if (e.getSource() == dialog) {
            if (mark != null && mark.isActive()) {
                // ignore
                return;
            }
            getLocation();
            super.componentMoved(e);
        }
    }

    // }}}
    // {{{ componentResized

    /**
     * Notify the RelativeDialogConfigWrapper that a window was resized.
     *
     * @param   e       Details about the event.
     */
    public void componentResized(ComponentEvent e)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "componentResized(ComponentEvent)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "e: " + e);
        }
        if (storeSize && e.getSource() == dialog) {
            super.componentResized(e);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Heper methods
    //----------------------------------------------------------------------
    // {{{ restore

    /**
     * Restore the dialog's size and location.
     *
     * @param   size     The size to set.
     * @param   location The location to set.
     */
    protected void restore (final Dimension size, final Point location)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "restore(Dimension,Point)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "size: " + size + ", " +
                       "location: " + location);
        }
        if (mark == null) {
            mark = new EventQueueMark ();
        }
        mark.activate(200);
        if (dialog instanceof JDialog) {
            SwingUtil.invokeAndWait(new Runnable() {
                public void run() {
                    dialog.setLocation (location);
                    dialog.pack();
                    dialog.setSize(size);
                    dialog.validate();
                }
            });
        } else {
            dialog.setLocation (location);
            dialog.pack();
            dialog.setSize(size);
            dialog.validate();
        }
        mark.activate();
    }

    // }}}
    // {{{ setLocation

    /**
     * Set the dialog's location relative to its master.
     *
     * @param   size    The new size for the dialog.
     */
    protected void setLocation(Dimension size)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setLocation()",
                       "size: " + size +
                       ", location: " + location +
                       ", reference: " + reference);
        }
        Point pos = new Point (location);
        if ("Absolute".equalsIgnoreCase(reference)
            || master == null) {
            restore(size, pos);
            return;
        }

        Point mPos = master.getLocation();
        Dimension mSize = master.getSize();
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "setLocation()",
                       "mpos: " + mPos + ", mSize: " + mSize);
        }

        if ("Northwest".equalsIgnoreCase(reference)) {
            pos.translate(mPos.x, mPos.y);
        } else if ("Northeast".equalsIgnoreCase(reference)) {
            pos.translate(mPos.x + mSize.width, mPos.y);
        } else if ("Southwest".equalsIgnoreCase(reference)) {
            pos.translate(mPos.x, mPos.y + mSize.height);
        } else if ("Southeast".equalsIgnoreCase(reference)) {
            pos.translate(mPos.x + mSize.width, mPos.y + mSize.height);
        } else if ("Center".equalsIgnoreCase(reference)) {
            pos.translate(mPos.x + mSize.width/2, mPos.y + mSize.height/2);
        }
        if (constrain) {
            pos.x = Math.max(pos.x, 0);
            pos.y = Math.max(pos.y, 0);
            Dimension screen = dialog.getToolkit().getScreenSize();
            pos.x = Math.min(pos.x, screen.width - size.width);
            pos.y = Math.min(pos.y, screen.height - size.height);
        }
        restore(size, pos);
    }

    // }}}
    // {{{ getLocation

    /**
     * Determine location and reference relative to the master window.
     */
    protected void getLocation()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getLocation()", "");
        }
        location = dialog.getLocation();
        if (master == null) {
            reference = "Absolute";
            return;
        }
        Dimension size = dialog.getSize();
        Point center = new Point (location.x + size.width / 2,
                                  location.y + size.height / 2);
        Point mPos = master.getLocation();

        Dimension mSize = master.getSize();
        double min = Math.sqrt((center.x - mPos.x) * (center.x - mPos.x)
                               + (center.y - mPos.y) * (center.y - mPos.y));
        reference = "Northwest";
        Point rel = new Point (location.x - mPos.x, location.y - mPos.y);

        double tst = Math.sqrt((center.x - mPos.x - mSize.width) *
                               (center.x - mPos.x - mSize.width)
                               + (center.y - mPos.y) * (center.y - mPos.y));
        if (tst < min) {
            min = tst;
            reference = "Northeast";
            rel = new Point (location.x - mPos.x - mSize.width,
                             location.y - mPos.y);
        }

        tst = Math.sqrt((center.x - mPos.x) * (center.x - mPos.x)
                        + (center.y - mPos.y - mSize.height) *
                          (center.y - mPos.y - mSize.height));
        if (tst < min) {
            min = tst;
            reference = "Southwest";
            rel = new Point (location.x - mPos.x,
                             location.y - mPos.y - mSize.height);
        }

        tst = Math.sqrt((center.x - mPos.x - mSize.width) *
                        (center.x - mPos.x - mSize.width)
                        + (center.y - mPos.y - mSize.height) *
                          (center.y - mPos.y - mSize.height));
        if (tst < min) {
            min = tst;
            reference = "Southeast";
            rel = new Point (location.x - mPos.x - mSize.width,
                             location.y - mPos.y - mSize.height);
        }

        tst = Math.sqrt((center.x - mPos.x - mSize.width/2) *
                        (center.x - mPos.x - mSize.width/2)
                        + (center.y - mPos.y - mSize.height/2) *
                          (center.y - mPos.y - mSize.height/2));
        if (tst < min) {
            min = tst;
            reference = "Center";
            rel = new Point (location.x - mPos.x - mSize.width/2,
                             location.y - mPos.y - mSize.height/2);
        }

        location = rel;
    }

    // }}}
}
