// {{{ copyright

/********************************************************************
 *
 * $Id: AWTTreeAdapter.java,v 1.1 2001/04/20 16:07:28 gs Exp $
 *
 * Copyright (C) 2000 Gregor Schmid, Quality First Software
 * All rights reserved
 *
 *******************************************************************/

// }}}

package de.qfs.lib.gui;

// {{{ imports

import java.awt.Component;
import java.awt.Container;

import javax.swing.JMenu;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.tree.IndexBasedTreeAdapter;

// }}}

/**
 * This is a TreeAdapter for the AWT Component hierarchy. It assumes that all
 * nodes passed are Components. It takes special care of JMenus, which violate
 * the standard AWT tree structure.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.1 $
 * @since       0.98.1
 */
public class AWTTreeAdapter
    extends IndexBasedTreeAdapter
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (AWTTreeAdapter.class);

    // }}}

    // {{{ getParentNode

    /**
     * Get the parent of a node.
     *
     * @param   child   The node to query for the parent.
     *
     * @return  The parent of the node.
     */
    public Object getParentNode(Object child)
    {
        return ((Component) child).getParent();
    }

    // }}}
    // {{{ isLeaf

    /**
     * Query whether a node is a leaf.
     *
     * @param   node    The node to query.
     *
     * @return  True if the node is a leaf.
     */
    public boolean isLeaf(Object node)
    {
        return ! (node instanceof Container);
    }

    // }}}

    //----------------------------------------------------------------------
    // Index based
    //----------------------------------------------------------------------
    // {{{ getChildCount

    /**
     * Get the number of children of a node.
     *
     * @param   parent  The node to query.
     *
     * @return  The number of children of the node.
     */
    public int getChildCount(Object parent)
    {
        if (parent instanceof JMenu) {
            return ((JMenu) parent).getMenuComponentCount();
        }
        if (parent instanceof Container) {
            return ((Container) parent).getComponentCount();
        }
        return 0;
    }

    // }}}
    // {{{ getChildNode

    /**
     * Get a child of a node.
     *
     * @param   parent  The node to query.
     * @param   index   The index of the child.
     *
     * @return  The child at the index.
     *
     * @throws   IndexOutOfBoundsException If the index is invalid.
     */
    public Object getChildNode(Object parent, int index)
         throws IndexOutOfBoundsException
    {
        if (parent instanceof JMenu) {
            return ((JMenu) parent).getMenuComponents()[index];
        }
        if (! (parent instanceof Container)) {
            throw new IndexOutOfBoundsException
                ("parent " + parent + " is not a Container");
        }
        return ((Container) parent).getComponents()[index];
    }

    // }}}
    // {{{ getIndexOfChildNode

    /**
     * Get the index of a child in a node's list of children.
     *
     * @param   parent  The node to query.
     * @param   child   The child to look for.
     *
     * @return  The index of the child or -1 if it is not a child of this node.
     */
    public int getIndexOfChildNode(Object parent, Object child)
    {
        if (! (parent instanceof Container)) {
            return -1;
        }
        Component[] children = ((Container) parent).getComponents();
        for (int i = 0; i < children.length; i++) {
            if (children[i].equals(child)) {
                return i;
            }
        }
        return -1;
    }

    // }}}

}
