// {{{ copyright

/********************************************************************
 *
 * $Id: DefaultTableModelSorter.jpp,v 1.12 2000/06/06 16:41:25 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.gui;

// {{{ imports

import java.text.Collator;

import java.io.Serializable;


import java.util.ArrayList;
import java.util.List;

import java.util.Properties;

import javax.swing.table.TableModel;

import de.qfs.lib.config.Configurable;
import de.qfs.lib.config.Configuration;
import de.qfs.lib.config.RegistrationHelper;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * A default implementation of the TableModelSorter interface. <p>
 *
 * The column class of the sort column determines the comparison for
 * the rows. String columns are compared via a Collator, Comparable's
 * via the Comparable interface and the rest via a Collator on their
 * toString values.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.12 $
 */
public class DefaultTableModelSorter
    implements TableModelSorter, Configurable
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (DefaultTableModelSorter.class);

    /**
     * The TableModelSorterListeners.
     */
    protected List listeners = new ArrayList ();

    /**
     * The column to sort on.
     */
    protected int column = -1;

    /**
     * Whether sort order is ascending or descending.
     */
    protected boolean ascending;

    /**
     * Whether setting of the column determines ascending order.
     */
    protected boolean autoAscending = true;

    /**
     * The configuration name.
     */
    protected String configurableName;

    /**
     * Helper for registration callbacks.
     */
    private RegistrationHelper regHelper = new RegistrationHelper ();

    // }}}

    // {{{ constructors

    /**
     * Create a new DefaultTableModelSorter. The initial column to sort
     * is -1, meaning no order.
     */
    public DefaultTableModelSorter()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "DefaultTableModelSorter(TableModel)", "");
        }
    }

    /**
     * Create a new DefaultTableModelSorter. The initial column to sort
     * is -1, meaning no order.
     *
     * @param   name    The {@link Configurable Configurable} name of the
     *                  DefaultTableModelSorter.
     */
    public DefaultTableModelSorter(String name)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "DefaultTableModelSorter(TableModel,String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }

        configurableName = name;
    }

    // }}}

    //----------------------------------------------------------------------
    // Parameter access
    //----------------------------------------------------------------------
    // {{{ setSortColumn

    /**
     * Set the column to sort in model coordinates. If {@link
     * #setAutoAscending autoAscending} is turned on ascending or descending
     * order will be determined as follows:
     * <ul>
     * <li>If the column changes, ascending sort order will be used</li>
     * <li>Otherwise the sort order for the old and new column is reversed</li>
     * </ul>
     *
     * @param   column  The column to sort on. -1 means no order.
     *
     * @see     #setAutoAscending
     */
    public final void setSortColumn(int column)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setSortColumn(int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "column: " + column);
        }

        if (autoAscending) {
            if (this.column == column) {
                ascending = ! ascending;
            } else {
                ascending = true;
            }
        }
        this.column = column;
        fireSortOrderChanged(new TableModelSorterEvent
                             (this, column, ascending));
    }

    // }}}
    // {{{ getSortColumn

    /**
     * Get the column being sorted on.
     *
     * @return  The column being sorted on.
     */
    public final int getSortColumn()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getSortColumn()", "");
        }
        return column;
    }

    // }}}
    // {{{ setSortAscending

    /**
     * Define whether the rows should be sorted in ascending or descending
     * order.
     *
     * @param   ascending       True for ascending order.
     */
    public final void setSortAscending(boolean ascending)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setSortAscending(boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "ascending: " + ascending);
        }
        this.ascending = ascending;
        fireSortOrderChanged(new TableModelSorterEvent
                             (this, column, ascending));
    }

    // }}}
    // {{{ isSortAscending

    /**
     * Query whether the rows are sorted in ascending order.
     *
     * @return  True for ascending sort order.
     */
    public final boolean isSortAscending()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "isSortAscending()", "");
        }
        return ascending;
    }

    // }}}
    // {{{ setAutoAscending

    /**
     * If autoAscending is set to true (the default), the sort order is
     * determined by {@link #setSortColumn setSortColumn}.  Otherwise sort
     * order has to be set explicitly through {@link #setSortAscending
     * setSortAscending}.
     *
     * @param   auto    True for autoAscending.
     */
    public final void setAutoAscending(boolean auto)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setAutoAscending(boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "auto: " + auto);
        }
        this.autoAscending = auto;
    }

    // }}}
    // {{{ isAutoAscending

    /**
     * Query whether autoAscending is turned on.
     *
     * @return  True if autoAscending is turned on.
     */
    public final boolean isAutoAscending()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "isAutoAscending()", "");
        }
        return autoAscending;
    }

    // }}}

    //----------------------------------------------------------------------
    // The TableModelSorter interface
    //----------------------------------------------------------------------
    // {{{ compare

    /**
     * Compare two rows of the table.
     *
     * @param   model   The original TableModel.
     * @param   idx1    Index of the first row.
     * @param   idx2    Index of the second row.
     *
     * @return  0 if the rows are equal, -1 if the first row is less than
     *          the second, 1 if it is greater.
     */
    public int compare(TableModel model, int idx1, int idx2)
    {
        // logging is too expensive - enable only for debugging
        //  if (logger.level >= Log.MTD) {
            //  logger.log(Log.MTD, "compare(int,int)",
                       //  logger.level < Log.MTDDETAIL ? "" :
                       //  "idx1: " + idx1 + ", " +
                       //  "idx2: " + idx2);
        //  }

        if (model == null
            || column < 0
            || column >= model.getColumnCount()) {
            return 0;
        }

        Class columnClass = model.getColumnClass(column);
        Object o1 = model.getValueAt(idx1, column);
        Object o2 = model.getValueAt(idx2, column);

        int ret;

        if (columnClass == String.class) {
            // Most common case: String
            //  if (logger.level >= Log.DBGDETAIL) {
                //  logger.log(Log.DBGDETAIL, "compare(int,int)",
                           //  "Strings - 1: " + o1 + ", 2: " + o2);
            //  }
            ret = Collator.getInstance().compare((String) o1, (String) o2);
        } else if (Comparable.class.isAssignableFrom(columnClass)) {
            // Common in JDK 1.2, probably very uncommon in JDK 1.1
            ret = ((Comparable) o1).compareTo((Comparable) o2);
            //  if (logger.level >= Log.DBGDETAIL) {
                //  logger.log(Log.DBGDETAIL, "compare(int,int)",
                           //  "Comparables - 1: " + o1 + ", 2: " + o2);
            //  }
        } else {
            // default:
            ret = Collator.getInstance().compare(o1.toString(), o2.toString());
            //  if (logger.level >= Log.DBGDETAIL) {
                //  logger.log(Log.DBGDETAIL, "compare(int,int)",
                           //  "toString - 1: " + o1 + ", 2: " + o2);
            //  }
        }
        //  if (logger.level >= Log.DBG) {
            //  logger.log(Log.DBG, "compare(int,int)",
                       //  "result: " + (ascending ? ret : -ret));
        //  }

        return ascending ? ret : -ret;
    }

    // }}}
    // {{{ addTableModelSorterListener

    /**
     * Add a TableModelSorterListener to the TableModelSorter.
     *
     * @param   listener        The TableModelSorterListener to add.
     */
    public void addTableModelSorterListener(TableModelSorterListener listener)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "addTableModelSorterListener(TableModelSorterListener)",
                       "listener: " + listener);
        }
        listeners.add(listener);
    }

    // }}}
    // {{{ removeTableModelSorterListener

    /**
     * Remove a TableModelSorterListener from the TableModelSorter.
     *
     * @param   listener        The TableModelSorterListener to remove.
     */
    public void removeTableModelSorterListener
        (TableModelSorterListener listener)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "removeTableModelSorterListener" +
                       "(TableModelSorterListener)",
                       "listener: " + listener);
        }
        listeners.remove(listener);
    }

    // }}}

    // {{{ fireSortOrderChanged

    /**
     * Notify the listeners of a change in the sort order.
     *
     * @param   e       The event containing the details.
     */
    protected void fireSortOrderChanged(TableModelSorterEvent e)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "fireSortOrderChanged(TableModelSorterEvent)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "e: " + e);
        }
        TableModelSorterListener[] tms = (TableModelSorterListener[])
            listeners.toArray(new TableModelSorterListener[0]);
        for (int i = tms.length - 1; i >= 0; i--) {
            tms[i].sortOrderChanged(e);
        }
        regHelper.stateChanged(this);
    }

    // }}}

    //----------------------------------------------------------------------
    // The Configurable interface
    //----------------------------------------------------------------------
    // {{{ getConfigurableName

    /**
     * Get the DefaultTableModelSorter's configuration name.
     *
     * @return The configuration name.
     */
    public final String getConfigurableName()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getConfigurableName()", "");
        }
        return configurableName;
    }

    // }}}
    // {{{ setConfigurableName

    /**
     * Set the DefaultTableModelSorter's configuration name.
     *
     * @param   name    The name to set.
     */
    public final void setConfigurableName(String name)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setConfigurableName(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }
        configurableName = name;
    }

    // }}}
    // {{{ getConfigurableState

    /**
     * Get the DefaultTableModelSorter's configurable state.
     *
     * @return    The configurable state.
     */
    public Properties getConfigurableState()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getConfigurableState()", "");
        }
        Properties state = new Properties ();
        state.put("column", "" + column);
        state.put("ascending", "" + ascending);
        return state;
    }

    // }}}
    // {{{ setConfigurableState

    /**
     * Set the DefaultTableModelSorter's configurable state.
     *
     * @param   state   The state to set.
     */
    public void setConfigurableState(Properties state)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setConfigurableState(Properties)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "state: " + state);
        }

        String col = state.getProperty("column");
        String asc = state.getProperty("ascending");
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "setConfigurableState(Properties)",
                       "column: " + col +
                       ", ascending: " + asc);
        }
        if (col == null || asc == null) {
            if (logger.level >= Log.WRN) {
                logger.log(Log.WRN, "setConfigurableState(Properties)",
                           "Invalid state: " +
                           "column: " + col +
                           ", ascending: " + asc);
            }
            return;
        }

        try {
            // avoid inversion of ascending through setSortColumn
            boolean tmp = autoAscending;
            autoAscending = false;
            ascending = Boolean.valueOf(asc).booleanValue();
            setSortColumn(Integer.parseInt(col));
            autoAscending = tmp;
        } catch (NumberFormatException ex) {
            logger.log(Log.WRN, "setConfigurableState(Properties)", ex);
        }
    }

    // }}}
    // {{{ registrationChanged

    /**
     * This callback is used to notify the Configurable whenever it is
     * registered with or unregistered from the Configuration, so it can
     * decide whether to notify the Configuration when its state changes.
     *
     * @param   config  The affected Configuration.
     * @param   name    The registration name of the Configurable which may
     *                  be different from the Configurable's name.
     * @param   registered True if the Configurable was just registered,
     *                  false if it was unregistered.
     * @param   valuesKept Whether the Configurable's values were kept, if
     *                  it was just unregistered.
     */
    public void registrationChanged(Configuration config, String name,
                                    boolean registered, boolean valuesKept)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "registrationChanged(Configuration,String," +
                       "boolean,boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "registered: " + registered);
        }
        regHelper.registrationChanged(config, registered);
    }

    // }}}
}
