// {{{ copyright

/********************************************************************
 *
 * $Id: FilteredAndSortedTableModel.jpp,v 1.18 2000/08/22 09:09:56 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.gui;

// {{{ imports

import javax.swing.ListSelectionModel;
import javax.swing.table.TableModel;
import javax.swing.event.TableModelEvent;


import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.ArrayList;
import java.util.ListIterator;


import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * A TableModelAdapter that enables filtering and sorting. <p>
 *
 * This is the class that does all the hard work needed to implement tables
 * with filter and sort mechanisms. You will rarely need to use this class
 * directly as the {@link SortedTableHelper SortedTableHelper} provides
 * everything that is necessary to set up filtered and sorted table. <p>
 *
 * The methods of interest to end users are {@link #getMappedRow
 * getMappedRow}, {@link #getAddedRows getAddedRows} and {@link
 * #clearAddedRows clearAddedRows}.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.18 $
 */
public class FilteredAndSortedTableModel
    extends TableModelAdapter
    implements TableModelSorterListener, TableModelFilterListener
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private static Logger logger =
        new Logger (FilteredAndSortedTableModel.class);

    /**
     * The filter for the data.
     */
    protected TableModelFilter filter;

    /**
     * The sorter for the data.
     */
    protected TableModelSorter sorter;

    /**
     * The sorter as Comparator.
     */
    protected SorterCompare sorterCompare = new SorterCompare ();

    /**
     * A list of indexes into the original TableModel.
     */
    public List indexMap;

    /**
     * Whether indexMap must be sorted before data can be accessed.
     */
    protected boolean mustSort;

    /**
     * Whether a rebuild is currently pending.
     */
    //  private boolean rebuildPending;

    /**
     * An array mapping original TableModel indices to mapped inices. Only
     * used temporarily, not updated regularly.
     */
    public int[] reverseMap;

    /**
     * The extra space the reverseMap was built with.
     */
    public int reverseExtra;

    /**
     * The rows added to the model.
     */
    private List addedRows = new ArrayList ();

    /**
     * SelectionModel whose selection to save.
     */
    private ListSelectionModel savedSelectionModel;

    /**
     * The selected Rows of the saved selectionModel
     */
    private List savedRows = new ArrayList ();

    /**
     * The saved lead selection index
     */
    private int savedLead = -1;

    /**
     * The saved anchor selection index
     */
    private int savedAnchor = -1;

    /**
     * Whether the selection has been saved.
     */
    private boolean selectionSaved;

    /**
     * The relative amount of changed rows in the underlying model that
     * triggers a complete rebuild
     */
    private double criticalSizeFactor = 0.3;

    // }}}

    // {{{ constructors

    /**
     * Construct a new FilteredAndSortedTableModel without filter and sorter.
     *
     * @param   model   The TableModel to wrap.
     */
    public FilteredAndSortedTableModel (TableModel model)
    {
        super (model);
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "FilteredAndSortedTableModel(TableModel)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "model: " + model);
        }
    }

    /**
     * Construct a new FilteredAndSortedTableModel.
     *
     * @param   model   The TableModel to wrap.
     * @param   filter  The TableModelFilter for the data.
     * @param   sorter  The TableModelSorter that imposes the order on
     *                  the data.
     */
    public FilteredAndSortedTableModel (TableModel model,
                                        TableModelFilter filter,
                                        TableModelSorter sorter)
    {
        super (model);
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "FilteredAndSortedTableModel(TableModel," +
                       "TableModelFilter,TableModelSorter)",
                       "model: " + model + ", " +
                       "filter: " + filter + ", " +
                       "sorter: " + sorter);
        }
        this.filter = filter;
        if (filter != null) {
            filter.addTableModelFilterListener(this);
        }
        this.sorter = sorter;
        if (sorter != null) {
            sorter.addTableModelSorterListener(this);
        }
        mustSort = true;
    }

    // }}}

    //----------------------------------------------------------------------
    // The filter
    //----------------------------------------------------------------------
    // {{{ getFilter

    /**
     * Get the filter for the data.
     *
     * @return  The filter.
     */
    public final TableModelFilter getFilter ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getFilter()", "");
        }
        return filter;
    }

    // }}}
    // {{{ setFilter

    /**
     * Set the filter for the data. This invalidates the mapping of table
     * rows and causes a complete rebuild.
     *
     * @param   filter  The filter to set.
     */
    public void setFilter (TableModelFilter filter)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setFilter(TableModelFilter)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "filter: " + filter);
        }
        if (this.filter != null) {
            this.filter.removeTableModelFilterListener(this);
        }
        saveSelection();
        this.filter = filter;
        if (filter != null) {
            filter.addTableModelFilterListener(this);
        }
        indexMap = null;
        mustSort = (sorter != null);
        invalidateMap();
        //  if (savedSelectionModel != null) {
            //  validateMap();
            //  restoreSavedSelection();
        //  }
    }

    // }}}
    // {{{ sortOrderChanged

    /**
     * Notify the FilteredAndSortedTableModel of a change in the sort order.
     *
     * @param   e       The event holding the details.
     */
    public void sortOrderChanged(TableModelSorterEvent e)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "sortOrderChanged(TableModelSorterEvent)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "e: " + e);
        }
        saveSelection();
        mustSort = true;
        invalidateMap();
        //  if (savedSelectionModel != null) {
            //  validateMap();
            //  restoreSavedSelection();
        //  }
    }

    // }}}

    //----------------------------------------------------------------------
    // The sorter
    //----------------------------------------------------------------------
    // {{{ getSorter

    /**
     * Get the TableModelSorter for the data.
     *
     * @return  The TableModelSorter.
     */
    public final TableModelSorter getSorter ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getSorter()", "");
        }
        return sorter;
    }

    // }}}
    // {{{ setSorter

    /**
     * Set the TableModelSorter for the data. This invalidates the row mapping
     * and causes a full sort.
     *
     * @param   sorter  The TableModelSorter to set.
     */
    public void setSorter (TableModelSorter sorter)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setSorter(TableModelSorter)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "sorter: " + sorter);
        }
        saveSelection();
        if (this.sorter != null) {
            this.sorter.removeTableModelSorterListener(this);
            if (sorter == null) {
                // must rebuild, since we cannot sort with new sorter
                indexMap = null;
            }
        }
        this.sorter = sorter;
        if (sorter != null) {
            sorter.addTableModelSorterListener(this);
            mustSort = true;
        } else {
            mustSort = false;
        }
        invalidateMap();
        //  if (savedSelectionModel != null) {
            //  validateMap();
            //  restoreSavedSelection();
        //  }
    }

    // }}}
    // {{{ filterChanged

    /**
     * Notify the FilteredAndSortedTableModel of a change in the
     * TableModelFilter.
     *
     * @param   event   Details about the change.
     */
    public void filterChanged(TableModelFilterEvent event)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "filterChanged(TableModelFilterEvent)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "event: " + event);
        }
        if (filter == event.getSource()) {
            saveSelection();
            indexMap = null;
            mustSort = (sorter != null);
            invalidateMap();
            //  if (savedSelectionModel != null) {
                //  validateMap();
                //  restoreSavedSelection();
            //  }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // The TableModel interface
    //----------------------------------------------------------------------
    // {{{ getValueAt

    /**
     * Return an attribute value for a cell.
     *
     * @param   row     The row whose value is to be looked up.
     * @param   column  The column whose value is to be looked up.
     *
     * @return  The value Object at the specified cell.
     */
    public Object getValueAt(int row, int column)
    {
        validateMap();
        if (row >= indexMap.size()) {
            logger.log(Log.ERR, "getValueAt(int,int)",
                       "Illegal row: " + row +
                       ", current size: " + indexMap.size());
            logger.dumpStack(Log.ERR, "getValueAt(int,int)", "StackTrace");
            return null;
        }
        return model.getValueAt(((Integer)indexMap.get(row)).intValue(),
                                column);
    }

    // }}}
    // {{{ setValueAt

    /**
     * Set the attribute value for a cell.
     *
     * @param   value   The value to set.
     * @param   row     The row whose value is to be set.
     * @param   column  The column whose value is to be set.
     */
    public void setValueAt(Object aValue, int row, int column)
    {
        validateMap();
        if (row >= indexMap.size()) {
            logger.log(Log.ERR, "setValueAt(Object,int,int)",
                       "Illegal row: " + row +
                       ", current size: " + indexMap.size());
            logger.dumpStack(Log.ERR, "setValueAt(Object,int,int)",
                             "StackTrace");
            return;
        }
        model.setValueAt(aValue, ((Integer)indexMap.get(row)).intValue(),
                         column);
    }

    // }}}
    // {{{ getRowCount

    /**
     * Get the number of rows in the table.
     *
     * @return  The number of rows.
     */
    public int getRowCount()
    {
        if (filter == null) {
            return model.getRowCount();
        } else {
            validateMap();
            return indexMap.size();
        }
    }

    // }}}
    // {{{ isCellEditable

    /**
     * Get the number of columns in the table.
     *
     * @return  The number of columns.
     */
    public boolean isCellEditable(int row, int column)
    {
        validateMap();
        if (row >= indexMap.size()) {
            // the JTable KeyEventHandler is stupid enough to ask for
            // non-existant rows :-(
            return false;
        }
        return model.isCellEditable(((Integer)indexMap.get(row)).intValue(),
                                    column);
    }

    // }}}
    // {{{ tableChanged

    /**
     * Called, when the wrapped model has changed in some way. Forward
     * the event to all registered TableModelListeners.
     *
     * @param   event   The event that happened.
     */
    public void tableChanged(TableModelEvent e)
    {
        int start = e.getFirstRow();
        int end = e.getLastRow();
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "tableChanged(TableModelEvent)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "e: " + e +
                       ", start: " + start +
                       ", end: " + end);
        }

        if (start == e.HEADER_ROW) {
            // structure change - invalidates everything
            indexMap = null;
            addedRows.clear();
            reverseMap = null;
            mustSort = (sorter != null);
            selectionSaved = false;
            super.tableChanged(new TableModelEvent (this, e.HEADER_ROW));
            return;
        }

        if (start == 0 && end >= model.getRowCount()
            && e.getType() == TableModelEvent.UPDATE) {
            // table data has changed completely
            indexMap = null;
            addedRows.clear();
            reverseMap = null;
            mustSort = (sorter != null);
            selectionSaved = false;
            invalidateMap();
            return;
        }

        saveSelection();
        switch (e.getType()) {
        case TableModelEvent.UPDATE:
            rowsUpdated(start, end, e);
            break;
        case TableModelEvent.INSERT:
            rowsAdded(start, end, e);
            break;
        case TableModelEvent.DELETE:
            rowsRemoved(start, end, e);
            break;
        }
        reverseMap = null;

        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "tableChanged(TableModelEvent)",
                       "savedSelectionModel: " + savedSelectionModel +
                       ", selectionSaved: " + selectionSaved);
        }
        if (savedSelectionModel != null
            && selectionSaved) {
            validateMap();
            restoreSavedSelection();
        }
    }

    // }}}

    // {{{ fireAdded

    /**
     * Fire an event that rows have been added. Also update addedRows
     *
     * @param   from    Index of he first row added.
     * @param   to      Index of he last row added.
     */
    private void fireAdded(int from, int to)
    {
        int num = to - from + 1;
        boolean done = false;

        for (ListIterator i = addedRows.listIterator(); i.hasNext();) {
            int idx = ((Integer) i.next()).intValue();
            if (idx >= from) {
                // insert new values before this one
                i.previous();
                for (int j = from; j <= to; j++) {
                    i.add(new Integer (j));
                }
                // move the rest
                for (; i.hasNext();) {
                    idx = ((Integer) i.next()).intValue();
                    i.set (new Integer (idx + num));
                }
                done = true;
                break;
            }
        }
        if (!done) {
            // just add
            for (int j = from; j <= to; j++) {
                addedRows.add(new Integer (j));
            }
        }

        super.tableChanged (new TableModelEvent
                            (this,
                             from, to,
                             TableModelEvent.ALL_COLUMNS,
                             TableModelEvent.INSERT));
    }


    // }}}
    // {{{ fireRemoved

    /**
     * Fire an event that rows have been removed. Also update addedRows
     *
     * @param   from    Index of he first row removed.
     * @param   to      Index of he first last removed.
     */
    private void fireRemoved(int from, int to)
    {
        int num = to - from + 1;

        for (ListIterator i = addedRows.listIterator(); i.hasNext();) {
            int idx = ((Integer) i.next()).intValue();
            if (idx >= from) {
                if (idx <= to) {
                    // remove value if inside region
                    i.remove();
                } else {
                    // move the rest
                    i.set (new Integer (idx - num));
                    for (; i.hasNext();) {
                        idx = ((Integer) i.next()).intValue();
                        i.set (new Integer (idx - num));
                    }
                    break;
                }
            }
        }

        super.tableChanged (new TableModelEvent
                            (this,
                             from, to,
                             TableModelEvent.ALL_COLUMNS,
                             TableModelEvent.DELETE));
    }


    // }}}

    //----------------------------------------------------------------------
    // Additional methods
    //----------------------------------------------------------------------
    // {{{ getMappedRow

    /**
     * Get the mapped row index for the underlying TableModel.
     *
     * @param   row     The index in the FilteredAndSortedTableModel.
     *
     * @return  The corresponding index in the wrapped TableModel or -1,
     *          if row is not a valid index.
     */
    public int getMappedRow (int row)
    {
        // too expensive
        //  if (Flags.LOG && TRACE) {
            //  logger.log(Log.MTD, "getMappedRow(int)",
                       //  logger.level < Log.MTDDETAIL ? "" :
                       //  "row: " + row);
        //  }
        validateMap();
        if (row >= 0 && row < indexMap.size()) {
            return ((Integer)indexMap.get(row)).intValue();
        }
        return -1;
    }

    // }}}
    // {{{ getAddedRows

    /**
     * Due to filtering and sorting, the result of adding rows to a TableModel
     * that is wrapped by a FilteredAndSortedTableModel is hardly predictable.
     * This method queries the FilteredAndSortedTableModel for the rows that
     * have been added since its creation or the last call to one of {@link
     * #clearAddedRows clearAddedRows}, {@link #setSorter setSorter}, {@link
     * #sortOrderChanged sortOrderChanged}, {@link #setFilter setFilter} or
     * {@link #filterChanged filterChanged}.  This method does NOT clear that
     * state.
     *
     * @return  An integer array containing the indices of the added rows.
     */
    public int[] getAddedRows()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getAddedRows()", "");
        }
        int[] ret = new int [addedRows.size()];
        for (int i = 0; i < ret.length; i++) {
            ret[i] = ((Integer) addedRows.get(i)).intValue();
        }
        return ret;
    }


    // }}}
    // {{{ clearAddedRows

    /**
     * Reset the buffer for added rows.
     */
    public void clearAddedRows()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "clearAddedRows()", "");
        }
        addedRows.clear();
    }

    // }}}

    //----------------------------------------------------------------------
    // The selection
    //----------------------------------------------------------------------
    // {{{ setSavedSelectionModel

    /**
     * The FilteredAndSortedTableModel can keep track of the selection state in
     * one ListSelectionModel and keep it up to date during changes to the
     * underlying TableModel.
     *
     * @param   model   The ListSelectionModel whose selection to save.
     */
    public void setSavedSelectionModel(ListSelectionModel model)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setSavedSelectionModel(ListSelectionModel)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "model: " + model);
        }
        savedSelectionModel = model;
    }


    // }}}
    // {{{ saveSelection

    /**
     * Save selected rows.
     */
    private void saveSelection()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "saveSelection()", "this: " + this);
        }
        if (selectionSaved) {
            return;
        }
        if (savedSelectionModel != null) {
            if (indexMap == null || mustSort) {
                savedLead = -1;
                savedAnchor = -1;
                savedRows.clear();
            } else {
                savedLead = savedSelectionModel.getLeadSelectionIndex();
                savedAnchor = savedSelectionModel.getAnchorSelectionIndex();
                savedRows.clear();
                int idx1 = savedSelectionModel.getMinSelectionIndex();
                int idx2 = savedSelectionModel.getMaxSelectionIndex();
                // we need to compare to savedLead before mapping it to model
                // coordinates
                if (idx1 >= 0) {
                    for (int i = idx1; i <= idx2; i++) {
                        if (i != savedLead
                            && savedSelectionModel.isSelectedIndex(i)
                            && i < indexMap.size()) {
                            savedRows.add(indexMap.get(i));
                            if (logger.level >= Log.DBG) {
                                logger.log(Log.DBG, "saveSelection()",
                                           "selected row " + i +
                                           " maps to " + indexMap.get(i));
                            }
                        }
                    }
                }
                // now we can map savedLead and savedAnchor
                if (savedLead >= 0
                    && savedSelectionModel.isSelectedIndex(savedLead)
                    && savedLead < indexMap.size()) {
                    savedLead = ((Integer) indexMap.get(savedLead)).intValue();
                } else {
                    savedLead = -1;
                }
                // anchor is 0 even if list is empty!
                if (savedAnchor >= 0 && savedAnchor < indexMap.size()) {
                    savedAnchor =
                        ((Integer) indexMap.get(savedAnchor)).intValue();
                } else {
                    savedAnchor = -1;
                }
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "saveSelection()",
                               "savedLead: " + savedLead +
                               ", savedAnchor: " + savedAnchor);
                }
            }
            selectionSaved = true;
        }
    }

    // }}}
    // {{{ restoreSavedSelection

    /**
     * Restore the saved selection rows.
     */
    private void restoreSavedSelection()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "restoreSavedSelection()", "");
        }
        if (savedSelectionModel != null
            && selectionSaved) {
            buildReverseMap(0);
            savedSelectionModel.clearSelection();
            savedSelectionModel.setValueIsAdjusting(true);
            for (ListIterator i = savedRows.listIterator(); i.hasNext(); ) {
                int idx = ((Integer) i.next()).intValue();
                if (reverseMap[idx] >= 0) {
                    savedSelectionModel.addSelectionInterval
                        (reverseMap[idx], reverseMap[idx]);
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG, "restoreSavedSelection()",
                                   "saved index " + idx +
                                   " maps to " + reverseMap[idx]);
                    }
                } else {
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG, "restoreSavedSelection()",
                                   "saved index " + idx + " is unmapped");
                    }
                }
            }
            if (savedLead >= 0 && reverseMap[savedLead] >= 0) {
                //  savedSelectionModel.setLeadSelectionIndex
                //  (reverseMap[savedLead]);
                savedSelectionModel.addSelectionInterval
                    (reverseMap[savedLead], reverseMap[savedLead]);
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "restoreSavedSelection()",
                               "savedLead maps to " + reverseMap[savedLead]);
                }

            }
            savedSelectionModel.setValueIsAdjusting(false);
            if (savedAnchor >= 0 && reverseMap[savedAnchor] >= 0) {
                savedSelectionModel.setAnchorSelectionIndex
                    (reverseMap[savedAnchor]);
                //  savedSelectionModel.addSelectionInterval
                    //  (reverseMap[savedAnchor], reverseMap[savedAnchor]);
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "restoreSavedSelection()",
                               "savedAnchor maps to " +
                               reverseMap[savedAnchor]);
                }
            }
            savedRows.clear();
            selectionSaved = false;
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ buildMap

    /**
     * Build the map from new rows to original rows.
     */
    private void buildMap()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "buildMap()", "");
        }
        int count = model.getRowCount();
        indexMap = new ArrayList();

        for (int i = 0; i < count; i++) {
            if (filter == null || filter.filter (model, i)) {
                indexMap.add (new Integer(i));
            }
        }
    }

    // }}}
    // {{{ validateMap

    /**
     * Make sure that indexMap exists and holds the correct values.
     */
    private void validateMap()
    {
        // called by getValueAt - too often
        //  if (Flags.LOG && TRACE) {
            //  logger.log(Log.MTD, "validateMap()", "");
        //  }
        if (indexMap == null) {
            buildMap();
            reverseMap = null;
        }
        if (sorter != null && mustSort) {
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "validateMap()",
                           "Sorting");
            }
            Collections.sort (indexMap, sorterCompare);
            mustSort = false;
            reverseMap = null;
        }
    }

    // }}}
    // {{{ invalidateMap

    /**
     * Notify the registered TableModelListeners that the mapping is
     * no longer valid.
     */
    private void invalidateMap()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "invalidateMap()", "");
        }
        reverseMap = null;
        addedRows.clear();
        super.tableChanged (new TableModelEvent (this));
        if (savedSelectionModel != null
            && selectionSaved) {
            validateMap();
            restoreSavedSelection();
        }
    }

    // }}}
    // {{{ rowsAdded

    /**
     * Some rows have been added to the wrapped TableModel.
     *
     * @param   start   The first new row.
     * @param   end     The last new row.
     * @param   e       The event of the change.
     */
    private void rowsAdded(int start, int end, TableModelEvent e)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "rowsAdded(int,int,TableModelEvent)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "start: " + start + ", " +
                       "end: " + end + ", " +
                       "e: " + e);
        }
        int num = end - start + 1;

        // Adjust saved selection
        if (savedSelectionModel != null) {
            for (ListIterator i = savedRows.listIterator(); i.hasNext();) {
                int idx = ((Integer) i.next()).intValue();
                if (idx >= start) {
                    i.set (new Integer (idx + num));
                }
            }
            if (savedLead >= start) {
                savedLead += num;
            }
            if (savedAnchor >= start) {
                savedAnchor += num;
            }
        }

        // We cannot check criticalSizeFactor here since addedRows needs to be
        // updated
        if (indexMap == null
            || mustSort) {
            indexMap = null;
            mustSort = (sorter != null);
            invalidateMap();
            return;
        }

        // Adjust the old mapping
        moveIndices (start, num);

        // Check which of the new rows must be added and sort them
        List added = getFilteredAndSortedPart (start, end);
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "rowsAdded(int,int,TableModelEvent)",
                       "Rows added: " + start + "-" + end);
        }
        insertRows(added);
    }

    // }}}
    // {{{ rowsRemoved

    /**
     * Some rows have been removed from the wrapped TableModel.
     *
     * @param   start   The first removed row.
     * @param   end     The last removed row.
     * @param   e       The event of the change.
     */
    private void rowsRemoved(int start, int end, TableModelEvent e)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "rowsRemoved(int,int,TableModelEvent)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "start: " + start + ", " +
                       "end: " + end + ", " +
                       "e: " + e);
        }
        int num = end - start + 1;

        // Adjust saved selection
        if (savedSelectionModel != null) {
            for (ListIterator i = savedRows.listIterator(); i.hasNext();) {
                int idx = ((Integer) i.next()).intValue();
                if (idx >= start) {
                    if (idx <= end) {
                        // remove value if inside region
                        i.remove();
                    } else {
                        // move the rest
                        i.set (new Integer (idx - num));
                    }
                }
            }
            if (savedLead >= start) {
                if (savedLead <= end) {
                    savedLead = -1;
                } else {
                    savedLead -= num;
                }
            }
            if (savedAnchor >= start) {
                if (savedLead <= end) {
                    savedAnchor = -1;
                } else {
                    savedAnchor -= num;
                }
            }
        }

        if (indexMap == null
            || mustSort
            || num > model.getRowCount() * criticalSizeFactor) {
            indexMap = null;
            mustSort = (sorter != null);
            invalidateMap();
            return;
        }

        buildReverseMap(end - start + 1);
        // Check which of the new rows must be added and sort them
        List removed = findFilteredAndSortedPart (start, end);
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "rowsRemoved(int,int,TableModelEvent)",
                       "Rows removed: " + start + "-" + end);
        }

        deleteRows(removed);
        // Adjust the old mapping
        moveIndices (end + 1, -num);
    }

    // }}}
    // {{{ rowsUpdated

    /**
     * Some rows have been updated in the wrapped TableModel.
     *
     * @param   start   The first updated row.
     * @param   end     The last updated row.
     * @param   e       The event of the change.
     */
    private void rowsUpdated(int start, int end, TableModelEvent e)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "rowsUpdated(int,int,TableModelEvent)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "start: " + start + ", " +
                       "end: " + end + ", " +
                       "e: " + e);
        }

        if (indexMap == null
            || mustSort
            || end - start >= model.getRowCount() * criticalSizeFactor) {
            indexMap = null;
            mustSort = (sorter != null);
            invalidateMap();
            return;
        }

        buildReverseMap(0);
        // Get the current indizes for the rows
        List updated = findFilteredAndSortedPart (start, end);

        List added = new ArrayList ();
        List removed = new ArrayList ();

        for (int i = start; i <= end; i++) {
            boolean filtered = (filter != null && !filter.filter(model, i));
            Integer mapped = (Integer) updated.get(i - start);
            if (mapped.intValue() == -1) {
                if (filtered) {
                    // stays unmapped
                    continue;
                } else {
                    // gets mapped
                    added.add(new Integer (i));
                }
            } else {
                if (filtered) {
                    // gets unmapped
                    removed.add(mapped);
                } else if (sorter != null) {
                    // has changed, must be moved
                    removed.add(mapped);
                    added.add(new Integer (i));
                }
            }
        }

        Collections.sort(removed, new IntegerCompare ());
        deleteRows(removed);

        Collections.sort(added, new IntegerCompare ());
        insertRows(added);
    }

    // }}}
    // {{{ addRow

    /**
     * Add one row to the table.
     *
     * @param   idx     The index of the row in the wrapped TableModel.
     *
     * @return  The new corresponding index in indexMap.
     */
    private int addRow (Integer idx)
    {
        // called too often
        //  if (Flags.LOG && TRACE) {
            //  logger.log(Log.MTD, "addRow(Integer)",
                       //  logger.level < Log.MTDDETAIL ? "" :
                       //  "idx: " + idx);
        //  }
        int newIdx;
        if (sorter == null) {
            // XXX Java 1.2 doesn't need this since Integer implements
            // Comparable
            newIdx = Collections.binarySearch (indexMap, idx,
                                               new IntegerCompare());
        } else {
            newIdx = Collections.binarySearch (indexMap, idx, sorterCompare);
        }
        if (newIdx < 0) {
            newIdx = -newIdx - 1;
        }
        indexMap.add(newIdx, idx);
        return newIdx;
    }

    // }}}
    // {{{ insertRows

    /**
     * Some rows have been added
     *
     * @param   added   List of model indices of the added rows, sorted.
     */
    private void insertRows(List added)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "insertRows(List)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "added: " + added.size() + " rows");
        }
        // Try to add new rows in contiguous blocks to minimize the number of
        // TableModelEvents that have to be generated.
        int oldstart = -2;
        int oldend = -2;
        for (int i = 0; i < added.size(); i++) {
            Integer tmp = (Integer)added.get(i);
            int idx = addRow (tmp);
            if (idx != oldend + 1) {
                if (oldstart >= 0) {
                    fireAdded(oldstart, oldend);
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG,
                                   "insertRows(List)",
                                   "changed: " + oldstart + "-" + oldend);
                    }
                }
                oldstart = idx;
                oldend = idx;
            } else {
                oldend++;
            }
        }
        if (oldstart >= 0) {
            fireAdded(oldstart, oldend);
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG,
                           "insertRows(List)",
                           "changed: " + oldstart + "-" + oldend);
            }
        }
    }

    // }}}
    // {{{ deleteRows

    /**
     * Some rows have been removed
     *
     * @param   removed List of mapped indices of the removed rows, sorted.
     */
    private void deleteRows(List removed)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "deleteRows(List)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "removed: " + removed);
        }
        // Try to remove rows in contiguous blocks to minimize the number of
        // TableModelEvents that have to be generated.
        int oldstart = -2;
        int oldend = -2;
        for (int i = removed.size() - 1; i >= 0; i--) {
            int idx = ((Integer) removed.get(i)).intValue();
            if (idx < 0) {
                continue;
            }
            indexMap.remove(idx);
            if (idx != oldstart - 1) {
                if (oldstart >= 0) {
                    fireRemoved(oldstart, oldend);
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG,
                                   "deleteRows(List)",
                                   "changed: " + oldstart + "-" + oldend);
                    }
                }
                oldstart = idx;
                oldend = idx;
            } else {
                oldstart--;
            }
        }
        if (oldstart >= 0) {
            fireRemoved(oldstart, oldend);
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG,
                           "deleteRows(List)",
                           "changed: " + oldstart + "-" + oldend);
            }
        }
    }

    // }}}
    // {{{ moveIndices

    /**
     * Some rows have been inserted into or removed from the underlying
     * TableModel, so the indices in the indexMap have to be updated
     * to reflect this change.
     *
     * @param   start   Every index greater or equal to this value will
     *                  be changed.
     * @param   num     The amount to change the indices.
     */
    private void moveIndices (int start, int num)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "moveIndices(int,int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "start: " + start + ", " +
                       "num: " + num);
        }
        if (indexMap != null) {
            for (ListIterator i = indexMap.listIterator(); i.hasNext();) {
                int idx = ((Integer) i.next()).intValue();
                if (idx >= start) {
                    i.set (new Integer (idx + num));
                }
            }
        }
    }

    // }}}
    // {{{ getFilteredAndSortedPart

    /**
     * Filter and sort some rows from the wrapped model.
     *
     * @param   start   The first row to consider.
     * @param   end     The last row to consider.
     *
     * @return  A List holding the remaining indices as Integers.
     */
    private List getFilteredAndSortedPart (int start, int end)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getFilteredAndSortedPart(int,int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "start: " + start + ", " +
                       "end: " + end);
        }
        List list = new ArrayList();
        for (int i = start; i <= end; i++) {
            if (filter == null || filter.filter(model, i)) {
                list.add (new Integer(i));
            }
        }
        if (sorter != null) {
            Collections.sort (list, sorterCompare);
        }
        return list;
    }

    // }}}
    // {{{ findFilteredAndSortedPart

    /**
     * Find some rows from the wrapped model. The caller must ensure that
     * reverseMap is uptodate.
     *
     * @param   start   The first row to consider.
     * @param   end     The last row to consider.
     *
     * @return  A List holding the remaining indices as Integers.
     */
    private List findFilteredAndSortedPart (int start, int end)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "findFilteredAndSortedPart(int,int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "start: " + start + ", " +
                       "end: " + end);
        }
        List list = new ArrayList();
        for (int i = start; i <= end; i++) {
            int idx = reverseMap[i];
            list.add(new Integer(idx));
        }
        Collections.sort (list, new IntegerCompare ());
        return list;
    }

    // }}}
    // {{{ buildReverseMap

    /**
     * Build a reverse map from rows in the model to the indexMap.
     *
     * @param   extra   Extra space for the reverseMap if rows have been
     *                  deleted.
     */
    private void buildReverseMap(int extra)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "buildReverseMap(int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "extra: " + extra);
        }
        if (reverseMap == null || extra > reverseExtra) {
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "buildReverseMap(int)",
                           "rebuilding");
            }
            reverseMap = new int[model.getRowCount() + extra];
            for (int i = 0; i < reverseMap.length; i++) {
                reverseMap[i] = -1;
            }
            for (int i = 0; i < indexMap.size(); i++) {
                reverseMap[((Integer) indexMap.get(i)).intValue()] = i;
            }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper classes
    //----------------------------------------------------------------------
    // {{{ class SorterCompare

    /**
     * Turn the sorter into a Comparator.
     */
    private class SorterCompare implements Comparator
    {
        /**
         * Compare two objects.
         *
         * @param       o1     The first object.
         * @param       o2     The second object.
         *
         * @return      -1 if o1 < o2, 0 if o1 = o2, 1 if o1 > o2.
         */
        public int compare(Object o1, Object o2)
        {
            int i1 = ((Integer)o1).intValue();
            int i2 = ((Integer)o2).intValue();

            return sorter.compare(model, i1, i2);
        }
    }

    // }}}
    // {{{ class IntegerCompare

    /**
     * Comparator for Integers.
     */
    private class IntegerCompare implements Comparator
    {
        /**
         * Compare two objects.
         *
         * @param       o1     The first object.
         * @param       o2     The second object.
         *
         * @return      -1 if o1 < o2, 0 if o1 = o2, 1 if o1 > o2.
         */
        public int compare(Object o1, Object o2)
        {
            int i1 = ((Integer)o1).intValue();
            int i2 = ((Integer)o2).intValue();

            return (i1 == i2) ? 0 : ((i1 < i2) ? -1 : 1);
        }
    }


    // }}}
}
