// {{{ copyright

/********************************************************************
 *
 * $Id: FocusFixer.jpp,v 1.7 2001/02/04 12:19:45 gs Exp $
 *
 * Copyright (C) 1999 Gregor Schmid, Quality First Software
 * All rights reserved
 *
 *******************************************************************/

// }}}

package de.qfs.lib.gui;

// {{{ imports

import java.awt.Component;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.KeyEvent;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;


import java.util.Collections;
import java.util.HashMap;
import java.util.Map;


import javax.swing.DefaultFocusManager;
import javax.swing.FocusManager;
import javax.swing.SwingUtilities;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * This class fixes an annoying bug that only occurs in JDK 1.1 with the motif
 * Toolkit. In that constellation, whenever a window loses the focus it
 * forgets which component had the focus last. When the window regains the
 * focus, no component is focused and navigation with the tab key restarts at
 * the first component. <p>
 *
 * To fix this problem, FocusFixer extends {@link
 * javax.swing.DefaultFocusManager DefaultFocusManager} to keep track of
 * focused components and implements {@link java.awt.event.WindowListener
 * WindowListener} in order to return the focus to the right component when a
 * window is opened or activated. <p>
 *
 * To fix focus management for any Window, simply get a FocusFixer instance
 * with {@link #instance instance} and use its {@link #addWindow addWindow}
 * method. Don't forget to remove the window with {@link #removeWindow
 * removeWindow} when you are getting rid of it.<p>
 *
 * You can use FocusFixer regardless of what JDK you use and what operating
 * system resp. what AWT Toolkit you are running on. If it is not the JDK 1.1
 * motif combination, FocusFixer will silently ignore all calls.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.7 $
 * @since       0.97.0
 */
public class FocusFixer
    extends DefaultFocusManager
    implements WindowListener
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (FocusFixer.class);

    /**
     * Whether the fix is needed for the current JDK Toolkit combination. Null
     * means not decide yet.
     */
    private static Boolean isNeeded;

    /**
     * Dummy instance used for JDK 1.2  or non-Motif systems.
     */
    private static FocusFixer dummyInstance;

    /**
     * Map from ThreadGroups to FocusFixer instances.
     */
    private static Map instances = new HashMap ();

    /**
     * Map from windows to their focus owners.
     */
    private Map focus = Collections.synchronizedMap(new HashMap ());

    // }}}

    // {{{ constructor

    /**
     * Create a new FocusFixer. Hidden, use the {@link #instance instance}
     * to get a FocusFixer.
     */
    private FocusFixer ()
    {
    }

    // }}}

    // {{{ instance

    /**
     * Get the FocusFixer instance for the ThreadGroup of the calling
     * Thread. If no FocusFixer for that ThreadGroup has been created yet,
     * create a new one and register it as the FocusManager for that
     * ThreadGroup.
     *
     * @return  The FocusFixer instance.
     *
     * @see     javax.swing.FocusManager#setCurrentManager
     */
    public static synchronized FocusFixer instance()
    {
        if (! isNeeded()) {
            if (dummyInstance == null) {
                dummyInstance = new FocusFixer ();
            }
            return dummyInstance;
        }

        ThreadGroup group = Thread.currentThread().getThreadGroup();
        FocusFixer instance = (FocusFixer) instances.get(group);
        if (instance == null) {
            instance = new FocusFixer ();
            instances.put(group, instance);
            FocusManager.setCurrentManager(instance);
        }
        return instance;
    }

    // }}}

    // {{{ addWindow

    /**
     * Register a window with a FocusFixer so it will keep track of the focus
     * owner and restore the focus whenever the window is openened or
     * activated. You should call {@link #removeWindow removeWindow} when the
     * window is disposed in order to clean up all references to it.
     *
     * @param   window  The window to register.
     */
    public void addWindow(Window window)
    {
        if (! isNeeded()) {
            return;
        }
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "addWindow(Window)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "window: " + window);
        }

        if (focus.containsKey(window)) {
            if (logger.level >= Log.WRN) {
                logger.log(Log.WRN, "addWindow(Window)",
                           "Already added: " + window);
            }
            return;
        }

        focus.put(window, SwingUtilities.findFocusOwner(window));
        window.addWindowListener(this);
    }

    // }}}
    // {{{ removeWindow

    /**
     * Remove a window from a FocusFixer so it will no longer handle its focus.
     *
     * @param   window  The window to remove.
     */
    public void removeWindow(Window window)
    {
        if (! isNeeded()) {
            return;
        }
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "removeWindow(Window)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "window: " + window);
        }

        if (!focus.containsKey(window)) {
            if (logger.level >= Log.WRN) {
                logger.log(Log.WRN, "removeWindow(Window)",
                           "Unknown window: " + window);
            }
        } else {
            focus.remove(window);
        }
        window.removeWindowListener(this);
    }

    // }}}

    //----------------------------------------------------------------------
    // Window events
    //----------------------------------------------------------------------
    // {{{ windowActivated

    /**
     * Restore the focus to the recent focus owner.
     *
     * @param   e       The event containing the details.
     */
    public void windowActivated(WindowEvent e)
    {
        fixFocus((Window) e.getSource());
    }

    // }}}
    // {{{ windowOpened

    /**
     * Restore the focus to the recent focus owner.
     *
     * @param   e       The event containing the details.
     */
    public void windowOpened(WindowEvent e)
    {
        fixFocus((Window) e.getSource());
    }

    // }}}
    // {{{ windowDeactivated

    /**
     * Does nothing.
     *
     * @param   e       The event containing the details.
     */
    public void windowDeactivated(WindowEvent e)
    {
    }

    // }}}
    // {{{ windowClosed

    /**
     * Does nothing.
     *
     * @param   e       The event containing the details.
     */
    public void windowClosed(WindowEvent e)
    {
    }

    // }}}
    // {{{ windowClosing

    /**
     * Does nothing.
     *
     * @param   e       The event containing the details.
     */
    public void windowClosing(WindowEvent e)
    {
    }

    // }}}
    // {{{ windowIconified

    /**
     * Does nothing.
     *
     * @param   e       The event containing the details.
     */
    public void windowIconified(WindowEvent e)
    {
    }

    // }}}
    // {{{ windowDeiconified

    /**
     * Does nothing.
     *
     * @param   e       The event containing the details.
     */
    public void windowDeiconified(WindowEvent e)
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // FocusManager
    //----------------------------------------------------------------------
    // {{{ processKeyEvent

    /**
     * Process one key event. If the component is inside a Window that has
     * been registered with the FocusFixer, remember it as the current focus
     * owner.
     *
     * @param   com     The component on which the key event occured.
     * @param   e       The event.
     */
    public void processKeyEvent (Component com, KeyEvent e)
    {
        Window win = SwingUtilities.windowForComponent(com);

        if (focus.containsKey(win)) {
            focus.put(win, com);
        }
        super.processKeyEvent(com, e);
    }

    // }}}
    // {{{ focusNextComponent

    /**
     * Focus the next component. If the component is inside a Window that has
     * been registered with the FocusFixer, remember the new focus owner.
     *
     * @param   com     Focus the component that is next in the focus order
     *                  from this one.
     */
    public void focusNextComponent (Component com)
    {
        Window win = SwingUtilities.windowForComponent(com);
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "focusNextComponent(Component)",
                       "win: " + win +
                       " \ncom: " + com);
        }
        super.focusNextComponent(com);
        Component owner = SwingUtilities.findFocusOwner(win);
        if (focus.containsKey(win)) {
            focus.put(win, owner);
        }
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "focusNextComponent(Component)",
                       "newOwner: " + owner);
        }
    }

    // }}}
    // {{{ focusPreviousComponent

    /**
     * Focus the previous component. If the component is inside a Window that
     * has been registered with the FocusFixer, remember the new focus owner.
     *
     * @param   com     Focus the component that is the previous one from
     *                  this one in the focus order.
     */
    public void focusPreviousComponent (Component com)
    {
        Window win = SwingUtilities.windowForComponent(com);
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "focusPreviousComponent(Component)",
                       "win: " + win +
                       " \ncom: " + com);
        }
        super.focusPreviousComponent(com);
        if (focus.containsKey(win)) {
            focus.put(win, SwingUtilities.findFocusOwner(win));
        }
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "focusPreviousComponent(Component)",
                       "newOwner: " + SwingUtilities.findFocusOwner(win));
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ isNeeded

    /**
     * Check whether the fix is needed for the current JDK/Toolkit version.
     *
     * @return  True if the fix is needed.
     */
    private static boolean isNeeded()
    {
        if (isNeeded == null) {
            String version = System.getProperty("java.version");
            // paranoia
            if (version == null) {
                version = "1.1.";
            }
            Toolkit tk = Toolkit.getDefaultToolkit();
            isNeeded = new Boolean
                (version.startsWith("1.1.")
                 && tk.getClass().getName().indexOf(".motif.") >= 0);
        }
        return isNeeded.booleanValue();
    }

    // }}}
    // {{{ fixFocus

    /**
     * Fix the focus for a Window.
     *
     * @param   win     The window to fix.
     */
    private void fixFocus(Window win)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "fixFocus(Window)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "win: " + win);
        }
        Component com = (Component) focus.get(win);
        Component current = SwingUtilities.findFocusOwner(win);
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "fixFocus(Window)",
                       "current owner: " + current +
                       " \nremembered owner: " + com);
        }
        if (current == null || current == win) {
            if (com != null) {
                com.requestFocus();
            } else {
                focusNextComponent(win);
            }
        }
    }

    // }}}
}
