// {{{ copyright

/********************************************************************
 *
 * $Id: Grid.java,v 1.15 2000/11/15 19:02:44 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.gui;

// {{{ imports

import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;

import javax.swing.Box;
import javax.swing.JComponent;
import javax.swing.JPanel;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * A Grid extends JPanel to provide convenient layout of its children through
 * the use of (possibly nested) GridBagLayout(s). <p>
 *
 * This class is still very rough and experimental.
 *
 * XXX Todo:
 *
 * Rewrite Grid so it can recreate itself anytime, allowing lines to be
 * inserted later and add the components in the lines directly instead of
 * packing them into yet another panel.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.15 $
 */

public class Grid extends JPanel
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (Grid.class);

    /**
     * The number of rows contained in the Grid.
     */
    protected int lines = 0;

    /**
     * The layout of the Grid.
     */
    protected GridBagLayout layout;

    /**
     * The horizontalDistance of the Grid.
     */
    protected int horDist = 5;

    /**
     * The verticalDistance of the Grid.
     */
    protected int verDist = 5;

    /**
     * The glue of the Grid.
     */
    private Component glue;

    /**
     * The total accumulated y-weight.
     */
    protected double totalWeight;

    // }}}

    // {{{ constructors

    /**
     * Create a new Grid.
     */
    public Grid ()
    {
        layout = new GridBagLayout();
        setLayout (layout);
    }

    /**
     * Create a new Grid.
     *
     * @param   hdist   Default horizontal distance between children.
     * @param   vdist   Default vertical distance between children.
     */
    public Grid (int hdist, int vdist)
    {
        this ();
        horDist = hdist;
        verDist = vdist;
    }

    /**
     * Create a new Grid.
     *
     * @param   components      Two dimensional array of components to add
     *                          to the Grid.
     */
    public Grid (Component[][] components)
    {
        this ();
        for (int i = 0; i < components.length; i++) {
            addLine(components[i]);
        }
    }

    // }}}

    // {{{ addLine(Component[])

    /**
     * Add one row of Components to the Grid.
     *
     * @param   line    An array containig the Components to add.
     */
    public void addLine(Component[] line)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "addLine(Component[])",
                       logger.level < Log.MTDDETAIL ? "" :
                       "line: " + line);
        }
        Component com;
        if (line.length > 1) {
            com = makeLine(line);
        } else {
            com = line[0];
        }
        add(com, 0.0);
    }

    // }}}
    // {{{ addLine(Component[],double)

    /**
     * Add one row of Components to the Grid.
     *
     * @param   line    An array containig the Components to add.
     * @param   weighty The weight to use for the GridBagConstraints.
     */
    public void addLine(Component[] line, double weighty)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "addLine(Component[],double)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "line: " + line + ", " +
                       "weighty: " + weighty);
        }
        Component com;
        if (line.length > 1) {
            com = makeLine(line);
        } else {
            com = line[0];
        }
        add(com, weighty);
    }

    // }}}
    // {{{ addGrid

    /**
     * Add several rows of Components to the Grid. Components will be
     * added left packed, i.e. if one line has fewer Components than
     * another, the last Component in that line will fill the empty cells.
     *
     * @param   grid    A two dimensional array containing the Components
     *                  to add.
     */
    public void addGrid(Component[][] grid)
    {
        JPanel panel = new JPanel ();
        GridBagLayout layout = new GridBagLayout ();

        panel.setLayout(layout);

        GridBagConstraints cons = new GridBagConstraints ();
        cons.anchor = GridBagConstraints.NORTHWEST;

        int max = 0;
        for (int i = 0; i < grid.length; i++) {
            if (grid[i].length > max) {
                max = grid[i].length;
            }
        }

        int y = 0;
        for (int i = 0; i < grid.length; i++) {
            int x = 0;
            cons.gridy = y++;
            for (int j = 0; j < grid[i].length; j++) {
                cons.gridx = x++;
                if (j < grid[i].length - 1) {
                    cons.fill = GridBagConstraints.NONE;
                    cons.weightx = 0;
                    cons.gridwidth = 1;
                } else {
                    cons.fill = GridBagConstraints.HORIZONTAL;
                    cons.weightx = 1.0;
                    cons.gridwidth = GridBagConstraints.REMAINDER;
                }
                layout.setConstraints(grid[i][j], cons);
                panel.add(grid[i][j]);
                if (j < grid[i].length - 1 && horDist > 0) {
                    Component dist = Box.createHorizontalStrut(horDist);
                    cons.gridx = x++;
                    layout.setConstraints(dist, cons);
                    panel.add(dist);
                }
            }
            if (i < grid.length - 1 && verDist > 0) {
                Component dist = Box.createVerticalStrut(verDist);
                cons.gridy = y++;
                cons.gridx = 0;
                layout.setConstraints(dist, cons);
                panel.add(dist);
            }
        }
        add(panel, 0.0);
    }

    // }}}
    // {{{ add

    /**
     * Add one Component as a full row to the Grid.
     *
     * @param   com     The Component to add.
     * @param   weighty The weight to use for the GridBagConstraints.
     */
    public void add(Component com, double weighty)
    {
        GridBagConstraints cons = new GridBagConstraints ();
        cons.anchor = GridBagConstraints.NORTHWEST;
        cons.gridwidth = GridBagConstraints.REMAINDER;

        if (lines > 0 && verDist > 0) {
            Component dist = Box.createVerticalStrut(verDist);
            cons.fill = GridBagConstraints.HORIZONTAL;
            layout.setConstraints(dist, cons);
            add(dist);
        }

        if (weighty > 0) {
            cons.fill = GridBagConstraints.BOTH;
        } else {
            cons.fill =  GridBagConstraints.HORIZONTAL;
        }
        cons.weightx = 1.0;
        cons.weighty = weighty;
        layout.setConstraints(com, cons);
        add(com);
        lines++;
        totalWeight += weighty;
    }

    // }}}
    // {{{ addGlue

    /**
     * Add vertical glue to the Grid to provide proper spacing. Should be
     * called only once per Grid.
     */
    public void addGlue()
    {
        if (glue == null) {
            glue = Box.createVerticalGlue();
            add (glue, 1.0);
        }
    }

    // }}}
    // {{{ removeGlue

    /**
     * Add vertical glue to the Grid to provide proper spacing.
     */
    public void removeGlue()
    {
        if (glue != null) {
            remove(glue);
            glue = null;
            totalWeight -= 1.0;
        }
    }

    // }}}

    // {{{ getHorizontalDistance

    /**
     * Get the horizontalDistance of the Grid.
     *
     * @return  The horizontalDistance of the Grid.
     */
    public final int getHorizontalDistance()
    {
        return horDist;
    }

    // }}}
    // {{{ setHorizontalDistance

    /**
     * Set the horizontalDistance of the Grid.
     *
     * @param   horizontalDistance      The horizontalDistance to set.
     */
    public final void setHorizontalDistance(int horizontalDistance)
    {
        this.horDist = horizontalDistance;
    }

    // }}}
    // {{{ getVerticalDistance

    /**
     * Get the verticalDistance of the Grid.
     *
     * @return  The verticalDistance of the Grid.
     */
    public final int getVerticalDistance()
    {
        return verDist;
    }

    // }}}
    // {{{ setVerticalDistance

    /**
     * Set the verticalDistance of the Grid.
     *
     * @param   verticalDistance        The verticalDistance to set.
     */
    public final void setVerticalDistance(int verticalDistance)
    {
        this.verDist = verticalDistance;
    }

    // }}}
    // {{{ needGlue

    /**
     * Whether the Grid still needs some glue to provide proper spacing.
     *
     * @return  True if glue is still needed.
     */
    public final boolean needGlue()
    {
        return glue == null && totalWeight < 0.0001;
    }

    // }}}
    // {{{ getTotalWeight

    /**
     * Get the total accumulated y-weight of the Grid.
     *
     * @return  The total y-weight.
     */
    public double getTotalWeight()
    {
        return totalWeight;
    }

    // }}}

    // {{{ removeAll

    /**
     * Overridden to reset line count.
     */
    public void removeAll()
    {
        super.removeAll();
        lines = 0;
    }

    // }}}

    // {{{ makeLine

    /**
     * Pack Components into a row for the Grid. The Components will be sized
     * equally.
     *
     * @param   line    The array of Components.
     *
     * @return  The complete row.
     */
    protected Component makeLine (Component[] line)
    {
        Box box = Box.createHorizontalBox();

        int max = 0;
        boolean mustgrow = true;
        for (int i = 0; i < line.length; i++) {
            Component com = line[i];
            if (com.getMinimumSize().width > max) {
                max = com.getMinimumSize().width;
            }
            if (com.getMaximumSize().width >= 30000) {
                // This is a bit of a hack: The items are placed in a box that
                // doesn't grow them beyond their maximum size. This is OK, if
                // at least one item can grow infinitely which is typically
                // signaled by a max width of 32768 or MAX_INT.
                mustgrow = false;
            }
            box.add(com);
            if (i < line.length - 1 && horDist > 0) {
                Component dist = Box.createHorizontalStrut(horDist);
                box.add(dist);
            }
        }
        for (int i = 0; i < line.length; i++) {
            Component com = line[i];
            if (com instanceof JComponent) {
                ((JComponent) com).setPreferredSize
                (new Dimension (max, com.getPreferredSize().height));
                if (mustgrow) {
                    // If none of the items can grow significantly we
                    // make them all grow equally
                    ((JComponent) com).setMaximumSize
                        (new Dimension (Integer.MAX_VALUE,
                                        com.getMaximumSize().height));
                }
            }
        }
        return box;
    }

    // }}}
}
