// {{{ copyright

/********************************************************************
 *
 * $Id: Message.java,v 1.16 2001/02/27 10:04:31 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.gui;

// {{{ imports

import java.awt.Component;

import java.text.MessageFormat;

import java.util.ListResourceBundle;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.Vector;

import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JOptionPane;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.MapResourceBundle;

// }}}

// {{{ doc

/**
 * A collection of convenience methods to bring up localized versions
 * of standard dialogs. <p>
 *
 * Before any dialogs can be shown, the resources to use must be set with
 * the {@link #setResources setResources} method. The resource bundle must
 * contain at least the following values:
 * <TABLE BORDER="1" CELLPADDING="3" CELLSPACING="0" WIDTH="100%">
 * <TR>
 * <TD WIDTH="40%"><B>dialog.cancel.name</B></TD>
 * <TD>Default label for the Cancel button</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><B>dialog.no.name</B></TD>
 * <TD>Default label for the No button</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><B>dialog.ok.name</B></TD>
 * <TD>Default label for the OK button</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><B>dialog.yes.name</B></TD>
 * <TD>Default label for the Yes button</TD>
 * </TR>
 * </TABLE>
 *
 * These resources are provided as
 * <code>/de/qfs/lib/resources/properties/qflib</code> in
 * <code>qflib.jar</code> and added automatically during initialization of
 * the <code>Message</code> class.<p>
 *
 * Additionally for each message to display the following resources are
 * used:
 * <TABLE BORDER="1" CELLPADDING="3" CELLSPACING="0" WIDTH="100%">
 * <TR>
 * <TD WIDTH="40%"><B>&lt;messageName&gt;.title</B></TD>
 * <TD>The title for &lt;messageName&gt;</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><B>&lt;messageName&gt;.message</B></TD>
 * <TD>The message for &lt;messageName&gt;</TD>
 * </TR>
 * <TD WIDTH="40%"><B>&lt;messageName&gt;.custom</B></TD>
 * <TD>Number of custom objects to display. If this is set,
 * &lt;messageName&gt;.message may be omitted. The custom objects to display
 * are taken from the param array in the call to {@link
 * #showMessage(java.awt.Component,java.lang.String,java.lang.Object[])
 * showMessage}. See {@link javax.swing.JOptionPane JOptionPane}.</TD>
 * <TR>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><B>&lt;messageName&gt;.type</B></TD>
 * <TD>The message type for &lt;messageName&gt;. Possible values are
 *  plain, info, question, warning and error, which is the default.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><B>&lt;messageName&gt;.buttons</B></TD>
 * <TD>The buttons to use for &lt;messageName&gt;. Possible values are custom,
 *  okcancel, yesno, yesnocancel and ok, which is the default.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><B>&lt;messageName&gt;.button&lt;num&gt;</B></TD>
 * <TD>If button type custom was chosen, use one button&lt;num&gt; resource
 * for each button with num starting at 0 in the order in which they are to
 * appear.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><B>&lt;messageName&gt;.icon</B></TD>
 * <TD>The icon resource name for &lt;messageName&gt; (optional)</TD>
 * </TR>
 * </TABLE> <p>
 *
 * The Dialogs shown will have their name set depending on the type of the
 * message as follows:
 * <TABLE BORDER="1" CELLPADDING="3" CELLSPACING="0" WIDTH="100%">
 * <TR>
 * <TD WIDTH="40%"><B>JOptionPane.PLAIN_MESSAGE</B></TD>
 * <TD>"Message.message"</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><B>JOptionPane.INFORMATION_MESSAGE</B></TD>
 * <TD>"Message.info"</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><B>JOptionPane.QUESTION_MESSAGE</B></TD>
 * <TD>"Message.question"</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><B>JOptionPane.WARNING_MESSAGE</B></TD>
 * <TD>"Message.warning"</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><B>JOptionPane.ERROR_MESSAGE</B></TD>
 * <TD>"Message.error"</TD>
 * </TR>
 * </TABLE>
 *
 * @see JOptionPane
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.16 $
 */

// }}}
public abstract class Message
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (Message.class);

    /**
     * The resources for the messages.
     */
    private static MapResourceBundle rb = new MapResourceBundle ();

    // try to initialize resources
    static {
        try {
            rb.fetchProperties("/de/qfs/lib/resources/properties/qflib",
                               Message.class);
        } catch (MissingResourceException ex) {
            // so be it
        }
    }

    // }}}

    // {{{ constructor

    /**
     * Hidden constructor.
     */
    private Message()
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // The resources
    //----------------------------------------------------------------------
    // {{{ getResources

    /**
     * Get the resources of the Message class.
     *
     * @return  The resources.
     */
    public final static ResourceBundle getResources()
    {
        return rb;
    }

    // }}}
    // {{{ setResources

    /**
     * Set the resources of the Message class.
     *
     * @param   resources       The resources to set.
     */
    public final static void setResources(ResourceBundle resources)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setResources(ResourceBundle)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "resources: " + resources);
        }
        if (resources == null) {
            rb = new MapResourceBundle ();
        } else if (rb instanceof MapResourceBundle) {
            rb = (MapResourceBundle) resources;
        } else {
            rb = new MapResourceBundle (resources);
        }
    }

    // }}}
    // {{{ addResources

    /**
     * Add some resources of the Message class.
     *
     * @param   resources       The resources to add.
     */
    public final static void addResources(ResourceBundle resources)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "addResources(ResourceBundle)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "resources: " + resources);
        }
        rb.addResources(resources);
    }

    // }}}

    // {{{ showYesNo

    /**
     * Bring up a localized Yes/No dialog.
     *
     * @param   parent  A Component belonging to the parent frame for the
     *                  dialog.
     * @param   message The message to display.
     * @param   title   The title of the dialog.
     * @param   messageType     The type of the message, see
     *                          {@link JOptionPane JOptionPane}.
     * @param   icon    An icon for the message.
     *
     * @return  YES_OPTION, NO_OPTION or CLOSED_OPTION
     *          if the dialog is closed without selecting an option (values
     *          taken from {@link JOptionPane JOptionPane}).
     */
    public static int showYesNo (Component parent,
                                 Object message,
                                 String title,
                                 int messageType,
                                 Icon icon)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "showYesNoCancel(Component,String,String,int,Icon)",
                       "parent: " + parent + ", " +
                       "message: " + message + ", " +
                       "title: " + title + ", " +
                       "messageType: " + messageType + ", " +
                       "icon: " + icon);
        }

        Object[] options = new Object[]
        {rb.getString("dialog.yes.name", "Yes"),
         rb.getString("dialog.no.name", "No")};
        Object ret = show
            (parent, message, title, JOptionPane.YES_NO_OPTION,
             messageType, icon, options, options[0]);
        if (options[0].equals(ret)) {
            return JOptionPane.YES_OPTION;
        }
        if (options[1].equals(ret)) {
            return JOptionPane.NO_OPTION;
        }
        return JOptionPane.CLOSED_OPTION;
    }

    // }}}
    // {{{ showYesNoCancel

    /**
     * Bring up a localized Yes/No/Cancel dialog.
     *
     * @param   parent  A Component belonging to the parent frame for the
     *                  dialog.
     * @param   message The message to display.
     * @param   title   The title of the dialog.
     * @param   messageType     The type of the message, see
     *                          {@link JOptionPane JOptionPane}.
     * @param   icon    An icon for the message.
     *
     * @return  YES_OPTION, NO_OPTION, CANCEL_OPTION or CLOSED_OPTION
     *          if the dialog is closed without selecting an option (values
     *          taken from {@link JOptionPane JOptionPane}).
     */
    public static int showYesNoCancel (Component parent,
                                       Object message,
                                       String title,
                                       int messageType,
                                       Icon icon)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "showYesNoCancel(Component,String,String,int,Icon)",
                       "parent: " + parent + ", " +
                       "message: " + message + ", " +
                       "title: " + title + ", " +
                       "messageType: " + messageType + ", " +
                       "icon: " + icon);
        }

        Object[] options = new Object[]
        {rb.getString("dialog.yes.name", "Yes"),
         rb.getString("dialog.no.name", "No"),
         rb.getString("dialog.cancel.name", "Cancel")};
        Object ret = show
            (parent, message, title, JOptionPane.YES_NO_CANCEL_OPTION,
             messageType, icon, options, options[0]);
        if (options[0].equals(ret)) {
            return JOptionPane.YES_OPTION;
        }
        if (options[1].equals(ret)) {
            return JOptionPane.NO_OPTION;
        }
        return JOptionPane.CANCEL_OPTION;
    }

    // }}}
    // {{{ showOKCancel

    /**
     * Bring up a localized OK/Cancel dialog.
     *
     * @param   parent  A Component belonging to the parent frame for the
     *                  dialog.
     * @param   message The message to display.
     * @param   title   The title of the dialog.
     * @param   messageType     The type of the message, see
     *                          {@link JOptionPane JOptionPane}.
     * @param   icon    An icon for the message.
     *
     * @return  OK_OPTION, CANCEL_OPTION or CLOSED_OPTION
     *          if the dialog is closed without selecting an option (values
     *          taken from {@link JOptionPane JOptionPane}).
     */
    public static int showOKCancel (Component parent,
                                    Object message,
                                    String title,
                                    int messageType,
                                    Icon icon)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "showOKCancel(Component,String,String,int,Icon)",
                       "parent: " + parent + ", " +
                       "message: " + message + ", " +
                       "title: " + title + ", " +
                       "messageType: " + messageType + ", " +
                       "icon: " + icon);
        }

        Object[] options = new Object[]
        {rb.getString("dialog.ok.name", "OK"),
         rb.getString("dialog.cancel.name", "Cancel")};
        Object ret = show
            (parent, message, title,
             JOptionPane.OK_CANCEL_OPTION, messageType, icon,
             options, options[0]);
        if (options[0].equals(ret)) {
            return JOptionPane.OK_OPTION;
        }
        return JOptionPane.CANCEL_OPTION;
    }

    // }}}
    // {{{ showOK

    /**
     * Bring up a localized confirm dialog with OK as the single option.
     *
     * @param   parent  A Component belonging to the parent frame for the
     *                  dialog.
     * @param   message The message to display.
     * @param   title   The title of the dialog.
     * @param   messageType     The type of the message, see
     *                          {@link JOptionPane JOptionPane}.
     * @param   icon    An icon for the message.
     *
     * @return  OK_OPTION or CLOSED_OPTION
     *          if the dialog is closed without selecting an option (values
     *          taken from {@link JOptionPane JOptionPane}).
     */
    public static int showOK (Component parent,
                              Object message,
                              String title,
                              int messageType,
                              Icon icon)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "showOK(Component,String,String,int,Icon)",
                       "parent: " + parent + ", " +
                       "message: " + message + ", " +
                       "title: " + title + ", " +
                       "messageType: " + messageType + ", " +
                       "icon: " + icon);
        }

        Object[] options = new Object[]
        {rb.getString("dialog.ok.name", "OK")};
        Object ret = show
            (parent, message, title,
             JOptionPane.OK_CANCEL_OPTION, messageType, icon,
             options, options[0]);
        if (options[0].equals(ret)) {
            return JOptionPane.OK_OPTION;
        }
        return JOptionPane.CANCEL_OPTION;
    }

    // }}}
    // {{{ showMessage

    /**
     * Bring up a localized message dialog. The message, title, type and
     * icon for the dialog are taken from the resources set through
     * {@link #setResources setResources}.
     *
     * @param   parent  A Component belonging to the parent frame for the
     *                  dialog.
     * @param   resource        The basename of the resources to use.
     *
     * @return  OK_OPTION or CLOSED_OPTION
     *          if the dialog is closed without selecting an option (values
     *          taken from {@link JOptionPane JOptionPane}).
     */
    public static int showMessage(Component parent, String resource)
    {
        return showMessage(parent, resource, null);
    }

    // }}}
    // {{{ showMessage

    /**
     * Bring up a parameterized, localized message dialog.
     * The message, title, type and
     * icon for the dialog are taken from the resources set through
     * {@link #setResources setResources}.
     *
     * @param   parent  A Component belonging to the parent frame for the
     *                  dialog.
     * @param   resource        The basename of the resources to use.
     * @param param     Array of parameters for the message. If the resource
     *                  for custom is set, the first n objects are used as
     *                  custom objects. The rest is used to format the
     *                  message.
     *
     * @return  OK_OPTION or CLOSED_OPTION
     *          if the dialog is closed without selecting an option (values
     *          taken from {@link JOptionPane JOptionPane}).
     */
    public static int showMessage(Component parent, String resource,
                                  Object[] param)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "showMessage(Component,String,Object[])",
                       logger.level < Log.MTDDETAIL ? "" :
                       "parent: " + parent + ", " +
                       "resource: " + resource + ", " +
                       "param: " + param);
        }
        String message;
        int nCustom = 0;
        Object[] oCustom = null;
        String title;
        int mtype;
        Icon icon = null;
        Object display;

        nCustom = rb.getInt(resource + ".custom", 0);
        message = rb.getString(resource + ".message", null);
        if (message == null && nCustom == 0) {
            message = "MISSING: " + resource + ".message";
        }

        if (nCustom > 0) {
            if (param.length < nCustom) {
                throw new IllegalArgumentException
                    ("More custom objects needed than param array holds.");
            }
            oCustom = new Object[nCustom + (message == null ? 0 : 1)];
            System.arraycopy(param, 0, oCustom, message == null ? 0 : 1,
                             nCustom);
            if (param.length > nCustom) {
                Object[] tmp = new Object[param.length - nCustom];
                System.arraycopy(param, nCustom, tmp, 0,
                                 param.length - nCustom);
                param = tmp;
            } else {
                param = null;
            }
        }

        if (param != null && param.length > 0 && message != null) {
            MessageFormat mf = new MessageFormat (message);
            message = mf.format(param);
        }

        if (oCustom != null && message != null) {
            oCustom[0] = message;
        }

        if (oCustom == null) {
            display = message;
        } else {
            display = oCustom;
        }

        title = rb.getString(resource + ".title", resource + ".title");
        String type = rb.getString(resource + ".type", "error");
        if (type.equalsIgnoreCase("plain")) {
            mtype = JOptionPane.PLAIN_MESSAGE;
        } else if (type.equalsIgnoreCase("info")) {
            mtype = JOptionPane.INFORMATION_MESSAGE;
        } else if (type.equalsIgnoreCase("question")) {
            mtype = JOptionPane.QUESTION_MESSAGE;
        } else if (type.equalsIgnoreCase("warning")) {
            mtype = JOptionPane.WARNING_MESSAGE;
        } else {
            mtype = JOptionPane.ERROR_MESSAGE;
        }

        String iconName = rb.getString(resource + ".icon", null);
        if (iconName != null) {
            icon = new ImageIcon
                (ClassLoader.getSystemResource (iconName));
        }

        String buttons = rb.getString(resource + ".buttons", "ok");
        if (buttons.equals("okcancel")) {
            return showOKCancel(parent, display, title, mtype, icon);
        } else if (buttons.equals("yesno")) {
            return showYesNo(parent, display, title, mtype, icon);
        } else if (buttons.equals("yesnocancel")) {
            return showYesNoCancel(parent, display, title, mtype, icon);
        } else if (buttons.equals("custom")) {
            return showCustom(parent, resource, display, title, mtype, icon);
        } else {
            return showOK(parent, display, title, mtype, icon);
        }
    }

    // }}}
    // {{{ format

    /**
     * Convenience method that creates a message from a resource string
     * and its parameters.
     *
     * @param   resource        The name of the resource.
     * @param   param           The parameter values.
     *
     * @return  The formatted, localized string.
     */
    public static String format(String resource, Object[] param)
    {
        String msg = rb.getString(resource, resource);
        if (param == null) {
            return msg;
        }
        return new MessageFormat (msg).format(param);
    }

    // }}}

    // {{{ helpers

    // {{{ makeButtons

    //  private static void makeButtons()
    //  {
        //  if (buttonsInitialized) {
            //  return;
        //  }
        //  yesButton = new JButton (rb.getString("dialog.yes.name", "Yes"));
        //  yesButton.setMnemonic
            //  (rb.getString("dialog.yes.mnemonic", "y").charAt(0));
        //  noButton = new JButton (rb.getString("dialog.no.name", "No"));
        //  noButton.setMnemonic
            //  (rb.getString("dialog.no.mnemonic", "n").charAt(0));
        //  cancelButton =
            //  new JButton (rb.getString("dialog.cancel.name", "Cancel"));
        //  cancelButton.setMnemonic
            //  (rb.getString("dialog.cancel.mnemonic", "c").charAt(0));
        //  buttonsInitialized = true;
    //  }

    // }}}
    // {{{ show

    /**
     * Bring up the Dialog similar to {@link JOptionPane JOptionPane},
     * additionlly setting its name depending on the messageType.
     *
     * @return  The value returned by {@link JOptionPane#getValue
     *          JOptionPane.getValue}.
     */
    private static Object show(Component parent, Object message, String title,
                               int optionType, int messageType, Icon icon,
                               Object[] options, Object initialValue)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "show(Component,Object,String,int,int,Icon,Object[],Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "parent: " + parent + ", " +
                       "message: " + message + ", " +
                       "title: " + title + ", " +
                       "messageType: " + messageType + ", " +
                       "optionType: " + optionType + ", " +
                       "icon: " + icon + ", " +
                       "options: " + options + ", " +
                       "initialValue: " + initialValue);
        }
        JOptionPane pane = new JOptionPane (message, messageType, optionType,
                                            icon, options, initialValue);
        JDialog dialog = pane.createDialog(parent, title);
        switch(messageType) {
        case JOptionPane.PLAIN_MESSAGE:
            dialog.setName("Message.message");
        case JOptionPane.INFORMATION_MESSAGE:
            dialog.setName("Message.info");
        case JOptionPane.QUESTION_MESSAGE:
            dialog.setName("Message.question");
        case JOptionPane.WARNING_MESSAGE:
            dialog.setName("Message.warning");
        case JOptionPane.ERROR_MESSAGE:
            dialog.setName("Message.error");
        }
        dialog.show();
        return pane.getValue();
    }

    // }}}
    // {{{ showCustom

    /**
     * Bring up a localized confirm dialog with user defined buttons.
     *
     * @param   parent  A Component belonging to the parent frame for the
     *                  dialog.
     * @param   resource        The basename of the resources to use.
     * @param   message The message to display.
     * @param   title   The title of the dialog.
     * @param   messageType     The type of the message, see
     *                          {@link JOptionPane JOptionPane}.
     * @param   icon    An icon for the message.
     *
     * @return  CLOSED_OPTION if the dialog is closed without selecting an
     *          option, zero based button index otherwise.
     */
    private static int showCustom (Component parent,
                                   String resource,
                                   Object message,
                                   String title,
                                   int messageType,
                                   Icon icon)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD,
                       "showCustom(Component,String,String,String,int,Icon)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "parent: " + parent + ", " +
                       "resource: " + resource + ", " +
                       "message: " + message + ", " +
                       "title: " + title + ", " +
                       "messageType: " + messageType + ", " +
                       "icon: " + icon);
        }
        Vector tmp = new Vector();

        for (int i = 0; true; i++) {
            String bname = rb.getString(resource + ".button" + i, null);
            if (bname == null) {
                break;
            }
            tmp.addElement(bname);
        }

        if (tmp.size() == 0) {
            if (logger.level >= Log.WRN) {
                logger.log(Log.WRN,
                      "showCustom(Component,String,String,String,int,Icon)",
                           "Custom message but no button resources.");
            }
            return showOK(parent, message, title, messageType, icon);
        }

        Object[] options = new Object[tmp.size()];
        tmp.copyInto(options);

        Object ret = show
            (parent, message, title,
             JOptionPane.OK_CANCEL_OPTION, messageType, icon,
             options, options[0]);
        for (int i = 0; i < options.length; i++) {
            if (options[i].equals(ret)) {
                return i;
            }
        }
        return JOptionPane.CANCEL_OPTION;
    }

    // }}}

    // }}}
}
