// {{{ copyright

/********************************************************************
 *
 * $Id: StatusLine.java,v 1.14 2001/02/04 12:19:48 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.gui;

// {{{ imports

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JProgressBar;
import javax.swing.Timer;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.CountEvent;
import de.qfs.lib.util.CountListener;

// }}}

// {{{ doc

/**
 * The StatusLine is a component that is typically placed at the bottom
 * of a Frame and used to display status and other messages. <P>
 *
 * The StatusLine is built from a simple JComponent with a BorderLayout,
 * containing one JLabel in the center, so it can easily be extended to
 * display additional information. <P>
 *
 * These are some common uses for a status line:
 * <ul>
 * <li>Displaying a 'Ready' message when the application is idle</li>
 * <li>Displaying a 'Running...' message for the duration of an operation</li>
 * <li>Displaying some information like 'Do something ... done', that is
 *     removed after a certain delay</li>
 * <li>Displaying a message that explains the function of the MenuItem or
 * toolbar button the mouse cursor is over</li>
 * </ul>
 *
 * These are supported with {@link #setPermanentMessage setPermanentMessage},
 * {@link #pushMessage pushMessage} and {@link #setTemporaryMessage
 * setTemporaryMessage}. <p>
 *
 * The status line will be redrawn immediately after changing the message,
 * since standard AWT redraw during idle time is typically too late. <p>
 *
 * New since <em>qflib</em> version 0.96.0 is support for a progress bar which
 * can be displayed with {@link #showProgressBar showProgressBar} and turned
 * off with {@link #removeProgressBar removeProgressBar}. The progress value
 * can be set with {@link #setProgressValue setProgressValue}. Alternatively
 * StatusLine now implements the {@link de.qfs.lib.util.CountListener
 * CountListener} interface and can be registered with a {@link
 * de.qfs.lib.util.CountingInputStream CountingInputStream}.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.14 $
 */

// }}}
public class StatusLine
    extends JComponent
    implements CountListener
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (StatusLine.class);

    /**
     * The Label displaying the messages.
     */
    protected JLabel message = new JLabel ();

    /**
     * The permanent message.
     */
    private String permanentMessage = " ";

    /**
     * The semi permanent message.
     */
    private String semiPermanentMessage = permanentMessage;

    /**
     * The current message.
     */
    private String currentMessage = permanentMessage;

    /**
     * The ActionListener that resets the status line to the previous message.
     */
    private ActionListener reset = new ActionListener () {
        public void actionPerformed(ActionEvent e) {
            currentMessage = semiPermanentMessage;
            forceUpdate();
        }
    };

    /**
     * The Timer responsible for resetting the StatusLine.
     */
    private Timer timer;

    /**
     * The default delay for the timer.
     */
    private static final int DEFAULT_DELAY = 1500;

    /**
     * Optional progress bar.
     *
     * @since   0.96.0
     */
    protected JProgressBar progress;

    /**
     * Whether the progress bar is currently shown.
     *
     * @since   0.96.0
     */
    protected boolean progressShowing;

    /**
     * Whether to display a percent string in the progress bar.
     *
     * @since   0.96.0
     */
    protected boolean showPercent;

    // }}}

    // {{{ constructor

    /**
     * Create a new StatusLine.
     */
    public StatusLine ()
    {
        message.setText(permanentMessage);
        message.setBorder(BorderFactory.createLoweredBevelBorder());
        timer = new Timer (DEFAULT_DELAY, reset);
        timer.setRepeats(false);
        timer.stop();
        init();
    }

    // }}}

    // {{{ init

    /**
     * Initialize the StatusLine. This method uses a BorderLayout with the
     * message Label at the center. It can be replaced or extended to create
     * a customized StatusLine.
     */
    protected void init()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "init()", "");
        }
        setLayout(new BorderLayout ());
        add(message, "Center");
    }


    // }}}

    // {{{ getMessageLabel

    /**
     * Get the JLabel component that displays the status message.
     *
     * @return  The StatusLine's message label.
     */
    public final JLabel getMessageLabel()
    {
        return message;
    }

    // }}}

    // {{{ setPermanentMessage

    /**
     * Set the message to display in the StatusLine whenever no other
     * message is displayed. The display is updated only, if the permanent
     * message is the active one.
     *
     * @param   msg     The permanent message to set.
     */
    public void setPermanentMessage(String msg)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setPermanentMessage(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "msg: " + msg);
        }
        boolean current = (permanentMessage == currentMessage);
        permanentMessage = msg;
        if (current) {
            currentMessage = permanentMessage;
            forceUpdate();
        }
    }

    // }}}
    // {{{ pushMessage

    /**
     * Display a message until it is either removed with {@link #popMessage
     * popMessage} or replaced with another call to pushMessage.
     *
     * @param   msg     The message to display.
     */
    public void pushMessage(String msg)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "pushMessage(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "msg: " + msg);
        }
        semiPermanentMessage = msg;
        currentMessage = msg;
        timer.stop();
        forceUpdate();
    }

    // }}}
    // {{{ popMessage

    /**
     * Remove the semi permanent message set with {@link #pushMessage
     * pushMessage} and display the permanent message.
     */
    public void popMessage()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "popMessage()", "");
        }
        if (currentMessage != permanentMessage) {
            semiPermanentMessage = permanentMessage;
            currentMessage = permanentMessage;
            timer.stop();
            forceUpdate();
        }
    }

    // }}}
    // {{{ setTemporaryMessage

    /**
     * Display a message temporarily. After a delay it will be replaced with
     * the previous message.
     *
     * @param   msg     The message to display.
     */
    public void setTemporaryMessage(String msg)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setTemporaryMessage(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "msg: " + msg);
        }
        currentMessage = msg;
        forceUpdate();
        timer.restart();
    }

    // }}}
    // {{{ setTemporaryMessageDelay

    /**
     * Set the delay in milliseconds after which to replace the temporary
     * message with the previous message. Default is 1500 ms.
     *
     * @param   delay   The delay to set.
     */
    public void setTemporaryMessageDelay(int delay)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setTemporaryMessageDelay(int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "delay: " + delay);
        }
        timer.setInitialDelay(delay);
    }

    // }}}
    // {{{ forceUpdate

    /**
     * Force redraw of StatusLine
     *
     * @since   0.96.0
     */
    public void forceUpdate()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "forceUpdate()", "");
        }
        message.setText(currentMessage);
        SwingUtil.invokeAndWait(new Runnable () {
            public void run() {
                //  message.paint(message.getGraphics());
                Graphics g = getGraphics();
                if (g != null) {
                    Dimension size = getSize();
                    g.setClip(0, 0, size.width, size.height);
                    g.clearRect(0, 0, size.width, size.height);
                    update(g);
                }
            }
        });
    }

    // }}}

    //----------------------------------------------------------------------
    // ProgrssBar and CountListener interface
    //----------------------------------------------------------------------
    // {{{ showProgressBar

    /**
     * Display a progress bar.
     *
     * @param   min             Minimum value for the progress bar.
     * @param   max             Maximum value for the progress bar.
     * @param   where           BorderLayout constraint, e.g. "West".
     * @param   showPercent     Whether to display a percent string inside the
     *                          progress bar.
     *
     * @since   0.96.0
     */
    public void showProgressBar(int min, int max, String where,
                                boolean showPercent)
    {
        if (progress == null) {
            progress = new JProgressBar ();
            progress.setBorder(BorderFactory.createLoweredBevelBorder());
        }

        progress.setMinimum(min);
        progress.setMaximum(max);
        this.showPercent = showPercent;
        progress.setStringPainted(showPercent);
        add(progress, where);
        validate();
        progressShowing = true;
        setProgressValue(min);
        forceUpdate();
    }

    // }}}
    // {{{ removeProgressBar

    /**
     * Remove the progress bar from the StatusLine.
     *
     * @since   0.96.0
     */
    public void removeProgressBar()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "removeProgressBar()", "");
        }
        if (progressShowing) {
            remove(progress);
            validate();
            progressShowing = false;
            forceUpdate();
        }
    }

    // }}}
    // {{{ setProgressValue

    /**
     * Set the value for the progress bar and force a redisplay of the
     * StatusLine.
     *
     * @param   value   The new value.
     *
     * @since   0.96.0
     */
    public void setProgressValue(int value)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setProgressValue(int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "value: " + value);
        }
        if (progressShowing) {
            progress.setValue(value);
            if (showPercent) {
                progress.setString
                    ("" + ((int) (progress.getPercentComplete() * 100)) + "%");
            }
            SwingUtil.invokeAndWait(new Runnable () {
                public void run() {
                    //  message.paint(message.getGraphics());
                    Graphics g = progress.getGraphics();
                    if (g != null) {
                        Dimension size = progress.getSize();
                        g.setClip(0, 0, size.width, size.height);
                        g.clearRect(0, 0, size.width, size.height);
                        progress.paint(g);
                    }
                }
            });
        }
    }

    // }}}
    // {{{ countChanged

    /**
     * Notify the progress bar of a new value.
     *
     * @param   event   The event containing the details.
     *
     * @since   0.96.0
     */
    public void countChanged(CountEvent event)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "countChanged(CountEvent)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "event: " + event);
        }
        setProgressValue((int) event.getCount());
    }

    // }}}
    // {{{ getProgressBar

    /**
     * Get the progress bar of the status line.
     *
     * @return  The status line's progress bar.
     *
     * @since   0.98.0
     */
    public JProgressBar getProgressBar()
    {
        return progress;
    }

    // }}}
    // {{{ isProgressShowing

    /**
     * Query whther the progess bar is currently showing.
     *
     * @return  True if the progess bar is showing.
     *
     * @since   0.98.0
     */
    public boolean isProgressShowing()
    {
        return progressShowing;
    }

    // }}}
}
