// {{{ copyright

/********************************************************************
 *
 * $Id: RemoteLogFilter.java,v 1.2 2000/10/26 15:04:12 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.logrmi;

// {{{ imports

import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.UnknownHostException;

import java.rmi.AccessException;
import java.rmi.AlreadyBoundException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.ServerNotActiveException;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.LogEntry;
import de.qfs.lib.log.Logger;
import de.qfs.lib.log.LogUser;

// }}}

// {{{ doc

/**
 * The RemoteLogFilter class takes care of all the necessary details to
 * connect to a {@link LogServer LogServer} and pass the log messages created
 * by a program to it.  The convenience methods {@link #logRemote logRemote}
 * and {@link #stopLogging stopLogging} handle all the details. <p>
 *
 * There are two possible modes of operation: Contact a server as soon as the
 * RemoteLogFilter is created or register with the RMI registry to get
 * contacted later by a LogServer. The most useful mode of operation is to
 * combine these two, since then the user doesn't need to concern himself with
 * the order in which server and clients are started. <p>
 *
 * If the RemoteLogFilter is to register itself with the RMI registry, a few
 * additional things have to be decided: <br>
 * - Which registry port to use<br>
 *
 * - Whether to create a registry, if the rmiregistry tool hasn't been
 * started. This can be useful when a long-running server program is using its
 * own registry port anyway, but should be used with care, as the termination
 * of the program that created the registry will destroy all other
 * registrations. <br>
 *
 * - Whether LogServers from different machines are allowed to connect. This
 * can be very useful but is also a great security hole. The decision depends
 * on your kind of program and whether your machine is safe behind a
 * firewall. <p>
 *
 * To handle the log messages, the RemoteLogFilter must be added to the Log
 * filter chain with {@link Log#addFilter Log.addFilter}. It implements the
 * {@link LogUser LogUser} interface to enable configuration of which messages
 * to pass on through the filter chain.  Its default filter mode is {@link
 * LogUser#FILTER_NONE FILTER_NONE}. <p>
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.2 $
 *
 * @deprecated Since qflib version 0.98.0 this class is superceded by
 * RemoteLogWriter.
 */

// }}}
public class RemoteLogFilter
    extends LogClientImplBase
    implements LogUser
{
    // {{{ variables

    /**
     * RemoteLogFilter instance for convenience methods.
     */
    private static RemoteLogFilter instance;

    /**
     * The LogUser mode.
     */
    private int mode = FILTER_NONE;

    /**
     * The name of the RemoteLogFilter.
     */
    private String name;

    /**
     * The name under which the RemoteLogFilter is bound in the Registry.
     */
    private String bound;

    /**
     * Whether only local connects from servers are allowed.
     */
    private boolean localOnly;

    /**
     * The server to connect to.
     */
    private LogServer logServer;

    /**
     * The id received from the server.
     */
    private int id;

    /**
     * Whether the RemoteLogFilter is connected to a LogServer.
     */
    private boolean connected;

    /**
     * The listener to forward log entries to.
     */
    private LogListener listener;

    /**
     * The RemoteLogLevelListener registered with the Logger class.
     */
    private RemoteLogLevelListener levelListener;

    /**
     * The LogLevelListener registered with the Logger class.
     */
    private RemoteLogLevelAdapter adapter;

    /**
     * Number of communication failures.
     */
    private int failure = 0;

    /**
     * Number of failures after which to stop trying to call the listener.
     */
    private final static int MAX_FAILURES = 3;

    // }}}

    // {{{ constructor

    /**
     * Create a new RemoteLogFilter and connect to a LogServer.
     *
     * @param   name    The name to connect under.
     * @throws  RemoteException         If something RMI specific goes wrong.
     */
    public RemoteLogFilter (String name)
         throws RemoteException
    {
        this.name = name;
    }

    // }}}

    //----------------------------------------------------------------------
    // Static convenience methods
    //----------------------------------------------------------------------
    // {{{ logRemote(server)

    /**
     * Create a new RemoteLogFilter and connect it to a LogServer and put it
     * into the Log filter chain unless the connection to the LogServer fails.
     *
     * @param   server  The servername to connect to in the format used
     *                  by {@link java.rmi.Naming#lookup
     *                  java.rmi.Naming.lookup}.
     * @param   name    The name to connect under.
     *
     * @throws  MalformedURLException   If the server URL is bad.
     * @throws  NotBoundException       If the server is not bound under that
     *                                  name.
     * @throws  RemoteException         If something RMI specific goes wrong.
     *
     * @deprecated Please use {@link RemoteLogWriter#logRemote(String,String)
     * RemoteLogWriter.logRemote} instead.
     */
    public static void logRemote(String server, String name)
        throws MalformedURLException, NotBoundException, RemoteException
    {
        instance = new RemoteLogFilter (name);

        instance.connect(server);
        Log.addFilter(instance);
    }

    // }}}
    // {{{ logRemote(RMI)

    /**
     * Create a new RemoteLogFilter, register it with the RMI registry and put
     * it into the Log filter chain unless the registration fails.
     *
     * @param   name            The name to use when connecting to a LogServer.
     * @param   port            The registry port to use. The value -1 means
     *                          use the standard RMI registry port.
     * @param   createRegistry  Whether to create a new Registry if none is
     *                          available.
     * @param   localOnly       If true, only LogServers from the same network
     *                          address as the RemoteLogFilter are allowed to
     *                          connect.
     * @throws  RemoteException         If something RMI specific goes wrong.
     *
     * @deprecated Please use {@link
     * RemoteLogWriter#logRemote(String,int,boolean,boolean)
     * RemoteLogWriter.logRemote} instead.
     */
    public static void logRemote (String name, int port,
                                  boolean createRegistry, boolean localOnly)
        throws RemoteException
    {
        instance = new RemoteLogFilter (name);

        instance.register(port, createRegistry, localOnly);
        Log.addFilter(instance);
    }

    // }}}
    // {{{ logRemote(server and RMI)

    /**
     * Create a new RemoteLogFilter and connect it to a LogServer, register it
     * with the RMI registry and put it into the Log filter chain unless both
     * the connection to the LogServer and the registration fail.
     *
     * @param   server  The servername to connect to in the format used
     *                  by {@link java.rmi.Naming#lookup
     *                  java.rmi.Naming.lookup}.
     * @param   name    The name to connect under.
     * @param   port            The registry port to use. The value -1 means
     *                          use the standard RMI registry port.
     * @param   createRegistry  Whether to create a new Registry if none is
     *                          available.
     * @param   localOnly       If true, only LogServers from the same network
     *                          address as the RemoteLogFilter are allowed to
     *                          connect.
     * @throws  MalformedURLException   If the server URL is bad.
     * @throws  NotBoundException       If the server is not bound under that
     *                                  name.
     * @throws  RemoteException         If something RMI specific goes wrong.
     *
     * @deprecated Please use {@link
     * RemoteLogWriter#logRemote(String,String,int,boolean,boolean)
     * RemoteLogWriter.logRemote} instead.
     */
    public static void logRemote (String server, String name, int port,
                                  boolean createRegistry, boolean localOnly)
        throws MalformedURLException, NotBoundException, RemoteException
    {
        instance = new RemoteLogFilter (name);

        boolean register = false;
        RemoteException rex = null;
        try {
            try {
                instance.register(port, createRegistry, localOnly);
                register = true;
            } catch (RemoteException ex) {
                rex = ex;
            }
            instance.connect(server);
            register = true;
            if (rex != null) {
                throw rex;
            }
        } finally {
            Log.addFilter(instance);
        }
    }

    // }}}

    // {{{ stopLogging

    /**
     * Remove the RemoteLogFilter instance from the Log filter chain,
     * disconnect from the LogServer and remove the instance from the RMI
     * registry.
     *
     * @deprecated Please use {@link RemoteLogWriter#stopLogging
     * RemoteLogWriter.stopLogging} instead.
     */
    public static void stopLogging()
    {
        if (instance != null) {
            Log.removeFilter(instance);
            instance.disconnect();
        }
    }

    // }}}

    // {{{ getInstance

    /**
     * Get the RemoteLogFilter instance created last through one of the {@link
     * #logRemote logRemote} methods.
     *
     * @return  The previously created RemoteLogFilter.
     *
     * @deprecated No replacement.
     */
    public static RemoteLogFilter getInstance()
    {
        return instance;
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ connect

    /**
     * Connect to a LogServer.
     *
     * @param   server  The servername to connect to in the format used
     *                  by {@link java.rmi.Naming#lookup
     *                  java.rmi.Naming.lookup}.
     *
     * @throws  MalformedURLException   If the server URL is bad.
     * @throws  NotBoundException       If the server is not bound under that
     *                                  name.
     * @throws  RemoteException If something RMI specific goes wrong.
     */
    public void connect(String server)
        throws MalformedURLException, NotBoundException, RemoteException
    {
        logServer = (LogServer) Naming.lookup(server);
        id = logServer.getId(name);
        connect();
    }

    // }}}
    // {{{ register

    /**
     * Register the RemoteLogFilter with the RMI registry.
     *
     * @param   port            The registry port to  use, -1 for default.
     * @param   createRegistry  Whether to create the registry if none exists.
     * @param   localOnly       If true, only LogServers from the same network
     *                          address as the RemoteLogFilter are allowed to
     *                          connect.
     *
     * @throws  RemoteException If no working Registry can be found or created.
     */
    public void register(int port, boolean createRegistry, boolean localOnly)
        throws RemoteException
    {
        this.localOnly = localOnly;
        Registry reg = null;
        RemoteException rex = null;
        try {
            reg = LocateRegistry.getRegistry
                (port > 0 ? port : Registry.REGISTRY_PORT);
        } catch (RemoteException ex) {
            rex = ex;
        }

        // Even if we got a registry Object it doesn't mean that the registry
        // is up and running

        bound = REGISTRY_BASE_NAME + name + "-" +
            (int)(Math.random() * 100000);
        if (reg != null) {
            try {
                reg.bind(bound, this);
                return;
            } catch (AlreadyBoundException ex) {
                // at least the registry is ok
            } catch (RemoteException ex) {
                // this registry is bad
                reg = null;
                rex = ex;
            }
        }

        if (reg == null) {
            if (createRegistry) {
                // if this fails just pass the Exception on
                reg = LocateRegistry.createRegistry
                    (port > 0 ? port : Registry.REGISTRY_PORT);
            } else {
                if (rex != null) {
                    throw rex;
                }
                // it should be impossible to get here
                System.err.println("RemoteLogFilter.register: internal error");
            }
        }

        boolean done = false;
        while (! done) {
            // try to generate a unique name
            bound = REGISTRY_BASE_NAME + name + "-" +
                (int)(Math.random() * 100000);
            try {
                reg.bind(bound, this);
                done = true;
            } catch (AlreadyBoundException ex) {
                // name already bound, try again
            }
            // RemoteException passes through, we have lost
        }
    }

    // }}}
    // {{{ disconnect

    /**
     * Disconnect from the LogServer.
     */
    public void disconnect ()
    {
        if (connected) {
            try {
                logServer.disconnect(id);
            } catch (RemoteException ex) {
            }
        }
        if (bound != null) {
            try {
                Naming.unbind(bound);
            } catch (RemoteException ex) {
                // ignore, maybe RMI registry is down
            } catch (NotBoundException ex) {
                // should NOT happen
                //  ex.printStackTrace();
            } catch (MalformedURLException ex) {
                // should NOT happen
                //  ex.printStackTrace();
            }
        }
        listener = null;
        if (adapter != null) {
            Logger.removeLogLevelListener(adapter);
            //  XXX adapter.disconnect();
            adapter = null;
            levelListener = null;
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // The LogUser interface
    //----------------------------------------------------------------------
    // {{{ filter

    /**
     * Take a LogEntry and pass it to the LogServer.
     *
     * @param   logEntry        The entry to pass on.
     *
     * @return  True if the entry should be passed on to the next filter.
     */
    public boolean filter (LogEntry logEntry)
    {
        boolean used = false;
        if (failure >= 0) {
            if (failure >= MAX_FAILURES) {
                failure = -1;
            } else if (listener != null) {
                try {
                    listener.log(logEntry);
                    used = true;
                } catch (RemoteException ex) {
                    failure++;
                }
            }
        }
        return mode != FILTER_ALL
            && ((used && mode != FILTER_USED)
                || (!used && mode != FILTER_UNUSED));
    }

    // }}}
    // {{{ filter

    /**
     * Send all messages from the list to the remote listener.
     *
     * @param   entries The list of LogEntries.
     */
    public LogEntry[] filter (LogEntry[] entries)
    {
        boolean used = false;
        if (failure >= 0) {
            if (failure >= MAX_FAILURES) {
                failure = -1;
            } else if (listener != null) {
                try {
                    listener.log(entries);
                    used = true;
                } catch (RemoteException ex) {
                    failure++;
                }
            }
        }
        if (mode == FILTER_ALL
            || (used && mode == FILTER_USED)
            || (!used && mode == FILTER_UNUSED)) {
            return new LogEntry[0];
        } else {
            return entries;
        }
    }

    // }}}
    // {{{ setFilterMode

    /**
     * Configure the RemoteLogFilter's filter mode.
     *
     * @param   mode    The mode to set.
     */
    public void setFilterMode(int mode)
    {
        this.mode = mode;
    }

    // }}}

    //----------------------------------------------------------------------
    // The LogClient interface
    //----------------------------------------------------------------------
    // {{{ connect

    /**
     * Tell the LogClient that it should disengage from the registry and
     * connect itself to a LogServer.
     *
     * @param   logServer  The LogServer to connect to.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     */
    public void connect(LogServer logServer)
        throws RemoteException
    {
        if (localOnly) {
            // check if the LogServer is local
            try {
                String host = getClientHost();
                InetAddress address = InetAddress.getByName(host);
                if (! (address.getHostAddress().startsWith("127.0.0.")
                       || address.equals(InetAddress.getLocalHost()))) {
                    throw new AccessException
                        ("RemoteLogFilter.connect: " +
                         host + " is not a local address");
                }
            } catch (UnknownHostException ex) {
                //  ex.printStackTrace();
            } catch (ServerNotActiveException ex) {
                //  ex.printStackTrace();
            }
        }

        // handle multiple connection by disconnecting from the old server
        if (connected) {
            try {
                this.logServer.disconnect(id);
            } catch (RemoteException ex) {
            }
        }
        if (adapter != null) {
            Logger.removeLogLevelListener(adapter);
            //  XXX adapter.disconnect();
            adapter = null;
            levelListener = null;
        }

        // now connect to the new server
        try {
            this.logServer = logServer;
            id = logServer.getId(name);

            connect();
        } catch (Exception ex) {
            // can't do much about it
            //  ex.printStackTrace();
        }
    }


    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ connect

    /**
     * Helper method that handles the connection business.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     */
    private void connect()
        throws RemoteException
    {
        listener = logServer.getLogListener(id);

        levelListener = logServer.getLogLevelListener(id);
        adapter = new RemoteLogLevelAdapter ();
        adapter.setRemoteListener(levelListener);
        Logger.addLogLevelListener(adapter);
        failure = 0;
        connected = true;
    }



    // }}}
}
