// {{{ copyright

/********************************************************************
 *
 * $Id: RemoteLogLevelCallback.java,v 1.2 2000/10/26 15:04:08 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.logrmi;

// {{{ imports

import java.rmi.Remote;
import java.rmi.RemoteException;

// }}}

/**
 * Callback that enables remote access to the log levels of an application as
 * well as the ouput and pre queue levels and the queue and flush buffer
 * parameters.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.2 $
 * @see         de.qfs.lib.log.Log
 */
public interface RemoteLogLevelCallback
    extends Remote
{
    /**
     * Get the current log levels from the callback. This method is needed to
     * synchronize the LogLevelListener and the LogLevelCallback after the
     * connection has been established.
     *
     * @return  An object array that contains an alternating sequence of
     *          class/package names and log levels. The levels are Integers
     *          that may be null for classes for which a level has been
     *          requested but not explicitly set.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     */
    public Object[] getLogLevels()
        throws RemoteException;

    /**
     * Callback method for a LogLevelListener to change the log level for
     * a class or package.
     *
     * @param   name    The name of the affected class or package.
     * @param   level   The new log level.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     */
    public void setLogLevel(String name, int level)
        throws RemoteException;

    /**
     * Callback method for a LogLevelListener to remove the log level for
     * a class or package.
     *
     * @param   name    The name of the affected class or package.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     */
    public void removeLogLevel(String name)
        throws RemoteException;

    /**
     * Get the current output log level from the callback's application.
     *
     * @return  The current output log level.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     *
     * @see     de.qfs.lib.log.Log#getOutputLevel
     */
    public int getOutputLevel()
        throws RemoteException;

    /**
     * Set the output log level for the callback's application.
     *
     * @param level     Output log level to set.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     *
     * @see     de.qfs.lib.log.Log#setOutputLevel
     */
    public void setOutputLevel(int level)
        throws RemoteException;

    /**
     * Return the current pre-queue level from the callback's application.
     *
     * @return  The current pre-queue level.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     *
     * @see     de.qfs.lib.log.Log#getPreQueueLevel
     */
    public int getPreQueueLevel()
        throws RemoteException;

    /**
     * Set the pre-queue level for the callback's application.
     *
     * @param level     The pre-queue level to set.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     *
     * @see     de.qfs.lib.log.Log#setPreQueueLevel
     */
    public void setPreQueueLevel(int level)
        throws RemoteException;

    /**
     * Check whether the callback's application is using a log queue.
     *
     * @return  Whether the callback's application is queueing log entries.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     *
     * @see     de.qfs.lib.log.Log#isQueueing
     */
    public boolean isQueueing()
        throws RemoteException;

    /**
     * Set whether the callback's application should use a log queue.
     *
     * @param   queue   Whether to queue log entries.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     *
     * @see     de.qfs.lib.log.Log#setQueueing
     */
    public void setQueueing(boolean queue)
        throws RemoteException;

    /**
     * Get the size of log queue of the callback's application.
     *
     * @return  The size of the log queue.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     *
     * @see     de.qfs.lib.log.Log#getQueueSize
     */
    public int getQueueSize()
        throws RemoteException;

    /**
     * Set the size of log queue of the callback's application.
     *
     * @param   size    The size of the log queue.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     *
     * @see     de.qfs.lib.log.Log#setQueueSize
     */
    public void setQueueSize(int size)
        throws RemoteException;

    /**
     * Check whether the callback's application is dropping entries when
     * its log queue overflows.
     *
     * @return  Whether entries are dropped on overflow.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     *
     * @see     de.qfs.lib.log.Log#isDropOnOverflow
     */
    public boolean isDropOnOverflow()
        throws RemoteException;

    /**
     * Set whether the callback's application should drop entries when
     * its log queue overflows.
     *
     * @param   drop    Whether to drop entries on overflow.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     *
     * @see     de.qfs.lib.log.Log#setDropOnOverflow
     */
    public void setDropOnOverflow(boolean drop)
        throws RemoteException;

    /**
     * Get the size of the flush buffer of the callback's application.
     *
     * @return  The flush buffer size or 0 to indicate no flush buffer.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     *
     * @see     de.qfs.lib.log.Log#getFlushBufferSize
     */
    public int getFlushBufferSize()
        throws RemoteException;

    /**
     * Set the size of the flush buffer of the callback's application. Setting
     * it to 0 will turn the flush buffer off.
     *
     * @param   size    The size of the flush buffer.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     *
     * @see     de.qfs.lib.log.Log#setFlushBufferSize
     */
    public void setFlushBufferSize(int size)
        throws RemoteException;

    /**
     * Get the level that will trigger a flush of the messages saved in the
     * flush buffer of the callback's application.
     *
     * @return  The flush buffer's trigger level.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     *
     * @see     de.qfs.lib.log.Log#getFlushTriggerLevel
     */
    public int getFlushTriggerLevel()
        throws RemoteException;

    /**
     * Set the level that will trigger a flush of the messages saved in the
     * flush buffer of the callback's application.
     *
     * @param   level   The trigger level to set.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     *
     * @see     de.qfs.lib.log.Log#setFlushTriggerLevel
     */
    public void setFlushTriggerLevel(int level)
        throws RemoteException;

    /**
     * Get the number of messages to pass unfiltered through the pre-queue
     * stage after a flush happened in the callback's application.
     *
     * @return  The number of messages to pass.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     *
     * @see     de.qfs.lib.log.Log#getPostFlushSize
     */
    public int getPostFlushSize()
        throws RemoteException;

    /**
     * Set the number of messages to pass unfiltered through the pre-queue
     * stage after a flush happened in the callback's application.
     *
     * @param   size    The number of messages to pass.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     *
     * @see     de.qfs.lib.log.Log#setPostFlushSize
     */
    public void setPostFlushSize(int size)
        throws RemoteException;

    /**
     * End the connection with the LogLevelCallback.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     */
    public void disconnect()
         throws RemoteException;

}
