// {{{ copyright

/********************************************************************
 *
 * $Id: BooleanOption.java,v 1.8 2000/06/06 16:41:27 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * A BooleanOption's value is of class Boolean. The value must not be null,
 * since the swing JCheckBox used to edit a BooleanOption does not support
 * the concept of neither true nor false.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.8 $
 */
public class BooleanOption
extends Option implements ConfigurableOption
{
    // {{{ variables

    /**
     * The UID used for serialization.
     */
    static final long serialVersionUID = -591653005109324305L;

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (BooleanOption.class);

    // }}}

    // {{{ constructors

    /**
     * Create a new BooleanOption.
     */
    public BooleanOption ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "BooleanOption()", "");
        }
    }

    /**
     * Create a new BooleanOption.
     *
     * @param   name    The name of the BooleanOption.
     * @param   value   The initial value.
     */
    public BooleanOption(String name, boolean value)
    {
        super(name, new Boolean(value));
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "BooleanOption(String,boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "value: " + value);
        }
    }

    // }}}

    // {{{ setNullOK

    /**
     * Override setNullOK, since a null value is illegal for a BooleanOption.
     *
     * @param   nullOK  Ignored.
     */
    public void setNullOK(boolean nullOK)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setNullOK(boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "nullOK: " + nullOK);
        }
        if (nullOK) {
            if (logger.level >= Log.WRN) {
                logger.log(Log.WRN, "setNullOK(boolean)",
                           "Warning: trying to allow null for BooleanOption");
            }
        }
    }

    // }}}
    // {{{ equals

    /**
     * Two BooleanOptions are considered equal if their names and values
     * are equal.
     *
     * @param   object  The object to compare to.
     *
     * @return  True if objects are equal
     */
    public boolean equals(Object object)
    {
        return (object instanceof BooleanOption) && super.equals(object);
    }

    // }}}

    //----------------------------------------------------------------------
    // the ConfigurableOption interface
    //----------------------------------------------------------------------
    // {{{ getConfigurableValue

    /**
     * Get the Option's value as a String.
     *
     * @return  The Option's value as a String.
     */
    public String getConfigurableValue()
    {
        Object value = getValue();
        return value == null ? "" : value.toString();
    }

    // }}}
    // {{{ setConfigurableValue

    /**
     * Set the Option's value from a String.
     *
     * @param   val     The value to set.
     */
    public void setConfigurableValue(String val)
    {
        setValue(val.equals("") ? null : new Boolean (val));
    }

    // }}}
}
