// {{{ copyright

/********************************************************************
 *
 * $Id: FileOption.java,v 1.7 2000/06/06 16:41:27 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.io.File;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * A FileOption's value is of class File. Additional flags determine, whether
 * the file must exist and if it must be a file, a directory or can be either.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.7 $
 */
public class FileOption
    extends Option
    implements ConfigurableOption
{
    // {{{ variables

    /**
     * The UID used for serialization.
     */
    static final long serialVersionUID = -8030424147783998707L;

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (FileOption.class);
    /**
     * The file may be either a plain file or a directory.
     */
    public final static int ANY_TYPE = 0;

    /**
     * The file must be a plain file.
     */
    public final static int FILE_TYPE = 1;

    /**
     * The file must be a directory.
     */
    public final static int DIRECTORY_TYPE = 2;

    /**
     * The mustExist state of the FileOption.
     * @serial
     */
    private boolean mustExist;

    /**
     * The fileType of the FileOption.
     * @serial
     */
    private int fileType;

    // }}}

    // {{{ constructors

    /**
     * Create a new FileOption.
     */
    public FileOption ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "FileOption()", "");
        }
    }

    /**
     * Create a new FileOption.
     *
     * @param   name    The name of the FileOption.
     * @param   value   The initial value.
     */
    public FileOption (String name, File value)
    {
        super (name, value);
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "FileOption(String,File)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "value: " + value);
        }
    }

    /**
     * Create a new FileOption.
     *
     * @param   name            The name of the FileOption.
     * @param   value           The initial value.
     * @param   nullOK          Whether the value may be null.
     * @param   mustExist       Whether the file must exist.
     * @param   fileType        The type of the file.
     */
    public FileOption (String name, File value, boolean nullOK,
                       boolean mustExist, int fileType)
    {
        super(name, value, nullOK);
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "FileOption(String,File,boolean,boolean,int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "value: " + value + ", " +
                       "nullOK: " + nullOK + ", " +
                       "mustExist: " + mustExist + ", " +
                       "fileType: " + fileType);
        }
        this.mustExist = mustExist;
        this.fileType = fileType;
    }

    // }}}

    // {{{ the mustExist flag

    /**
     * Get the mustExist state of the FileOption.
     *
     * @return  The mustExist state of the FileOption.
     */
    public final boolean isMustExist()
    {
        return mustExist;
    }

    /**
     * Set the mustExist state of the FileOption.
     *
     * @param   mustExist       The mustExist state to set.
     */
    public final void setMustExist(boolean mustExist)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setMustExist(boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "mustExist: " + mustExist);
        }
        this.mustExist = mustExist;
    }

    // }}}
    // {{{ the fileType

    /**
     * Get the fileType of the FileOption.
     *
     * @return  The fileType of the FileOption.
     */
    public final int getFileType()
    {
        return fileType;
    }

    /**
     * Set the fileType of the FileOption.
     *
     * @param   fileType        The fileType to set.
     */
    public final void setFileType(int fileType)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setFileType(int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "fileType: " + fileType);
        }
        this.fileType = fileType;
    }

    // }}}

    // {{{ equals

    /**
     * Two FileOptions are considered equal if their names and values
     * are equal.
     *
     * @param   object  The object to compare to.
     *
     * @return  True if objects are equal
     */
    public boolean equals(Object object)
    {
        return (object instanceof FileOption) && super.equals(object);
    }

    // }}}

    // {{{ the ConfigurableOption interface

    /**
     * Get the Option's value as a String.
     *
     * @return  The Option's value as a String.
     */
    public String getConfigurableValue()
    {
        File f = (File) getValue();
        return f != null ? f.getPath() : "";
    }

    /**
     * Set the Option's value from a String.
     *
     * @param   val     The value to set.
     */
    public void setConfigurableValue(String val)
    {
        setValue(val.equals("") ? null : new File (val));
    }

    // }}}
}
