// {{{ copyright

/********************************************************************
 *
 * $Id: IntegerOptionEdit.java,v 1.11 2001/01/31 13:42:41 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.BorderLayout;

import javax.swing.JTextField;

import de.qfs.lib.gui.Message;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

// {{{ doc

/**
 * An OptionEdit for an {@link IntegerOption IntegerOption}. It uses the
 * following resources: <p>
 *
 * <TABLE BORDER="1" CELLPADDING="3" CELLSPACING="0" WIDTH="100%">
 * <TR>
 * <TH WIDTH="40%">Resource name</TH>
 * <TH>Usage</TH>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.label</code></TD>
 * <TD>The label for the Option (optional, default empty).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.mnemonic</code></TD>
 * <TD>Keyboard accelerator (optional, default none).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.disabled</code></TD>
 * <TD>If true, the edit will be disabled and remain so even if {@link
 * #setEnabled setEnabled(true)} is called (optional, default false).</TD>
 * </TR>
 * </TABLE>
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.11 $
 */

// }}}
public class IntegerOptionEdit
    extends SimpleOptionEdit
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (IntegerOptionEdit.class);

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ IntegerOptionEdit

    /**
     * Create a new IntegerOptionEdit. Don't call this directly, use {@link
     * OptionEdit#makeEdit OptionEdit.makeEdit} instead.
     */
    public IntegerOptionEdit ()
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ init

    /**
     * Initialize the OptionEdit.
     */
    public void init()
    {
        super.init();
        text = new JTextField () {
            public boolean isFocusTraversable()
            {
                return isEditable() && this.isEnabled();
            }
        };
        label.setLabelFor(text);
        text.setName("Text-" + option.getName());
        add(text, BorderLayout.CENTER);
        if (option.getValue() != null) {
            text.setText(option.getValue().toString());
            text.setCaretPosition(0);
        }
        if (disabled) {
            setEnabled(false);
        }
    }

    // }}}
    // {{{ checkValue

    /**
     * Check the currently entered value of the OptionEdit. If the value is
     * not valid, request the focus.
     *
     * @return  True if the value is valid, false otherwise.
     */
    public boolean checkValue()
    {
        String val = text.getText();
        Integer iVal = null;
        if (!val.equals("")) {
            try {
                iVal = new Integer (val);
            } catch (NumberFormatException ex) {
                Message.showMessage(this, "option.error.notANumber");
                text.requestFocus();
                return false;
            }
        }
        if (! option.checkValue(this, iVal)) {
            text.requestFocus();
            return false;
        }
        return true;
    }

    // }}}
    // {{{ readValue

    /**
     * Read the currently entered value and store it in the Option.
     */
    public void readValue()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "readValue()", "");
        }
        String val = text.getText();
        Integer iVal = null;
        if (!val.equals("")) {
            try {
                iVal = new Integer (val);
            } catch (NumberFormatException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log("readValue()", ex,
                               "Should have called checkValue");
                }
            }
        }
        option.setValue(iVal);
    }

    // }}}
}
