// {{{ copyright

/********************************************************************
 *
 * $Id: OptionDialog.jpp,v 1.4 2001/01/24 15:55:58 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.Component;
import java.awt.Dialog;
import java.awt.Frame;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JButton;
import javax.swing.JLabel;

import de.qfs.lib.gui.Grid;
import de.qfs.lib.gui.ModalDialog;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.MapResourceBundle;

// }}}

/**
 * This class offers a convenient way to bring up a dialog in which an
 * application's options can be edited. <p>
 *
 * Since version 0.98.0 this class is based on {@link
 * de.qfs.lib.gui.ModalDialog ModalDialog}, so the buttons of the dialog can
 * be customized easily. Please see the documentation for ModalDialog about
 * the resources used. <p>
 *
 * OptionDialog defines one additional resource for each button:
 * <code>dialog.button.<em>name</em>.set</code>. If true, pressing the button
 * causes the values the user entered to be read from the dialog and stored in
 * the options. If this fails due to illegal values, the action is aborted.
 * The default for the <code>set</code> property is false. <p>
 *
 * If no <code>name</code> resource is defined for the dialog, its AWT name is
 * set to <code>"OptionDialog-&lt;root&gt;-&lt;options&gt;"</code> where
 * <code>&lt;root&gt;</code> is the resource name of the root OptionGroup and
 * <code>&lt;options&gt;</code> is the configurable name of the OptionSet
 * being edited.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.4 $
 */
public class OptionDialog
    extends ModalDialog
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (OptionDialog.class);

    /**
     * The Options to edit.
     */
    protected OptionSet options;

    /**
     * The name of the root OptionGroup.
     */
    protected String root;

    /**
     * The OptionGroup displayed in the dialog.
     */
    protected OptionGroup optionGroup;

    /**
     * The parent Window for the Dialog.
     */
    protected Window parent;

    // }}}

    //----------------------------------------------------------------------
    // Constructors
    //----------------------------------------------------------------------
    // {{{ OptionDialog(Frame,String,OptionSet,String)

    /**
     * Create a new OptionDialog.
     *
     * @param   parent  The parent Frame.
     * @param   title   The title to display.
     * @param   options The Options to edit.
     * @param   root    The resource name of the root of the OptionGroups. If
     *                  null, the configurable name of the OptionSet will be
     *                  used.
     *
     * @deprecated      Replaced with {@link
     *                  #OptionDialog(String,Frame,OptionSet,String)
     *                  OptionDialog} as of version 0.98.0.
     */
    public OptionDialog (Frame parent, String title, OptionSet options,
                         String root)
    {
        super("optiondialog", parent);
        this.parent = parent;
        this.options = options;
        this.root = root;
        init(Option.getResources());
        setName("OptionDialog-" + root + "-" + options.getConfigurableName());
        setTitle(title);
    }

    // }}}
    // {{{ OptionDialog(String,Frame,OptionSet,String)

    /**
     * Create a new OptionDialog.
     *
     * @param   name    The resource name of the dialog.
     * @param   parent  The parent Frame.
     * @param   options The Options to edit.
     * @param   root    The resource name of the root of the OptionGroups. If
     *                  null, the configurable name of the OptionSet will be
     *                  used.
     *
     * @since   0.98.0
     */
    public OptionDialog (String name, Frame parent, OptionSet options,
                         String root)
    {
        super(name, parent);
        this.parent = parent;
        this.options = options;
        this.root = root;
        init(Option.getResources());
        if (Option.getResources().getString(resName("name"), null) == null) {
            setName("OptionDialog-" + root + "-" +
                    options.getConfigurableName());
        }
    }

    // }}}

    // {{{ OptionDialog(String,Dialog,OptionSet,String)

    /**
     * Create a new OptionDialog.
     *
     * @param   name    The resource name of the dialog.
     * @param   parent  The parent Dialog.
     * @param   options The Options to edit.
     * @param   root    The resource name of the root of the OptionGroups. If
     *                  null, the configurable name of the OptionSet will be
     *                  used.
     *
     * @since   0.98.0
     */
    public OptionDialog (String name, Dialog parent, OptionSet options,
                         String root)
    {
        super(name, parent);
        this.parent = parent;
        this.options = options;
        this.root = root;
        init(Option.getResources());
        if (Option.getResources().getString(resName("name"), null) == null) {
            setName("OptionDialog-" + root + "-" +
                    options.getConfigurableName());
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ readValues

    /**
     * Read the current values from the dialog, store them in the options and
     * notify the observers.
     *
     * @return  True if everything was OK, false if there are illegal values.
     *          In the latter case no options are changed.
     *
     * @since   0.98.0
     */
    public boolean readValues()
    {
        if (optionGroup != null) {
            if (! optionGroup.checkValues()) {
                return false;
            }
            optionGroup.readValues();
            optionGroup.notifyOptionObservers();
            options.notifyObservers();
        }
        return true;
    }

    // }}}
    // {{{ update

    /**
     * Update the display of the OptionGroups.
     *
     * @param   options The Options to use. Null means use original Options.
     */
    public void update(OptionSet options)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "update()", "");
        }
        optionGroup.update(options == null ? this.options : options);
    }

    // }}}
    // {{{ doModal

    /**
     * Bring up the Dialog and wait for the user to finish editing. If the
     * user confirms the actions with OK, all input will be checked for
     * validity and the Options will be set accordingly. All Observers
     * will be notified. <p>
     *
     * To get the full benefit of the configurable buttons, use {@link
     * #doModal(Component,boolean) doModal(Component,boolean)} instead.
     *
     * @return True, if completed with the OK Button, false if canceled.
     */
    public boolean doModal()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "doModal()", "");
        }
        doModal(parent, true);
        return retval == 0;
    }

    // }}}
    // {{{ setParent

    /**
     * Update the parent of the dialog, so the same Dialog can be reused with
     * different parent Frames.
     *
     * @param   parent  The parent Frame to set.
     *
     * @deprecated      The window's owner cannot really be changed. To place
     *                  the dialog relative to some component use {@link
     *                  #doModal(Component,boolean)
     *                  doModal(Component,boolean)}.
     */
    public void setParent(Frame parent)
    {
        this.parent = parent;
    }

    // }}}
    // {{{ getGroup

    /**
     * Get the OptionGroup of the dialog.
     *
     * @return  The dialog's OptionGroup.
     * @since   0.98.0
     */
    public OptionGroup getGroup()
    {
        return optionGroup;
    }

    // }}}
    // {{{ getResult

    /**
     * Get the result for the last {@link #doModal doModal} operation. This is
     * necessary to determine the button the user selected.
     *
     * @return -1 if the dialog was canceled, 0 for the first custom or the OK
     * button, 1 for the second custom or the Cancel button, 2 and above for
     * further custom buttons.
     */
    public final int getResult()
    {
        return retval;
    }

    // }}}

    //----------------------------------------------------------------------
    // Initialization
    //----------------------------------------------------------------------
    // {{{ initComponent

    /**
     * Initialize the OptionGroup.
     */
    protected void initComponent()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "initComponent()", "");
        }

        optionGroup = OptionGroup.makeGroup(root, options);
        mainComponent = null;
        if (optionGroup != null) {
            mainComponent = optionGroup.getComponent();
            if (mainComponent instanceof Grid
                && ((Grid) mainComponent).needGlue()) {
                ((Grid) mainComponent).addGlue();
            }
        }
        if (mainComponent == null) {
            mainComponent = new JLabel ("MISSING OptionGroup: " + root);
        }
    }

    // }}}
    // {{{ initButton

    /**
     * Initialize one button. Overridden to create an OptionButtonData instead
     * of a ButtonData.
     *
     * @param   bname   The resource name of the button.
     * @param   rb      The resources to use.
     *
     * @return  The initialized button.
     */
    protected ButtonData initButton(String bname, MapResourceBundle rb)
    {
        return new OptionButtonData (bname, rb);
    }

    // }}}

    //----------------------------------------------------------------------
    // The ActionListener interface
    //----------------------------------------------------------------------
    // {{{ actionPerformed

    /**
     * Handler for a button press. If the button's close property is set, its
     * return value is determined and the dialog is closed.
     *
     * @param   e       The event containing the details.
     */
    public void actionPerformed(ActionEvent e)
    {
        if (e.getSource() instanceof JButton) {
            int idx = findButton ((JButton) e.getSource());
            if (idx >= 0) {
                if (buttons[idx] instanceof OptionButtonData
                    && ((OptionButtonData) buttons[idx]).set) {
                    if (! readValues()) {
                        return;
                    }
                }
                if (buttons[idx].close) {
                    if (buttons[idx].hasRetval) {
                        retval = buttons[idx].retval;
                    } else {
                        retval = idx;
                    }
                    setVisible(false);
                }
            }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Extended button data
    //----------------------------------------------------------------------
    // {{{ class OptionButtonData

    /**
     * Extend the structure that encapsulates the values for a button.
     */
    protected static class OptionButtonData
        extends ModalDialog.ButtonData
    {
        // {{{ variables

        /**
         * Whether to try to set the options' values.
         */
        public boolean set;

        // }}}

        //------------------------------------------------------------------
        // Constructors
        //------------------------------------------------------------------
        // {{{ OptionButtonData()

        /**
         * Create a new ButtonData.
         */
        public OptionButtonData ()
        {
        }

        // }}}
        // {{{ OptionButtonData(String,MapResourceBundle)

        /**
         * Create a new ButtonData, reading its value from the resources.
         *
         * @param       name    The resource name of the button.
         * @param       rb      The resources to use.
         */
        public OptionButtonData (String name, MapResourceBundle rb)
        {
            super (name, rb);
            set = rb.getBoolean(ModalDialog.buttonResName
                                (name, "set"), false);
        }

        // }}}
    }

    // }}}
}
