// {{{ copyright

/********************************************************************
 *
 * $Id: OptionSet.jpp,v 1.19 2001/02/04 12:19:53 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;

import java.util.Enumeration;
import java.util.Observable;
import java.util.Observer;
import java.util.Properties;


import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;


import de.qfs.lib.config.Configurable;
import de.qfs.lib.config.Configuration;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * An OptionSet bundles Options together, offering convenience methods
 * for easy, typed access to their values. Default values must be provided
 * in many cases, to avoid exception handling. <p>
 *
 * New in qflib 0.97.0: <br>
 *
 * Options from nested OptionSets can be queried and set by building a kind of
 * Option path name with a slash '/' as separator. <br>
 *
 * Example: <br>
 *
 * If OptionSet options has a nested OptionSet in an Option named 'a', which
 * in turn has a StringOption named 'b', then <br>
 *
 * <code>options.getString("a/b", null)</code><br>
 *
 * will get the string value from the nested Option 'b'. If options contains
 * an Option named a/b' however, it will be accessed instead.
 *
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.19 $
 */
public class OptionSet extends Observable
implements Serializable, Cloneable, Configurable, Observer
{
    // {{{ variables

    /**
     * The UID used for serialization.
     */
    static final long serialVersionUID = -2172456472965747088L;
    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (OptionSet.class);

    /**
     * The Options of the OptionSet.
     */
    protected transient Map options =
            Collections.synchronizedMap(new HashMap ());

    /**
     * An array of the Options, used only for serialization.
     * @serial
     */
    private Option[] savedOptions;

    /**
     * The name to use when registering with the Configuration.
     * @serial
     */
    private String configurableName;

    // }}}

    // {{{ constructor

    /**
     * Create a new OptionSet.
     *
     * @param   name    The name to use as a Configurable.
     */
    public OptionSet (String name)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "OptionSet(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }
        configurableName = name;
    }

    // }}}

    //----------------------------------------------------------------------
    // Option management
    //----------------------------------------------------------------------
    // {{{ add

    /**
     * Add an option to the OptionSet.
     *
     * @param   option  The Option to add.
     */
    public void add(Option option)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "add(Option)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "option: " + option);
        }
        options.put(option.getName(), option);
        option.addObserver(this);
        setChanged();
    }

    // }}}
    // {{{ addAll

    /**
     * Add all options from another OptionSet to the OptionSet.
     *
     * @param   set     The source of the Options to add.
     */
    public void addAll(OptionSet set)
    {
        Option[] opts = set.getOptions();
        for (int i = 0; i < opts.length; i++) {
            add(opts[i]);
        }
    }

    // }}}
    // {{{ remove

    /**
     * Remove an Option from the OptionSet.
     *
     * @param   name    The name of the Option to remove.
     */
    public void remove(String name)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "remove(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }
        Option option = (Option) options.get(name);
        if (option != null) {
            option.deleteObserver(this);
        }
        options.remove(name);
        setChanged();
    }

    // }}}
    // {{{ getOption

    /**
     * Get an Option from the OptionSet.
     *
     * @param   name    The name of the Option.
     *
     * @return  The requested Option or null, if the name is unknown.
     */
    public Option getOption(String name)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getOption(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }
        Option option = (Option) options.get(name);
        if (option == null) {
            int idx = name.indexOf('/');
            if (idx > 0 && idx < name.length() - 1) {
                Option nested = (Option) options.get(name.substring(0, idx));
                if (nested != null
                    && (nested.getValue() instanceof OptionSet)) {
                    return ((OptionSet) nested.getValue())
                        .getOption(name.substring(idx + 1));
                }
            }
        }
        return option;
    }

    // }}}
    // {{{ getOptions

    /**
     * Get all Options from the OptionSet.
     *
     * @return  The Options.
     */
    public Option[] getOptions()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getOptions()", "");
        }
        return (Option[]) options.values().toArray(new Option [0]);
    }

    // }}}

    //----------------------------------------------------------------------
    // Values
    //----------------------------------------------------------------------
    // {{{ getValue

    /**
     * Get the value from an Option in the OptionSet.
     *
     * @param   name    The name of the option.
     * @param   def     A default value in case the name is unknown.
     *
     * @return  The value.
     */
    public Object getValue(String name, Object def)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getValue(String,Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "def: " + def);
        }
        Option option = getOption(name);
        if (option != null) {
            return option.getValue();
        }
        return def;
    }

    // }}}
    // {{{ getBoolean

    /**
     * Get the boolean value from an Option in the OptionSet.
     *
     * @param   name    The name of the option.
     * @param   def     A default value in case the name is unknown.
     *
     * @return  The value.
     */
    public boolean getBoolean(String name, boolean def)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getBoolean(String,boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "def: " + def);
        }
        Option option = getOption(name);
        if (option != null && (option.getValue() instanceof Boolean)) {
            return ((Boolean) option.getValue()).booleanValue();
        }
        return def;
    }

    // }}}
    // {{{ getInt

    /**
     * Get the int value from an Option in the OptionSet.
     *
     * @param   name    The name of the option.
     * @param   def     A default value in case the name is unknown.
     *
     * @return  The value.
     */
    public int getInt(String name, int def)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getInt(String,int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "def: " + def);
        }
        Option option = getOption(name);
        if (option != null &&
            (option.getValue() instanceof Integer)) {
            return ((Integer) option.getValue()).intValue();
        }
        return def;
    }

    // }}}
    // {{{ getString

    /**
     * Get the String value from an Option in the OptionSet.
     *
     * @param   name    The name of the option.
     * @param   def     A default value in case the name is unknown.
     *
     * @return  The value.
     */
    public String getString(String name, String def)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getString(String,String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "def: " + def);
        }
        Option option = getOption(name);
        if (option != null &&
            (option.getValue() == null
             || (option.getValue() instanceof String))) {
            return (String) option.getValue();
        }
        return def;
    }

    // }}}
    // {{{ getFile

    /**
     * Get the File value from an Option in the OptionSet.
     *
     * @param   name    The name of the option.
     * @param   def     A default value in case the name is unknown.
     *
     * @return  The value.
     */
    public File getFile(String name, File def)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getFile(String,File)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "def: " + def);
        }
        Option option = getOption(name);
        if (option != null &&
            (option.getValue() == null
            || (option.getValue() instanceof File))) {
            return (File) option.getValue();
        }
        return def;
    }

    // }}}

    //----------------------------------------------------------------------
    // The Configurable interface
    //----------------------------------------------------------------------
    // {{{ getConfigurableName

    /**
     * Get the Configurable's name.
     *
     * @return The Configurable's name.
     */
    public final String getConfigurableName ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getConfigurableName()", "");
        }

        return configurableName;
    }

    // }}}
    // {{{ setConfigurableName

    /**
     * Set the Configurable's name.
     *
     * @param   name    The name to set.
     */
    public final void setConfigurableName (String configurableName)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setConfigurableName(String)",
                        logger.level < Log.MTDDETAIL ? "" :
                        "configurableName: " + configurableName);
        }

        this.configurableName = configurableName;
    }

    // }}}
    // {{{ getConfigurableState

    /**
     * Get the Configurable's state.
     *
     * @return The Configurable's state.
     */
    public Properties getConfigurableState()
    {
        Properties state = new Properties ();
        Option[] opts = getOptions();
        for (int i = 0; i < opts.length; i++) {
            if (opts[i] instanceof ConfigurableOption) {
                state.put(opts[i].getName(),
                          ((ConfigurableOption) opts[i])
                          .getConfigurableValue());
            }
        }

        return state;
    }

    // }}}
    // {{{ setConfigurableState

    /**
     * Set the Configurable's state.
     *
     * @param   state   The state to set.
     */
    public void setConfigurableState(Properties state)
    {
        for (Enumeration e = state.keys(); e.hasMoreElements();) {
            String key = (String) e.nextElement();
            Option opt = getOption(key);
            if (opt instanceof ConfigurableOption) {
                ((ConfigurableOption) opt).setConfigurableValue
                    (state.getProperty(key));
                opt.notifyObservers(this);
            }
        }
    }

    // }}}
    // {{{ registrationChanged

    /**
     * This callback is used to register or unregister nested OptionSets.
     *
     * @param   config  The affected Configuration.
     * @param   name    The registration name of the Configurable which may
     *                  be different from the Configurable's name.
     * @param   registered True if the Configurable was just registered,
     *                  false if it was unregistered.
     * @param   valuesKept Whether the Configurable's values were kept, if
     *                  it was just unregistered.
     */
    public void registrationChanged(Configuration config, String name,
                                    boolean registered, boolean valuesKept)

    {
        Option[] opts = getOptions();
        for (int i = 0; i < opts.length; i++) {
            if (opts[i] instanceof NestedOption) {
                OptionSet nested = (OptionSet) opts[i].getValue();
                if (nested != null) {
                    if (registered) {
                        config.registerAs
                            (nested, name + "." + opts[i].getName());
                    } else {
                        config.unregisterAs
                            (nested, name + "." + opts[i].getName(),
                             valuesKept);
                    }
                }
            }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // The Observer interface
    //----------------------------------------------------------------------
    // {{{ update

    /**
     * Update the OptionSet. An OptionSet observes its Options.
     *
     * @param   obs     The Option that has changed.
     * @param   arg     Details about the change.
     */
    public void update(Observable obs, Object arg)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "update(Observable,Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "obs: " + obs + ", " +
                       "arg: " + arg);
        }
        setChanged();
    }

    // }}}

    //----------------------------------------------------------------------
    // Serialization extensions
    //----------------------------------------------------------------------
    // {{{ readObject

    /**
     * Modify serialization. Only save the Options, the names in the map are
     * redundant.
     *
     * @param   stream  The stream to restore from.
     *
     * @throws  IOException     If an I/O error occurs.
     * @throws  ClassNotFoundException  If a class that should be restored
     *                          cannot be found.
     */
    private void readObject(ObjectInputStream stream)
         throws IOException, ClassNotFoundException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "readObject(ObjectInputStream)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "stream: " + stream);
        }
        stream.defaultReadObject();
        options = new HashMap ();
        for (int i = 0; i < savedOptions.length; i++) {
            add(savedOptions[i]);
        }
        savedOptions = null;
    }

    // }}}
    // {{{ writeObject

    /**
     * Modify serialization. Only save the Options, the names in the map are
     * redundant.
     *
     * @param   stream  The stream to write to.
     *
     * @throws  IOException     If an I/O error occurs.
     */
    private void writeObject(ObjectOutputStream stream)
         throws IOException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "writeObject(ObjectOutputStream)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "stream: " + stream);
        }
        savedOptions = getOptions();
        stream.defaultWriteObject();
        savedOptions = null;
    }

    // }}}

    //----------------------------------------------------------------------
    // Standard methods
    //----------------------------------------------------------------------
    // {{{ clone

    /**
     * Creates a deep copy of the OptionSet. The Option's of the OptionSet are
     * cloned as well, but not their values, except for {@link NestedOption
     * NestedOptions}. <p>
     * A shallow copy can be obtained via
     * <pre>
     * OptionSet newSet = new OptionSet("name");
     * newSet.addAll(originalSet);
     * </pre>
     *
     * @return  The new OptionSet.
     */
    public Object clone()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "clone()", "");
        }
        try {
            OptionSet ret = (OptionSet) super.clone();
            ret.options = Collections.synchronizedMap(new HashMap ());
            Option[] opts = getOptions();
            for (int i = 0; i < opts.length; i++) {
                if (logger.level >= Log.DBGDETAIL) {
                    logger.log(Log.DBGDETAIL, "clone()",
                               "cloning " + opts[i]);
                }
                Option nOpt = (Option) opts[i].clone();
                if (nOpt instanceof NestedOption) {
                    OptionSet nested = (OptionSet) nOpt.getValue();
                    if (nested != null) {
                        nOpt.setValue(nested.clone());
                    }
                }
                ret.options.put(opts[i].getName(), nOpt);
            }
            return ret;
        } catch (CloneNotSupportedException ex) {
            // shouldn't happen
            if (logger.level >= Log.ERR) {
                logger.log("clone()", ex);
            }
            return null;
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Miscellaneous
    //----------------------------------------------------------------------
    // {{{ listOptions

    /**
     * Print the options in the OptionSet to a string. Useful for logging.
     *
     * @return  A String containing all options.
     *
     * @since   0.98.0
     */
    public String listOptions()
    {
        StringBuffer sb = new StringBuffer ();
        Option[] opts = getOptions();

        for (int i = 0; i < opts.length; i++) {
            sb.append(opts[i].toString());
            if (i < opts.length - 1) {
                sb.append(TextOption.LINESEP);
            }
        }
        return sb.toString();

    }

    // }}}
}
