// {{{ copyright

/********************************************************************
 *
 * $Id: TabbedOptionGroup.jpp,v 1.11 2001/04/17 16:41:12 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.Component;
import java.awt.BorderLayout;


import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;


import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.DefaultSingleSelectionModel;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;

import de.qfs.lib.gui.Grid;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.MapResourceBundle;

// }}}

// {{{ doc

/**
 * A TabbedOptionGroup uses a JTabbedPane to distribute its OptionGroup
 * members into several tabs. It usees the following resoureces: <p>
 *
 * <TABLE BORDER="1" CELLPADDING="3" CELLSPACING="0" WIDTH="100%">
 * <TR>
 * <TH WIDTH="40%">Resource name</TH>
 * <TH>Usage</TH>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.type</code></TD>
 * <TD>A TabbedOptionGroup is designated by the type <code>tab</code>.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.disabled</code></TD>
 * <TD>If true, the group will be disabled and remain so even if
 * {@link OptionGroup#setEnabled setEnabled(true)} is called (optional,
 * default false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.numtabs</code></TD>
 * <TD>The number of tabs in the group.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.tab<em>n</em>.group</code></TD>
 * <TD>The name of the OptionGroup to put on the n<sup>th</sup> tab,
 * where n is between 1 and numtabs.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.tab<em>n</em>.title</code></TD>
 * <TD>The title for the n<sup>th</sup> tab, where n is between 1 and
 * numtabs.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.frame</code></TD>
 * <TD>Whether the group should have a framed border (optional, default
 * false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.title</code></TD>
 * <TD>The title for a framed border of the group. Ignored if no frame is
 * specified(optional, default empty).</TD>
 * </TR>
 * </TABLE>
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.11 $
 */

// }}}
public class TabbedOptionGroup extends OptionGroup
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (TabbedOptionGroup.class);

    /**
     * The JTabbedPane Component.
     */
    protected JTabbedPane pane;

    // }}}

    //----------------------------------------------------------------------
    // Constructors
    //----------------------------------------------------------------------
    // {{{ TabbedOptionGroup()

    /**
     * No arg constructor for customized OptionGroups.
     */
    protected TabbedOptionGroup ()
    {
    }

    // }}}
    // {{{ TabbedOptionGroup(String,OptionSet)

    /**
     * Create a new TabbedOptionGroup.
     *
     * @param   name    The name of the group.
     * @param   options The Options to edit.
     */
    public TabbedOptionGroup (String name, OptionSet options)
    {
        super (name, options);
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "TabbedOptionGroup(String,OptionSet)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }
        createMembers();
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ getComponent

    /**
     * Get the AWT Component for the OptionGroup.
     *
     * @return  The Component.
     */
    public Component getComponent()
    {
        if (pane == null) {
            makePane();
        }
        return pane;
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ createMembers

    /**
     * Create the tabs for the TabbedOptionGroup.
     */
    protected void createMembers()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "createMembers()", "");
        }
        MapResourceBundle rb = Option.getResources();
        members = new ArrayList();

        int num = getInt("numtabs", -1);
        if (num < 0) {
            if (logger.level >= Log.ERR) {
                logger.log(Log.ERR, "createMembers()",
                           "No numtabs for TabbedOptionGroup " + name);
                return;
            }
        }

        for (int i = 1; i <= num; i++) {
            String group = getString("tab" + i + ".group", null);
            if (group == null) {
                if (logger.level >= Log.ERR) {
                    logger.log(Log.ERR, "createMembers()",
                               "No group for tab " + i +
                               " in TabbedOptionGroup " + name);
                }
            } else {
                OptionGroup og = OptionGroup.makeGroup(group, options);
                if (og != null) {
                    og.setParentGroup(this);
                    members.add(og);
                }
            }
        }
    }

    // }}}
    // {{{ makePane

    /**
     * Create the JTabbedPane for the tab members.
     */
    protected void makePane()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "makePane()", "");
        }
        MapResourceBundle rb = Option.getResources();

        pane = new JTabbedPane ();
        pane.setName("OptionGroup-" + name + "-" +
                     options.getConfigurableName());

        for (int i = 0; i < members.size(); i++) {
            String title = getString("tab" + (i+1) + ".title",
                                     "Missing title");
            Component com = ((OptionGroup) members.get(i)).getComponent();
            if (com instanceof Grid
                && ((Grid) com).needGlue()) {
                ((Grid) com).addGlue();
            }
            if (! (com instanceof JComponent)
                // Compund border doesn't quite work
                || ((JComponent) com).getBorder() != null) {
                JPanel panel = new JPanel ();
                panel.setLayout(new BorderLayout ());
                panel.add(com, BorderLayout.CENTER);
                com = panel;
            }
                ((JComponent) com).setBorder
                (BorderFactory.createEmptyBorder(5, 5, 5, 5));
            pane.addTab(title, com);
        }

        // check values before changing tabs
        pane.setModel(new DefaultSingleSelectionModel () {
            public void setSelectedIndex (int index) {
                int oldidx = getSelectedIndex();
                if (index != oldidx && oldidx != -1) {
                    if (! ((OptionGroup) members.get(oldidx)).checkValues()) {
                        return;
                    }
                }
                super.setSelectedIndex(index);
            }
        });

        pane.setSelectedIndex(0);

        if (disabled) {
            setEnabled(false);
        }

        makeBorder(pane);
    }

    // }}}
}
