// {{{ copyright

/********************************************************************
 *
 * $Id: TableOptionEdit.jpp,v 1.7 2001/04/25 13:10:34 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Insets;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;

import java.util.EventObject;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.DefaultCellEditor;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.JScrollPane;
import javax.swing.JLabel;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.border.Border;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableCellRenderer;

import de.qfs.lib.config.Configuration;
import de.qfs.lib.config.ConfigWrapper;
import de.qfs.lib.gui.Borders;
import de.qfs.lib.gui.SwingUtil;
import de.qfs.lib.gui.WindowsTableCellRenderer;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.Reflector;

// }}}

// {{{ doc

/**
 * JTable based Edit for a TableOption. It uses the following resources:
 * <p>
 *
 * <TABLE BORDER="1" CELLPADDING="3" CELLSPACING="0" WIDTH="100%">
 * <TR>
 * <TH WIDTH="40%">Resource name</TH>
 * <TH>Usage</TH>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.label</code></TD>
 * <TD>The label for the Option (optional, default empty).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.mnemonic</code></TD>
 * <TD>Keyboard accelerator (optional, default none).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.disabled</code></TD>
 * <TD>If true, the edit will be disabled and remain so even if {@link
 * #setEnabled setEnabled(true)} is called (optional, default false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.width</code></TD>
 * <TD>Preferred width of the edit (optional, no default).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.height</code></TD>
 * <TD>Preferred height of the edit (optional, no default).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.weight</code></TD>
 * <TD>The weight (a double) to use for the edit when placed in the {@link
 * de.qfs.lib.gui.Grid Grid} of an {@link OptionGroup OptionGroup} (optional,
 * default 1.0).</TD>
 * </TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.configure</code></TD>
 * <TD>Whether a {@link de.qfs.lib.config.ConfigWrapper ConfigWrapper}
 * should be created for the table and registered with the {@link
 * de.qfs.lib.config.Configuration Configuration} under the name
 * <code><em>name</em>.table</code> (optional, default false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.canedit</code></TD>
 * <TD>Whether the cell values can be edited (optional, default true).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.caninsert</code></TD>
 * <TD>Whether new rows can be added (optional, default true).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.canremove</code></TD>
 * <TD>Whether rows can be deleted (optional, default true).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.canmove</code></TD>
 * <TD>Whether rows can be moved up or down (optional, default true).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.directinsert</code></TD>
 * <TD>Whether new rows are inserted directly (true) or created via an
 * {@link OptionDialog OptionDialog} (false) (optional, default
 * <code>(!canedit)</code>).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.directedit</code></TD>
 * <TD>Whether typing a key starts editing the focused cell. This is the
 * default value for the whole table and can be overridden for each column
 * separately (optional, default true).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.columns</code></TD>
 * <TD>Whitespace separated list of column identifiers (required).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.column.<em>id</em>.name</code></TD>
 * <TD>The name of the column to display in the table's header (optional,
 * default <code><em>id</em></code>).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.column.<em>id</em>.type</code></TD>
 * <TD>The type of the column. Currently supported are <code>string</code>,
 * <code>taxt, </code><code>boolean</code>and <code>int</code> (optional,
 * default string).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.column.<em>id</em>.nullok</code></TD>
 * <TD>Whether empty values are legal in this column (optional, default
 * false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.column.<em>id</em>.negativeok</code></TD>
 * <TD>Whether negative values are legal in this column (optional, default
 * false, only for columns of type <code>int</code>).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.column.<em>id</em>.default</code></TD>
 * <TD>Default value for new rows (optional, default null).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.column.<em>id</em>.disabled</code></TD>
 * <TD>If true, values in this column cannot be edited (optional, default
 * false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.column.<em>id</em>.directedit</code></TD>
 * <TD>Whether typing a key starts editing the focused cell (optional, default
 * taken from the value for the whole table).</TD>
 * </TR>
 * </TABLE>
 * <p>
 *
 * To edit a row's values in a dialog, an {@link OptionDialog OptionDialog}
 * with the same name as the edit's option is used. You can customize this
 * dialog in the usual way. The {@link OptionGroup OptionGroup} for this
 * dialog also has the same name. If the resource
 * <code>option.group.<em>name</em>.type</code> is not defined, default values
 * are set to create a plain group with an edit for each column. The labels of
 * these edits are copied from the table hader. If you define the resource
 * <code>option.group.<em>name</em>.type</code> you take full responsibility
 * for all of the groups resources.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.7 $
 * @since       0.98.1
 */

// }}}
public class TableOptionEdit
    extends OptionEdit
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (TableOptionEdit.class);

    /**
     * The label of the edit.
     * @serial
     */
    protected JLabel label;

    /**
     * The table of the edit.
     * @serial
     */
    protected Table table;

    /**
     * The config wrapper for the table.
     */
    protected ConfigWrapper wrapper;

    /**
     * The columns of the table.
     */
    protected Column columns[];

    /**
     * Whether the edit is currently enabled.
     * @serial
     */
    protected boolean enabled = true;

    /**
     * Whether editing via dialog is allowed.
     */
    protected boolean canedit;

    /**
     * Whether adding rows is allowed.
     */
    protected boolean caninsert;

    /**
     * Whether removing rows is allowed.
     */
    protected boolean canremove;

    /**
     * Whether moving rows is allowed.
     */
    protected boolean canmove;

    /**
     * Whether to insert new rows directly.
     */
    protected boolean directinsert;

    /**
     * Whether to start editing immediately on a keypress.
     */
    protected boolean directedit;

    /**
     * The y-weight of the table.
     */
    protected double weight = 1.0;

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ TableOptionEdit

    /**
     * Create a new TableOptionEdit.
     */
    public TableOptionEdit ()
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ init

    /**
     * Initialize the Edit.
     */
    protected void init()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "init()", "");
        }
        String cols = getString("columns", null);
        if (cols == null) {
            if (logger.level >= Log.ERR) {
                logger.log(Log.ERR, "init()",
                           "No columns for TableOptionEdit " +
                           option.getName());
            }
            return;
        }

        // needed as default for the columns
        directedit = getBoolean("directedit", true);

        int cnt = 0;
        for (StringTokenizer st = new StringTokenizer (cols);
             st.hasMoreTokens(); st.nextToken()) {
            cnt++;
        }
        columns = new Column[cnt];
        cnt = 0;
        for (StringTokenizer st = new StringTokenizer (cols);
             st.hasMoreTokens();) {
            columns[cnt++] = new Column (st.nextToken());
        }

        canedit = getBoolean("canedit", true);
        caninsert = getBoolean("caninsert", true);
        canremove = getBoolean("canremove", true);
        canmove = getBoolean("canmove", true);
        directinsert = getBoolean("directinsert", ! canedit);

        setLayout(new BorderLayout ());

        disabled = getBoolean("disabled", false);

        table = new Table ();

        String name = getString("label", null);
        if (name != null) {
            label = new JLabel (name);
            label.setName("Label-" + option.getName());
            add(label, BorderLayout.NORTH);
            label.setForeground(UIManager.getColor
                                (disabled ? "textInactiveText" : "textText"));
            label.setLabelFor(table);
            String mnemo = getString("mnemonic", null);
            if (mnemo != null && mnemo.length() > 0) {
                label.setDisplayedMnemonic(mnemo.charAt(0));
            }
        }

        Component buttons = table.getButtons();
        Box box = Box.createVerticalBox();
        box.add(buttons);
        box.add(new JScrollPane (table));
        add(box, BorderLayout.CENTER);

        setValue(option.getValue());
        table.setName("Table-" + option.getName());
        table.getTableHeader().setName("TableHeader-" + option.getName());

        int width = getInt("width", 50);
        int height = getInt("height", 100);
        setPreferredSize(new Dimension (width, height));
        if (getBoolean("configure", false)) {
            wrapper = ConfigWrapper.makeWrapper
                (table, option.getName() + ".table");
            Configuration.instance().register(wrapper);
        }

        String sWeight = getString("weight", null);
        if (sWeight != null) {
            try {
                weight = new Double(sWeight).doubleValue();
            } catch (NumberFormatException ex) {
                if (logger.level >= Log.WRN) {
                    logger.log(Log.WRN, "init()", ex);
                }
            }
        }

        if (disabled) {
            setEnabled(false);
        }
    }

    // }}}
    // {{{ update

    /**
     * Update the display for a possibly new Option.
     *
     * @param   option The new option.
     */
    public void update(Option option)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "update(Option)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "option: " + option);
        }
        this.option = option;
        Object val = option.getValue();
        if (wrapper != null) {
            Configuration.instance().unregister(wrapper, true);
        }
        setValue(val);
        if (getBoolean("configure", false)) {
            wrapper = ConfigWrapper.makeWrapper
                (table, option.getName() + ".table");
            Configuration.instance().register(wrapper);
        }
    }

    // }}}
    // {{{ checkValue

    /**
     * Check the currently entered value of the ValueEdit. If the value is
     * not valid, request the focus and bring up a dialog box explaining why.
     *
     * @return  True if the value is valid, false otherwise.
     */
    public boolean checkValue()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "checkValue()", "");
        }
        return true;
    }

    // }}}
    // {{{ readValue

    /**
     * Read the current value of the table.
     */
    public void readValue()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "readValue()", "");
        }
        int rows = table.getRowCount();
        int columns = table.getColumnModel().getColumnCount();

        if (rows > 0 && columns > 0) {
            Object[][] val = new Object[rows][];
            for (int i = 0; i < val.length; i++) {
                val[i] = new Object[columns];
                for (int j = 0; j < columns; j++) {
                    val[i][j] = table.getModel().getValueAt(i, j);
                }
            }
            option.setValue(val);
        } else {
            option.setValue(null);
        }
    }

    // }}}
    // {{{ getWeight

    /**
     * Get the y-weight for the TableOptionEdit.
     *
     * @return  1.0
     */
    public double getWeight()
    {
        return weight;
    }

    // }}}

    // {{{ setEnabled

    /**
     * Set the enabled flag.
     *
     * @param   enabled The flag to set.
     */
    public void setEnabled(boolean enabled)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setEnabled(boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "enabled: " + enabled +
                       ", disabled: " + disabled);
        }
        if (disabled) {
            enabled = false;
        }

        this.enabled = enabled;
        table.setEnabled(enabled);
        if (label != null) {
            if (enabled) {
                label.setForeground(UIManager.getColor("textText"));
            } else {
                label.setForeground(UIManager.getColor("textInactiveText"));
            }
        }
    }

    // }}}
    // {{{ isEnabled

    /**
     * See if the ValueEdit is enabled.
     *
     * @return  True if it is enabled.
     */
    public boolean isEnabled()
    {
        return enabled;
    }

    // }}}
    // {{{ requestFocus

    /**
     * Request the focus for the ValueEdit.
     */
    public void requestFocus()
    {
        table.requestFocus();
    }

    // }}}
    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ setValue

    /**
     * Set the Edit's value. Same format as the data parameter in
     * {@link #TableOptionEdit the constructor}.
     *
     * @param   value   The value of the Edit.
     */
    public void setValue (Object value)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setValue(Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "value: " + value);
        }
        Object[][] data =
            (value == null ? new Object[0][] : (Object[][]) value);

        table.setModel (new Model (data, columns.length));
        if (table.getRowCount() > 0) {
            table.setRowSelectionInterval(0,0);
            table.setColumnSelectionInterval(0,0);
        } else {
            table.clearSelection();
        }
        for (int i = 0; i < columns.length; i++) {
            table.getColumnModel().getColumn(i).setIdentifier
                (columns[i].id);
        }

    }

    // }}}

    //----------------------------------------------------------------------
    // Helper classes
    //----------------------------------------------------------------------
    // {{{ class Column

    /**
     * Structure that holds column data.
     */
    protected class Column
    {
        // {{{ variables

        /**
         * The identifier for the column.
         */
        public String id;

        /**
         * The name to display.
         */
        public String name;

        /**
         * The class of the column's values.
         */
        public Class clazz;

        /**
         * Default value for a new cell.
         */
        public Object dflt;

        /**
         * An option for the column's value.
         */
        public Option option;

        /**
         * Whether the column is disabled - i.e. not editable.
         */
        public boolean disabled;

        /**
         * Whether the column should start editing on any keypress.
         */
        public boolean directedit;

        // }}}

        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ Column

        /**
         * Crate a new column.
         *
         * @param       id    The column identifier.
         */
        public Column(String id)
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "Column(String)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "id: " + id);
            }

            String col =
                TableOptionEdit.this.option.getName() + ".column." + id;
            this.id = id;
            name = getString("column." + id + ".name", id);

            String type = getString("column." + id + ".type", "string");
            if ("string".equals(type)) {
                clazz = String.class;
                option = new StringOption
                    (col, null,
                     getBoolean("column." + id + ".nullok", false));
            } else if ("text".equals(type)) {
                clazz = String.class;
                option = new TextOption
                    (col, null,
                     getBoolean("column." + id + ".nullok", true));
            } else if ("int".equals(type)) {
                clazz = Integer.class;
                option = new IntegerOption
                    (col, 0,
                     getBoolean("column." + id + ".nullok", false),
                     getBoolean("column." + id + ".negativeok", false));
            } else if ("boolean".equals(type)) {
                clazz = Boolean.class;
                option = new BooleanOption
                    (col, false);
            }

            String df = getString("column." + id + ".default", null);
            if (df != null) {
                try {
                    dflt = clazz.getConstructor(new Class[] {String.class})
                        .newInstance(new String[] {df});
                } catch (Exception ex) {
                    if (logger.level >= Log.ERR) {
                        logger.log(Log.ERR, "Column(String)", ex);
                    }
                }
            }

            disabled = getBoolean("column." + id + ".disabled", false);

            this.directedit = getBoolean("column." + id + ".directedit",
                                    TableOptionEdit.this.directedit);

            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "Column(String)",
                           "New column: " + this);
            }

        }

        // }}}

        //------------------------------------------------------------------
        // Public methods
        //------------------------------------------------------------------
        // {{{ toString

        /**
         * Get a string representation of the Column.
         *
         * @return      The string for the column.
         */
        public String toString()
        {
            return super.toString() + "[" +
                "id=" + id +
                ",name=" + name +
                ",clazz=" + clazz +
                ",dflt=" + dflt +
                ",disabled=" + disabled +
                ",directedit=" + directedit +
                "]";
        }

        // }}}
    }

    // }}}
    // {{{ class Model

    /**
     * The table model used by the edit.
     */
    protected class Model
        extends DefaultTableModel
    {
        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ Model

        /**
         * Create a new Model.
         *
         * @param   data    The table data.
         * @param   col     The number of columns (javac doesn't allow access
         *                  to columns in this constructor).
         */
        public Model (Object[][] data, int col)
        {
            super(data, new Object[col]);
            Vector ids = new Vector ();
            for (int i = 0; i < columns.length; i++) {
                ids.addElement(columns[i].id);
            }
            setColumnIdentifiers(ids);

        }

        // }}}

        //------------------------------------------------------------------
        // Public methods
        //------------------------------------------------------------------
        // {{{ getColumnClass

        /**
         * Get the class for a column.
         *
         * @param   column  The column index.
         *
         * @return  The class for the column.
         */
        public Class getColumnClass(int column)
        {
            return columns[column].clazz;
        }

        // }}}
        // {{{ getColumnName

        /**
         * Get the name for a column.
         *
         * @param   column  The column index.
         *
         * @return  The name for the column.
         */
        public String getColumnName(int column)
        {
            return columns[column].name;
        }

        // }}}
    }

    // }}}
    // {{{ class Table

    protected class Table
        extends JTable
    {
        // {{{ variables

        /**
         * The handler for the new/del/up/down actions.
         */
        protected ActionHandler handler;

        /**
         * The "edit row" button.
         */
        protected JButton editButton;

        /**
         * The "insert new row" button.
         */
        protected JButton newButton;

        /**
         * The "remove row" button.
         */
        protected JButton removeButton;

        /**
         * The "move row up" button.
         */
        protected JButton upButton;

        /**
         * The "move row down" button.
         */
        protected JButton downButton;

        /**
         * The box with the buttons.
         */
        protected Box buttonBox;

        /**
         * The dialog for the row edit.
         */
        protected OptionDialog editDialog;

        /**
         * The OptionSet for the dialog.
         */
        protected OptionSet rowOptions;

        // }}}

        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ Table

        /**
         * Create a new Table.
         */
        public Table ()
        {
            JTableHeader header = new JTableHeader (getColumnModel()) {
                public boolean isFocusTraversable()
                {
                    return false;
                }
            };
            header.setRequestFocusEnabled(false);
            setTableHeader(header);
            setRowSelectionAllowed(false);
            setAutoResizeMode(AUTO_RESIZE_OFF);
            // Renderers
            setDefaultRenderer(String.class, new WindowsTableCellRenderer ());
            WindowsTableCellRenderer wr = new WindowsTableCellRenderer ();
            wr.setHorizontalAlignment(JLabel.RIGHT);
            setDefaultRenderer(Integer.class, wr);
            setDefaultRenderer(Boolean.class, new BooleanRenderer ());
            // Editors
            setDefaultEditor(String.class, new StringEditor ());
            setDefaultEditor(Integer.class, new NumberEditor ());
            setDefaultEditor(Boolean.class, new BooleanEditor (this));
            // Override buggy F2 key binding
            this.registerKeyboardAction(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    int row = getSelectionModel().getAnchorSelectionIndex();
                    int column = getColumnModel().getSelectionModel()
                        .getAnchorSelectionIndex();
                    if ((getColumnClass(column) == String.class
                         || getColumnClass(column) == Integer.class)
                        && editCellAt(row, column, e)) {
                        Component other = Table.this.hasFocus()
                            ? getEditorComponent() : Table.this;
                        if (other != null) {
                            other.requestFocus();
                        }
                    }
                }
            }, "F2",
            KeyStroke.getKeyStroke(KeyEvent.VK_F2, 0),
            WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
            // Direct edit
            this.addKeyListener(new KeyAdapter() {
                public void keyPressed(KeyEvent e) {
                    int row = getSelectionModel().getAnchorSelectionIndex();
                    int column = getColumnModel().getSelectionModel()
                        .getAnchorSelectionIndex();
                    if (columns[column].directedit
                        && e.getKeyChar() != 0
                        && e.getKeyChar() != 65535
                        && e.getKeyCode() != KeyEvent.VK_ENTER
                        && (e.getModifiers() &
                            (e.ALT_MASK | e.CTRL_MASK | e.META_MASK)) == 0
                        && (getColumnClass(column) == String.class
                            || getColumnClass(column) == Integer.class)
                        && editCellAt(row, column, e)) {
                        Component other = Table.this.hasFocus()
                            ? getEditorComponent() : Table.this;
                        if (other != null) {
                            other.requestFocus();
                        }
                        if (getCellEditor() != null) {
                            getCellEditor().shouldSelectCell(e);
                        }
                    }
                }
            });
            // Edit booleans with ' ' key binding
            this.registerKeyboardAction(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    int row = getSelectionModel().getAnchorSelectionIndex();
                    int column = getColumnModel().getSelectionModel()
                        .getAnchorSelectionIndex();
                    if (getColumnClass(column) == Boolean.class
                        && editCellAt(row, column, e)) {
                        BooleanEditor editor = (BooleanEditor)
                            getCellEditor();
                        JCheckBox cb = (JCheckBox) editor.getComponent();
                        cb.setSelected(! cb.isSelected());
                        editor.stopCellEditing();
                    }
                }
            }, "space",
            KeyStroke.getKeyStroke(KeyEvent.VK_SPACE, 0),
            WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);

            // Remove default escape and return bindings.
            this.unregisterKeyboardAction
                (KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0));
            this.unregisterKeyboardAction
                (KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0));
            // Special version for JDK 1.3
            Object inputMap = Reflector.safeCall
                (this, "getInputMap", Integer.TYPE,
                 new Integer (WHEN_ANCESTOR_OF_FOCUSED_COMPONENT));
            if (inputMap != null) {
                Reflector.safeCall
                    (inputMap, "put",
                     new Class[] {KeyStroke.class, Object.class},
                     new Object[] {
                         KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0),
                         "dummy"});
                Reflector.safeCall
                    (inputMap, "put",
                     new Class[] {KeyStroke.class, Object.class},
                     new Object[] {
                         KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0),
                         "dummy"});
            }
            handler = new ActionHandler ();

            if (canedit) {
                this.registerKeyboardAction
                    (handler, "edit",
                     KeyStroke.getKeyStroke(KeyEvent.VK_ENTER,
                                            KeyEvent.SHIFT_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
                this.registerKeyboardAction
                    (handler, "edit",
                     KeyStroke.getKeyStroke(KeyEvent.VK_ENTER,
                                            KeyEvent.ALT_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
            }
            if (caninsert) {
                this.registerKeyboardAction
                    (handler, "insert",
                     KeyStroke.getKeyStroke(KeyEvent.VK_INSERT,
                                            KeyEvent.SHIFT_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
            }
            if (canremove) {
                this.registerKeyboardAction
                    (handler, "remove",
                     KeyStroke.getKeyStroke(KeyEvent.VK_DELETE,
                                            KeyEvent.SHIFT_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
            }
            if (canmove) {
                this.registerKeyboardAction
                    (handler, "up",
                     KeyStroke.getKeyStroke(KeyEvent.VK_UP,
                                            KeyEvent.SHIFT_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
                this.registerKeyboardAction
                    (handler, "down",
                     KeyStroke.getKeyStroke(KeyEvent.VK_DOWN,
                                            KeyEvent.SHIFT_MASK),
                     WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
            }

        }

        // }}}

        //------------------------------------------------------------------
        // Getters and setters
        //------------------------------------------------------------------
        // {{{ getButtons

        /**
         * Get the component with the buttons.
         *
         * @return  The button box.
         */
        public Component getButtons()
        {
            if (buttonBox == null) {
                buttonBox = Box.createHorizontalBox();
                if (canedit) {
                    editButton = new JButton () {
                        public boolean isFocusTraversable() {
                            return false;
                        }
                    };
                    editButton.setName("EditButton-" + option.getName());
                    editButton.setRequestFocusEnabled(false);
                    editButton.setMargin(new Insets (0,0,0,0));
                    editButton.setIcon
                        (Option.getResources().getIcon
                         ("option.table.row.edit.icon", null));
                    editButton.setToolTipText
                        (Option.getResources().getString
                         ("option.table.row.edit.tooltip", null));
                    editButton.setActionCommand("edit");
                    editButton.addActionListener(handler);
                    buttonBox.add(editButton);
                }
                if (caninsert) {
                    newButton = new JButton () {
                        public boolean isFocusTraversable() {
                            return false;
                        }
                    };
                    newButton.setName("NewButton-" + option.getName());
                    newButton.setRequestFocusEnabled(false);
                    newButton.setMargin(new Insets (0,0,0,0));
                    newButton.setIcon
                        (Option.getResources().getIcon
                         ("option.table.row.new.icon", null));
                    newButton.setToolTipText
                        (Option.getResources().getString
                         ("option.table.row.new.tooltip", null));
                    newButton.setActionCommand("insert");
                    newButton.addActionListener(handler);
                    buttonBox.add(newButton);
                }
                if (canremove) {
                    removeButton = new JButton () {
                        public boolean isFocusTraversable() {
                            return false;
                        }
                    };
                    removeButton.setName("RemoveButton-" + option.getName());
                    removeButton.setRequestFocusEnabled(false);
                    removeButton.setMargin(new Insets (0,0,0,0));
                    removeButton.setIcon
                        (Option.getResources().getIcon
                         ("option.table.row.remove.icon", null));
                    removeButton.setToolTipText
                        (Option.getResources().getString
                         ("option.table.row.remove.tooltip", null));
                    removeButton.setActionCommand("remove");
                    removeButton.addActionListener(handler);
                    buttonBox.add(removeButton);
                }
                if (canmove) {
                    upButton = new JButton () {
                        public boolean isFocusTraversable() {
                            return false;
                        }
                    };
                    upButton.setName("UpButton-" + option.getName());
                    upButton.setRequestFocusEnabled(false);
                    upButton.setMargin(new Insets (0,0,0,0));
                    upButton.setIcon
                        (Option.getResources().getIcon
                         ("option.table.row.up.icon", null));
                    upButton.setToolTipText
                        (Option.getResources().getString
                         ("option.table.row.up.tooltip", null));
                    upButton.setActionCommand("up");
                    upButton.addActionListener(handler);
                    buttonBox.add(upButton);

                    downButton = new JButton () {
                        public boolean isFocusTraversable() {
                            return false;
                        }
                    };
                    downButton.setName("DownButton-" + option.getName());
                    downButton.setRequestFocusEnabled(false);
                    downButton.setMargin(new Insets (0,0,0,0));
                    downButton.setIcon
                        (Option.getResources().getIcon
                         ("option.table.row.down.icon", null));
                    downButton.setToolTipText
                        (Option.getResources().getString
                         ("option.table.row.down.tooltip", null));
                    downButton.setActionCommand("down");
                    downButton.addActionListener(handler);
                    buttonBox.add(downButton);
                }
                buttonBox.add(Box.createHorizontalGlue());
            }
            return buttonBox;
        }

        // }}}

        //------------------------------------------------------------------
        // Actions
        //------------------------------------------------------------------
        // {{{ moveUp

        /**
         * Move the selected row up by one.
         */
        protected void moveUp()
        {
            int row = getSelectedRow();
            if (row > 0) {
                if (isEditing()) {
                    if (getCellEditor() != null) {
                        getCellEditor().cancelCellEditing();
                    }
                }
                ((Model) getModel()).moveRow(row, row, row - 1);
                setRowSelectionInterval(row - 1, row - 1);
            }
        }

        // }}}
        // {{{ moveDown

        /**
         * Move the selected row down by one.
         */
        protected void moveDown()
        {
            int row = getSelectedRow();
            if (row >= 0 && row < getRowCount() - 1) {
                if (isEditing()) {
                    if (getCellEditor() != null) {
                        getCellEditor().cancelCellEditing();
                    }
                }
                ((Model) getModel()).moveRow(row, row, row + 1);
                setRowSelectionInterval(row + 1, row + 1);
            }
        }

        // }}}
        // {{{ insertRow

        /**
         * Create and insert a new row.
         */
        protected void insertRow()
        {
            Object[] data;
            if (isEditing()) {
                if (getCellEditor() != null) {
                    getCellEditor().cancelCellEditing();
                }
            }
            int row = getSelectedRow() + 1;
            if (canedit && ! directinsert) {
                initDialog(-1);
                int ret = editDialog.doModal
                    (SwingUtilities.windowForComponent(this), true);
                this.requestFocus();
                if (ret != 0) {
                    return;
                }
                data = new Object[columns.length];
                for (int i = 0; i < columns.length; i++) {
                    data[i] = rowOptions.getValue
                        (option.getName() + ".column." + columns[i].id, null);
                }
            } else {
                data = new Object[columns.length];
                for (int i = 0; i < columns.length; i++) {
                    data[i] = columns[i].dflt;
                }
                this.requestFocus();
            }

            ((Model) getModel()).insertRow(row, data);
            setRowSelectionInterval(row, row);
        }

        // }}}
        // {{{ removeRow

        /**
         * Remove the selected row.
         */
        protected void removeRow()
        {
            int row = getSelectedRow();
            if (row >= 0 && row < getRowCount()) {
                if (isEditing()) {
                    if (getCellEditor() != null) {
                        getCellEditor().cancelCellEditing();
                    }
                }
                ((Model) getModel()).removeRow(row);
                if (getRowCount() == 0) {
                    clearSelection();
                } else if (row == getRowCount()) {
                    setRowSelectionInterval(row - 1, row - 1);
                } else {
                    setRowSelectionInterval(row, row);
                }
            }
        }

        // }}}
        // {{{ editRow

        /**
         * Bring up a dialog for the selected row.
         */
        protected void editRow()
        {
            int row = getSelectedRow();
            if (row >= 0 && row < getRowCount()) {
                if (isEditing()) {
                    if (getCellEditor() != null) {
                        if (! getCellEditor().stopCellEditing()) {
                            getCellEditor().cancelCellEditing();
                        }
                    }
                }
                initDialog(row);
                if (editDialog.doModal(SwingUtilities.windowForComponent(this),
                                       true) == 0) {
                    for (int i = 0; i < columns.length; i++) {
                        ((Model) getModel()).setValueAt
                            (rowOptions.getValue
                             (option.getName() + ".column." + columns[i].id,
                              null),
                             row, i);
                    }
                }
                this.requestFocus();
            }
        }

        // }}}

        //------------------------------------------------------------------
        // JTable methods
        //------------------------------------------------------------------
        // {{{ isManagingFocus

        public boolean isManagingFocus()
        {
            return false;
        }

        // }}}
        // {{{ isFocusTraversable

        public boolean isFocusTraversable()
        {
            return this.isEnabled();
        }

        // }}}
        // {{{ isRequestFocusEnabled

        public boolean isRequestFocusEnabled()
        {
            return this.isEnabled();
        }

        // }}}
        // {{{ isCellEditable

        public boolean isCellEditable(int row, int column)
        {
            return this.isEnabled()
                && ! columns[column].disabled
                && super.isCellEditable(row, column);
        }

        // }}}
        // {{{ setEnabled

        /**
         * Set the enabled state of the table. Also changes the buttons.
         *
         * @param   enabled The new enabled state.
         */
        public void setEnabled(boolean enabled)
        {
            super.setEnabled(enabled);
            if (buttonBox != null) {
                newButton.setEnabled(enabled);
                removeButton.setEnabled(enabled);
                upButton.setEnabled(enabled);
                downButton.setEnabled(enabled);
            }
        }

        // }}}

        //------------------------------------------------------------------
        // Helper methods
        //------------------------------------------------------------------
        // {{{ initDialog

        protected void initDialog(int row)
        {
            String name = option.getName() + ".row";
            if (rowOptions == null) {
                rowOptions = new OptionSet (name);
                for (int i = 0; i < columns.length; i++) {
                    rowOptions.add(columns[i].option);
                }
            }
            for (int i = 0; i < columns.length; i++) {
                if (row >= 0) {
                    rowOptions.getOption
                        (option.getName() + ".column." + columns[i].id)
                        .setValue(getModel().getValueAt(row, i));
                } else {
                    rowOptions.getOption
                        (option.getName() + ".column." + columns[i].id)
                        .setValue(columns[i].dflt);
                }
            }
            if (editDialog == null) {
                // fixup option resources
                Properties props = new Properties ();
                if (Option.getResources().getString
                    ("option.group." + name + ".type", null) == null) {
                    props.put("option.group." + name + ".type", "plain");
                    StringBuffer sb = new StringBuffer ();
                    for (int i = 0; i < columns.length; i++) {
                        if (i > 0) {
                            sb.append(' ');
                        }
                        sb.append(option.getName());
                        sb.append(".column.");
                        sb.append(columns[i].id);
                        props.put("option." + option.getName() + ".column." +
                                  columns[i].id + ".label", columns[i].name);
                    }
                    props.put("option.group." + name + ".members",
                              sb.toString());
                }
                if (Option.getResources().getString
                    ("dialog." + name + ".title", null) == null) {
                    props.put("dialog." + name + ".title",
                              Option.getResources().getString
                              ("option.table.row.dialog.title",
                               "Edit table row"));
                }
                if (props.size() > 0) {
                    Option.getResources().addResources(props);
                }
                Window win = SwingUtilities.windowForComponent(this);
                if (win instanceof Frame) {
                    editDialog = new OptionDialog
                        (name, (Frame) win, rowOptions, name);

                } else if (win instanceof Dialog) {
                    editDialog = new OptionDialog
                        (name, (Dialog) win, rowOptions, name);

                } else {
                    editDialog = new OptionDialog
                        (name, (Frame) null, rowOptions, name);
                }
                SwingUtil.closeOnEscape(editDialog);
            } else {
                editDialog.update(rowOptions);
            }
        }

        // }}}

        //------------------------------------------------------------------
        // Helper classes
        //------------------------------------------------------------------
        // {{{ class ActionHandler

        /**
         * Action listener for the table row actions.
         */
        protected class ActionHandler
            implements ActionListener
        {
            // {{{ actionPerformed

            /**
             * Perform the action.
             *
             * @param       e     The event containing the details.
             */
            public void actionPerformed(ActionEvent e)
            {
                String name = e.getActionCommand();
                if ("up".equals(name)) {
                    moveUp();
                } else if ("down".equals(name)) {
                    moveDown();
                } else if ("insert".equals(name)) {
                    insertRow();
                } else if ("remove".equals(name)) {
                    removeRow();
                } else if ("edit".equals(name)) {
                    editRow();
                }
            }

            // }}}
        }

        // }}}
    }

    // }}}
    // {{{ class BooleanRenderer

    /**
     * Rendere for boolean values.
     */
    private class BooleanRenderer
        extends JPanel
        implements TableCellRenderer
    {
        // {{{ variables

        /**
         * The CheckBox that renders the value.
         */
        private JCheckBox cb;

        // }}}

        // {{{ BooleanRenderer

        /**
         * Create a new BooleanRenderer.
         */
        public BooleanRenderer ()
        {
            this.setLayout(new BorderLayout());
            cb = new JCheckBox ();
            cb.setHorizontalAlignment(cb.CENTER);
            this.add(cb, BorderLayout.CENTER);
            this.setBorder(BorderFactory.createLineBorder(Color.black));
        }

        // }}}
        // {{{ getTableCellRendererComponent

        /**
         * Get the component that renders the value.
         *
         * @param   table   The table to paint.
         * @param   value   The value of the cell.
         * @param   selected        Whether the cell is selected.
         * @param   focus           Whether the cell has the focus - ignored.
         * @param   row     The cell's row index.
         * @param   column  The cell's column index.
         *
         * @return      This
         */
        public Component getTableCellRendererComponent
            (JTable table, Object value,
             boolean isSelected, boolean hasFocus, int row, int column) {
	    if (isSelected) {
	        cb.setForeground(table.getSelectionForeground());
	        cb.setBackground(table.getSelectionBackground());
	    }
	    else {
	        cb.setForeground(table.getForeground());
	        cb.setBackground(table.getBackground());
	    }
            cb.setSelected((value != null && ((Boolean)value).booleanValue()));
            this.setBorder(hasFocus ? Borders.STIPPLE : null);
            return this;
        }

        // }}}
    }

    // }}}
    // {{{ class Editor

    /**
     * Baseclass for the cell editors.
     */
    protected class Editor
        extends DefaultCellEditor
    {
        // {{{ variables

        /**
         * The column being edited.
         */
        Column column;

        // }}}

        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ Editor(JTextField)

        /**
         * Create a new Editor.
         *
         * @param       com     The actual editor component.
         */
        public Editor (JTextField com)
        {
            super (com);
        }

        // }}}
        // {{{ Editor(JCheckBox)

        /**
         * Create a new Editor.
         *
         * @param       com     The actual editor component.
         */
        public Editor (JCheckBox com)
        {
            super (com);
        }

        // }}}

        //------------------------------------------------------------------
        // Public methods
        //------------------------------------------------------------------
        // {{{ getTableCellEditorComponent

        /**
         * Get the editor component for a cell.
         *
         * @param       table   The table being edited.
         * @param       value   The cell value.
         * @param       isSelected      Whether the cell is selected.
         * @param       row     The cell's row index.
         * @param       column  The cell's column index.
         *
         * @return      The editor component.
         */
	public Component getTableCellEditorComponent
            (JTable table, Object value, boolean isSelected,
             int row, int column)
        {
            this.column = columns[column];
	    return super.getTableCellEditorComponent
                (table, value, isSelected, row, column);
	}

        // }}}
        // {{{ isCellEditable

        /**
         * Query whether a cell is editable.
         *
         * @param       anEvent The event initiating the edit.
         *
         * @return      False if anEvent is null, otherwise the result of the
         *              superclass' method.
         */
        public boolean isCellEditable(EventObject anEvent)
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "isCellEditable(EventObject)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "anEvent: " + anEvent);
            }

            // Don't just start editing on any keypress
            if (anEvent == null) {
                if (logger.level >= Log.DBG) {
                    logger.log(Log.DBG, "isCellEditable(EventObject)",
                               "isCellEditable: ignore");
                }
                return false;
            }
            return super.isCellEditable(anEvent);
        }

        // }}}
        // {{{ shouldSelectCell

        /**
         * Query whether the cell should be selected.
         *
         * @param       anEvent The event that started the editing.
         *
         * @return      Depends on the superclass.
         */
        public boolean shouldSelectCell(EventObject anEvent)
        {
            if (anEvent instanceof KeyEvent) {
                if (editorComponent instanceof JTextField) {
                    ((JTextField) editorComponent).setText(null);
                }
            }
            return super.shouldSelectCell(anEvent);
        }

        // }}}
    }

    // }}}
    // {{{ class BooleanEditor

    /**
     * Editor for Boolean table cell.
     */
    protected class BooleanEditor
        extends Editor
    {
        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ BooleanEditor

        /**
         * Create a new BooleanEditor.
         */
        public BooleanEditor (JTable table)
        {
            super (new JCheckBox ());
            JCheckBox cb = (JCheckBox) editorComponent;
            cb.setForeground(table.getForeground());
            cb.setBackground(table.getBackground());
            cb.setHorizontalAlignment(cb.CENTER);
        }

        // }}}

        //------------------------------------------------------------------
        // Public methods
        //------------------------------------------------------------------
        // {{{ shouldSelectCell

        /**
         * Query whether the cell should be selected.
         *
         * @param       anEvent The event that started the editing.
         *
         * @return      False if the event is a mouse event.
         */
        public boolean shouldSelectCellXXX(EventObject anEvent)
        {
            if (anEvent instanceof MouseEvent) {
                return false;
            }
            return super.shouldSelectCell(anEvent);
        }

        // }}}
    }

    // }}}
    // {{{ class StringEditor

    /**
     * Editor for String table cell.
     */
    public class StringEditor
        extends Editor
    {
        // {{{ variables

	Object value;

        // }}}

        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ StringEditor

        /**
         * Create a new StringEditor.
         */
        public StringEditor ()
        {
            super (new JTextField ());
            JTextField tf = (JTextField) editorComponent;
            // Reset binding for return key
            tf.addKeyListener(new KeyAdapter () {
                public void keyPressed(KeyEvent e) {
                    if (e.getKeyCode() == KeyEvent.VK_ENTER
                        && e.getModifiers() == 0) {
                        stopCellEditing();
                        e.consume();
                    }
                }
            });
            // Bind escape to cancel editing
            tf.registerKeyboardAction(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    cancelCellEditing();
                }
            },
            KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0),
            WHEN_FOCUSED);
        }

        // }}}

        //------------------------------------------------------------------
        // Public methods
        //------------------------------------------------------------------
        // {{{ stopCellEditing

	public boolean stopCellEditing()
        {
	    value = super.getCellEditorValue();
	    if (! column.option.checkValue
                (null, "".equals(value) ? null : value)) {
		((JComponent) getComponent()).setBorder
                    (BorderFactory.createCompoundBorder
                     (BorderFactory.createLineBorder(Color.red),
                      BorderFactory.createEmptyBorder(1, 1, 1, 1)));
		return false;
	    }
	    return super.stopCellEditing();
	}

        // }}}
        // {{{ getTableCellEditorComponent

        /**
         * Get the editor component for a cell.
         *
         * @param       table   The table being edited.
         * @param       value   The cell value.
         * @param       isSelected      Whether the cell is selected.
         * @param       row     The cell's row index.
         * @param       column  The cell's column index.
         *
         * @return      The editor component.
         */
	public Component getTableCellEditorComponent
            (JTable table, Object value, boolean isSelected,
             int row, int column)
        {
	    this.value = null;
            this.column = columns[column];
	    ((JComponent)getComponent()).setBorder
                (BorderFactory.createCompoundBorder
                 (BorderFactory.createLineBorder(Color.black),
                  BorderFactory.createEmptyBorder(1, 1, 1, 1)));
	    return super.getTableCellEditorComponent
                (table, value, isSelected, row, column);
	}

        // }}}
        // {{{ getCellEditorValue

	public Object getCellEditorValue()
        {
	    return value;
	}

        // }}}
    }

    // }}}
    // {{{ class NumberEditor

    /**
     * Editor for Number table cell.
     */
    public class NumberEditor
        extends Editor
    {
        // {{{ variables

	Class[] argTypes = new Class [] {String.class};
	java.lang.reflect.Constructor constructor;
	Object value;
        Column column;

        // }}}

        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ NumberEditor

        /**
         * Create a new NumberEditor.
         */
        public NumberEditor ()
        {
            super (new JTextField ());
            JTextField tf = (JTextField) editorComponent;
            tf.setHorizontalAlignment(tf.RIGHT);
            // This listener is necessary for Swing 1.1.1
            tf.addKeyListener(new KeyAdapter () {
                public void keyPressed(KeyEvent e) {
                    if (e.getKeyCode() == KeyEvent.VK_ENTER
                        && e.getModifiers() == 0) {
                        stopCellEditing();
                        e.consume();
                    }
                }
            });
            // Bind escape to cancel editing
            tf.registerKeyboardAction(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    cancelCellEditing();
                }
            },
            KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0),
            WHEN_FOCUSED);
        }

        // }}}

        //------------------------------------------------------------------
        // Public methods
        //------------------------------------------------------------------
        // {{{ stopCellEditing

	public boolean stopCellEditing()
        {
	    String s = (String) super.getCellEditorValue();
            boolean ok = false;
	    try {
		value = "".equals(s) ? null
                    : constructor.newInstance(new Object[]{s});
                ok = column.option.checkValue(null,  value);
	    }
	    catch (Exception e) {
            }
            if (! ok) {
		((JComponent) getComponent()).setBorder
                (BorderFactory.createCompoundBorder
                 (BorderFactory.createLineBorder(Color.red),
                  BorderFactory.createEmptyBorder(1, 1, 1, 1)));
		return false;
	    }
	    return super.stopCellEditing();
	}

        // }}}
        // {{{ getTableCellEditorComponent

        /**
         * Get the editor component for a cell.
         *
         * @param       table   The table being edited.
         * @param       value   The cell value.
         * @param       isSelected      Whether the cell is selected.
         * @param       row     The cell's row index.
         * @param       column  The cell's column index.
         *
         * @return      The editor component.
         */
	public Component getTableCellEditorComponent
            (JTable table, Object value, boolean isSelected,
             int row, int column)
        {
	    this.value = null;
            this.column = columns[column];
	    ((JComponent)getComponent()).setBorder
                (BorderFactory.createCompoundBorder
                 (BorderFactory.createLineBorder(Color.black),
                  BorderFactory.createEmptyBorder(1, 1, 1, 1)));
	    try {
		Class type = table.getColumnClass(column);
		// Since our obligation is to produce a value which is
		// assignable for the required type it is OK to use the
		// String constructor for columns which are declared
		// to contain Objects. A String is an Object.
		if (type == Object.class) {
		    type = String.class;
		}
		constructor = type.getConstructor(argTypes);
	    }
	    catch (Exception e) {
		return null;
	    }
	    return super.getTableCellEditorComponent
                (table, value, isSelected, row, column);
	}

        // }}}
        // {{{ getCellEditorValue

	public Object getCellEditorValue()
        {
	    return value;
	}

        // }}}
    }

    // }}}
}
