// {{{ copyright

/********************************************************************
 *
 * $Id: TreeOptionGroup.java,v 1.7 2001/04/24 16:06:50 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.option;

// {{{ imports

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Toolkit;
import java.awt.event.FocusEvent;
import java.awt.event.KeyEvent;

import java.util.StringTokenizer;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.Icon;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTree;
import javax.swing.KeyStroke;
import javax.swing.UIManager;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeSelectionModel;
import javax.swing.tree.TreePath;

import de.qfs.lib.config.Configuration;
import de.qfs.lib.config.ConfigWrapper;
import de.qfs.lib.config.JSplitPaneConfigWrapper;
import de.qfs.lib.gui.AdapterTreeModel;
import de.qfs.lib.gui.Grid;
import de.qfs.lib.gui.MultiPane;
import de.qfs.lib.gui.SwingUtil;
import de.qfs.lib.gui.DelayedTreeSelectionListener;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.tree.IndexBasedTreeAdapter;
import de.qfs.lib.tree.TreeAdapter;
import de.qfs.lib.tree.TreeUtil;
import de.qfs.lib.util.Reflector;

// }}}

// {{{ doc

/**
 * This OptionGroup builds a tree structure from the resources. Each node of
 * the tree can have a set of child nodes and/or an associated OptionGroup
 * that is displayed to the right of the tree. It uses the following
 * resoureces: <p>
 *
 * <TABLE BORDER="1" CELLPADDING="3" CELLSPACING="0" WIDTH="100%">
 * <TR>
 * <TH WIDTH="40%">Resource name</TH>
 * <TH>Usage</TH>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.type</code></TD>
 * <TD>A TreeOptionGroup is designated by the type <code>tree</code>.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.<em>name</em>.disabled</code></TD>
 * <TD>If true, the group will be disabled and remain so even if
 * {@link OptionGroup#setEnabled setEnabled(true)} is called (optional,
 * default false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.title</code></TD>
 * <TD>Title to display for the tree (optional, default none).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.mnemonic</code></TD>
 * <TD>Keyboard accelerator for the title (optional, default none).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.divider</code></TD>
 * <TD>Initial setting for the divider of the split pane. A double between 0
 * and 1 designates a relative position, an integer designates an absolute
 * position (optional, default none).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.configure</code></TD>
 * <TD>If true, create a {@link de.qfs.lib.config.ConfigWrapper ConfigWrapper}
 * for the split pane of the group and register it under the name
 * <code><em>groupname</em>.splitpane</code> (optional, default false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.root</code></TD>
 * <TD>The identifier of the root node of the tree (optional, default
 * "root").</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.rootvisible</code></TD>
 * <TD>Whether to display the root node of the tree (optional, default
 * false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.roothandles</code></TD>
 * <TD>Whether to show handles for the toplevel nodes of the tree (optional,
 * default (! rootvisible)).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>lines</em>.roothandles</code></TD>
 * <TD>Whether to show lines for metal look and feel (optional, default
 * false).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.<em>path</em>.name</code></TD>
 * <TD>The name to display in the tree for the node at <em>path</em>.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.<em>path</em>.icon</code></TD>
 * <TD>The icon to display in the tree for the node at <em>path</em>.</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.<em>path</em>.children</code></TD>
 * <TD>Whitespace separated list of children of the node at <em>path</em>
 * (optional, default none).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.<em>path</em>.group</code></TD>
 * <TD>The name of the OptionGroup to display for the node at <em>path</em>
 * (optional, default none).</TD>
 * </TR>
 * <TR>
 * <TD WIDTH="40%"><code>option.group.<em>groupname</em>.<em>path</em>.title</code></TD>
 * <TD>The title for the OptionGroup of the node at <em>path</em> (optional,
 * default none).</TD>
 * </TR>
 * </TABLE>
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.7 $
 * @since       0.98.1
 */

// }}}
public class TreeOptionGroup
    extends OptionGroup
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (TreeOptionGroup.class);

    /**
     * The multi pane.
     */
    private JSplitPane pane;

    /**
     * The tree component.
     */
    private JTree tree;

    /**
     * The tree's model.
     */
    private Tree model;

    /**
     * The right panel.
     */
    private JPanel panel;

    /**
     * The inner right panel.
     */
    private JPanel innerPanel;

    /**
     * The title label;
     */
    private JLabel title;

    /**
     * The scroll pane for the right panel
     */
    private JScrollPane scroll;

    /**
     * The currently selected node.
     */
    private Node current;

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ TreeOptionGroup()

    /**
     * No arg constructor for customized OptionGroups.
     */
    protected TreeOptionGroup ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "TreeOptionGroup()", "");
        }
    }

    // }}}
    // {{{ TreeOptionGroup(String, OptionSet)

    /**
     * Create a new TreeOptionGroup.
     *
     * @param   name    The name of the group.
     * @param   options The Options to edit.
     */
    public TreeOptionGroup (String name, OptionSet options)
    {
        super (name, options);
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "TreeOptionGroup(String,OptionSet)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "options: " + options);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Public methods
    //----------------------------------------------------------------------
    // {{{ getComponent

    /**
     * Get the AWT Component for the OptionGroup.
     *
     * @return  The Component.
     */
    public Component getComponent()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getComponent()", "");
        }
        if (pane == null) {
            makePane();
        }
        return pane;
    }

    // }}}
    // {{{ selectNode

    /**
     * Select a node.
     *
     * @param   path The path of the node. Null to select nothing.
     */
    public void seledtNode(String path)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "selectNode(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "path: " + path);
        }

        TreePath tp = null;
        if (path != null) {
            Node node = model.findNode(path);
            if (node != null) {
                tp = new TreePath (TreeUtil.getPath(model, node));
            }
        }
        if (tp != null) {
            tree.makeVisible(tp);
            tree.scrollPathToVisible(tp);
        }
        tree.setSelectionPath(tp);
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ makePane

    /**
     * Create the split pane component.
     */
    protected void makePane()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "makePane()", "");
        }
        panel = new JPanel ();
        panel.setLayout(new BorderLayout ());
        panel.setBorder(null);

        innerPanel = new JPanel ();
        innerPanel.setLayout(new BorderLayout ());
        innerPanel.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));

        scroll = new JScrollPane ();
        scroll.setName("OptionGroup-" + name + "-Groupscrollpane");
        scroll.setBorder(BorderFactory.createEtchedBorder());

        title = new JLabel ();
        title.setName("OptionGroup-" + name + "-Grouptitle");
        title.setForeground(UIManager.getColor("textText"));
        title.setBorder(BorderFactory.createEmptyBorder(2, 5, 2, 2));
        makeTree();

        JScrollPane sp = new JScrollPane (tree);
        title.setName("OptionGroup-" + name + "-Treescrollpane");
        SwingUtil.constrainScroll(sp);
        sp.setBorder(BorderFactory.createEtchedBorder());

        JPanel tp = new JPanel ();
        tp.setLayout(new BorderLayout ());
        tp.setBorder(null);
        tp.add(sp, BorderLayout.CENTER);

        String top = getString("title", null);
        if (top != null) {
            JLabel lt = new JLabel (top);
            lt.setName("OptionGroup-" + name + "-Title");
            lt.setForeground(UIManager.getColor("textText"));
            lt.setBorder(BorderFactory.createEmptyBorder(2, 5, 2, 2));
            tp.add(lt, BorderLayout.NORTH);
            String mnemo = getString("mnemonic", null);
            if (mnemo != null && mnemo.length() > 0) {
                lt.setDisplayedMnemonic(mnemo.charAt(0));
            }
            lt.setLabelFor(tree);
        }

        pane = new JSplitPane (JSplitPane.HORIZONTAL_SPLIT, tp, panel);
        String div = getString("divider", null);
        ConfigWrapper wrapper = null;
        if (div != null) {
            try {
                Double val = new Double (div);
                if (val.doubleValue() >= 0 && val.doubleValue() <= 1.0) {
                    wrapper = new JSplitPaneConfigWrapper
                        (pane, name + ".splitpane", val.doubleValue());
                } else if (val.intValue() >= 1) {
                    wrapper = new JSplitPaneConfigWrapper
                        (pane, name + ".splitpane", val.intValue());
                }
            } catch (NumberFormatException ex) {
            }
        }
        if (getBoolean("configure", false)) {
            if (wrapper == null) {
                wrapper = ConfigWrapper.makeWrapper
                    (pane, name + ".splitpane");
            }
            Configuration.instance().register(wrapper);
        }
        // pane = new MultiPane (MultiPane.HORIZONTAL, tp, panel);
        pane.setName("OptionGroup-" + name + "-" +
                     options.getConfigurableName());
        pane.setBorder(BorderFactory.createEtchedBorder());
    }

    // }}}
    // {{{ makeTree

    /**
     * Create the tree component.
     */
    protected void makeTree()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "makeTree()", "");
        }
        model = new Tree ();
        tree = new JTree (new AdapterTreeModel (model, model.getRoot())) {
            /**
             * Overridden to force Focus on the tree if necessary.
             */
            public void requestFocus()
            {
                super.requestFocus();
                if (! hasFocus()) {
                    Toolkit.getDefaultToolkit().getSystemEventQueue()
                        .postEvent(new FocusEvent
                                   (this, FocusEvent.FOCUS_GAINED));
                }
            }
        };
        tree.setName("OptionGroup-" + name + "-Tree");
        tree.setBorder(null);

        // Remove default escape and return bindings.
        tree.unregisterKeyboardAction
            (KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0));
        tree.unregisterKeyboardAction
            (KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0));
        // Special version for JDK 1.3
        Object inputMap = Reflector.safeCall
            (tree, "getInputMap", Integer.TYPE,
             new Integer (tree.WHEN_FOCUSED));
        if (inputMap != null) {
            Reflector.safeCall
                (inputMap, "put",
                 new Class[] {KeyStroke.class, Object.class},
                 new Object[] {
                     KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0),
                     "dummy"});
        }
        inputMap = Reflector.safeCall
            (tree, "getInputMap", Integer.TYPE,
             new Integer (tree.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT));
        if (inputMap != null) {
            Reflector.safeCall
                (inputMap, "put",
                 new Class[] {KeyStroke.class, Object.class},
                 new Object[] {
                     KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0),
                     "dummy"});
        }

        boolean rootvisible = getBoolean("rootvisible", false);
        tree.setRootVisible(rootvisible);
        tree.setShowsRootHandles(getBoolean("roothandles", ! rootvisible));
        if (getBoolean ("lines", false)
            && (tree.getUI() instanceof javax.swing.plaf.metal.MetalTreeUI)) {
            tree.putClientProperty("JTree.lineStyle", "Angled");
        }
        tree.setCellRenderer(new Renderer ());

        ((DefaultTreeSelectionModel) tree.getSelectionModel())
            .setSelectionMode
            (DefaultTreeSelectionModel.SINGLE_TREE_SELECTION);
        if (tree.getRowCount() > 0) {
            tree.setSelectionRow(0);
            nodeChanged((Node) tree.getPathForRow(0).getLastPathComponent());
        }
        tree.addTreeSelectionListener(new Listener ());
    }

    // }}}
    // {{{ nodeChanged

    /**
     * Notify the group that the selected node in the tree has changed. If
     * there is an option group displayed for the old selected node, check its
     * values and don't allow the selection to change, if the values aren't
     * legal.
     *
     * @param   node    The new node.
     */
    protected void nodeChanged(Node node)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "nodeChanged(Node)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "node: " + node);
        }
        if (current == node) {
            return;
        }

        if (current != null
            && current.group != null
            && ! current.group.checkValues()) {
            tree.setSelectionPath
                (new TreePath (TreeUtil.getPath(model, current)));
            return;
        }

        current = node;
        if (node == null || node.title == null) {
            panel.remove(title);
        } else {
            title.setText(node.title);
            panel.add(title, BorderLayout.NORTH);
        }

        if (node == null || node.group == null) {
            panel.remove(scroll);
        } else {
            Component com = node.group.getComponent();
            if ((com instanceof Grid)
                && ((Grid) com).needGlue()) {
                ((Grid) com).addGlue();
            }
            innerPanel.removeAll();
            innerPanel.add(com, BorderLayout.CENTER);
            scroll.setViewportView(innerPanel);
            panel.add(scroll, BorderLayout.CENTER);
        }

        panel.revalidate();
        panel.repaint();
    }

    // }}}

    //----------------------------------------------------------------------
    // The tree model
    //----------------------------------------------------------------------
    // {{{ class Tree

    /**
     * A straightforward tree, implemented in terms of the Node class and
     * the TreeAdapter concept.
     */
    protected class Tree
        extends IndexBasedTreeAdapter
    {
        // {{{ variables

        /**
         * The root node of the tree.
         */
        private Node root;

        // }}}

        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ Tree

        /**
         * Create a new tree.
         */
        public Tree ()
        {
            String rt = getString("root", "root");
            root = new Node (rt);
        }

        // }}}

        //------------------------------------------------------------------
        // Public methods
        //------------------------------------------------------------------
        // {{{ getRoot

        /**
         * Get the root of the tree.
         *
         * @return      The root of the tree.
         */
        public Object getRoot()
        {
            return root;
        }

        // }}}
        // {{{ findNode

        /**
         * Find a node by its path name.
         *
         * @param       path    The path of the node.
         */
        public Node findNode(String path)
        {
            if (! path.startsWith(root.path)) {
                return null;
            }
            Node node = root;
            while (true) {
                if (path.equals(node.path)) {
                    return node;
                }
                if (node.children != null) {
                    boolean found = false;
                    for (int i = 0; i < node.children.size(); i++) {
                        if (path.startsWith
                            (((Node) node.children.elementAt(i)).path)) {
                            node = (Node) node.children.elementAt(i);
                            found = true;
                            break;
                        }
                    }
                    if (! found) {
                        return null;
                    }
                } else {
                    return null;
                }
            }
        }

        // }}}

        //------------------------------------------------------------------
        // The TreeAdapter interface
        //------------------------------------------------------------------
        // {{{ getParentNode

        /**
         * Get the parent of a node.
         *
         * @param   child   The node to query for the parent.
         *
         * @return  The parent of the node.
         */
        public Object getParentNode(Object child)
        {
            return ((Node) child).parent;
        }

        // }}}
        // {{{ isLeaf

        /**
         * Query whether a node is a leaf.
         *
         * @param   node    The node to query.
         *
         * @return  True if the node is a leaf.
         */
        public boolean isLeaf(Object node)
        {
            return ((Node) node).isLeaf();
        }

        // }}}
        // {{{ getChildCount

        /**
         * Get the number of children of a node.
         *
         * @param   parent  The node to query.
         *
         * @return  The number of children of the node.
         */
        public int getChildCount(Object parent)
        {
            return ((Node) parent).children == null ? 0
                : ((Node) parent).children.size();
        }

        // }}}
        // {{{ getChildNode

        /**
         * Get a child of a node.
         *
         * @param   parent  The node to query.
         * @param   index   The index of the child.
         *
         * @return  The child at the index.
         *
         * @throws   IndexOutOfBoundsException If the index is invalid.
         */
        public Object getChildNode(Object parent, int index)
             throws IndexOutOfBoundsException
        {
            if (((Node) parent).children == null) {
                throw new IndexOutOfBoundsException
                    ("No children for : " + ((Node) parent).path);
            }
            return ((Node) parent).children.elementAt(index);
        }

        // }}}
        // {{{ getIndexOfChildNode

        /**
         * Get the index of a child in a node's list of children.
         *
         * @param   parent  The node to query.
         * @param   child   The child to look for.
         *
         * @return  The index of the child or -1 if it is not a child of this
         *          node.
         */
        public int getIndexOfChildNode(Object parent, Object child)
        {
            if (((Node) parent).children == null) {
                return -1;
            }
            return ((Node) parent).children.indexOf(child);
        }

        // }}}
    }

    // }}}

    //----------------------------------------------------------------------
    // The tree nodes
    //----------------------------------------------------------------------
    // {{{ class Node

    /**
     * These elements of the tree model are simple structures.
     */
    protected class Node
    {
        // {{{ variables

        /**
         * The parent node.
         */
        public Node parent;

        /**
         * The path name of the node.
         */
        public String path;

        /**
         * The children of the node.
         */
        public Vector children;

        /**
         * The name of the node that is displayed in the tree.
         */
        public String name;

        /**
         * The icon of the node that is displayed in the tree.
         */
        public Icon icon;

        /**
         * The title fot the group display.
         */
        public String title;

        /**
         * The OptionGroup of the Node.
         */
        public OptionGroup group;

        // }}}

        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ Node

        /**
         * Create a new Node. Fetch the parameters for the Node from the
         * resources and create the children as well.
         *
         * @param       path            The path name of the node.
         */
        public Node (String path)
        {
            this.path = path;
            String ch = getString(path + ".children", null);
            if (ch != null) {
                children = new Vector ();
                for (StringTokenizer st = new StringTokenizer (ch);
                     st.hasMoreTokens(); ) {
                    Node child = new Node (path + "." +  st.nextToken());
                    child.parent = this;
                    children.addElement(child);
                }
            }

            name = getString(path + ".name", "");

            icon = Option.getResources().getIcon
                ("option.group." + TreeOptionGroup.this.name +
                 "." + path + ".icon", null);

            title = getString(path + ".title", null);

            String gr = getString(path + ".group", null);
            if (gr != null) {
                group = OptionGroup.makeGroup(gr, options);
                if (group != null) {
                    group.setParentGroup(TreeOptionGroup.this);
                    boolean enabled = getBoolean
                        (path + ".enabled", true);
                    if (! enabled) {
                        group.setEnabled(false);
                    }
                    members.add(group);
                }
            }
        }

        // }}}

        // {{{ isLeaf

        /**
         * Query whether the node is a leaf. Currently checks whether the node
         * has children.
         *
         * @return      True if the node has no children.
         */
        public boolean isLeaf()
        {
            return children == null;
        }

        // }}}
    }

    // }}}

    //----------------------------------------------------------------------
    // The Renderer for the tree nodes
    //----------------------------------------------------------------------
    // {{{ class Renderer

    /**
     * The Renderer for the tree. Displays the Node's name and possibly its
     * icon. The Swing default icons for branches and leaves are not used.
     */
    protected class Renderer
        extends DefaultTreeCellRenderer
    {
        //------------------------------------------------------------------
        // Constructor
        //------------------------------------------------------------------
        // {{{ Renderer

        /**
         * Create a new Renderer.
         */
        public Renderer ()
        {
        }

        // }}}

        // {{{ getTreeCellRendererComponent

        /**
         * Get the Component that renders a tree node.
         *
         * @param   tree            The tree to draw.
         * @param   value           The tree node's value.
         * @param   selected        Whether the node is selected.
         * @param   expanded        Whether the node is expanded.
         * @param   leaf            Whether the node is a leaf.
         * @param   row             The node's row index.
         * @param   hasFocus        Whether the node is selected.
         *
         * @return  A Component to render the node.
         */
        public Component getTreeCellRendererComponent
        (JTree tree, Object value, boolean selected, boolean expanded,
         boolean leaf, int row, boolean hasFocus)
        {
            DefaultTreeCellRenderer com = (DefaultTreeCellRenderer)
                super.getTreeCellRendererComponent
                (tree, value, selected, expanded,
                 leaf, row, hasFocus);
            com.setText(((Node) value).name);
            com.setIcon(((Node) value).icon);
            return com;
        }

        // }}}
    }

    // }}}

    //----------------------------------------------------------------------
    // The selection listener
    //----------------------------------------------------------------------
    // {{{ class Listener

    /**
     * Delayed listener for changes in the tree selection.
     */
    protected class Listener
        extends DelayedTreeSelectionListener
    {
        // {{{ delayedValueChanged

        /**
         * Notify the listener of a change in the selection.
         *
         * @param       e       The event containing the details.
         */
        public void delayedValueChanged(TreeSelectionEvent e)
        {
            Node node = null;

            if (tree.getSelectionPath() != null) {
                node = (Node) tree.getSelectionPath().getLastPathComponent();
            }

            nodeChanged(node);
        }

        // }}}
    }

    // }}}
}
