// {{{ copyright

/********************************************************************
 *
 * $Id: CommitableNotify.java,v 1.5 2000/06/14 16:51:59 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.transaction;

// {{{ imports

import java.util.Observable;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * A CommitableNotify delays the notification of an {@link Observable
 * Observable's} {@link java.util.Observer Observer} to the time of commit.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.5 $
 * @since       0.97.0
 */
public class CommitableNotify
implements Commitable
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (CommitableNotify.class);

    /**
     * The Observable that should notify its {@link java.util.Observer
     * Observers}.
     */
    protected Observable observable;

    /**
     * The argument for the {@link java.util.Observable#notifyObservers
     * notifyObservers} call.
     */
    protected Object arg;

    // }}}

    // {{{ constructors

    /**
     * Create a new CommitableNotify.
     *
     * @param   observable      The Observable that should notify its
     *                          {@link java.util.Observer Observers}.
     */
    public CommitableNotify (Observable observable)
    {
        this (observable, null);
    }

    /**
     * Create a new CommitableNotify.
     *
     * @param   observable      The Observable that should notify its
     *                          {@link java.util.Observer Observers}.
     * @param   arg             The argument for the
     *          {@link Observable#notifyObservers notifyObservers} call.
     */
    public CommitableNotify (Observable observable, Object arg)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "CommitableNotify(Observable,Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "observable: " + observable + ", " +
                       "arg: " + arg);
        }
        this.observable = observable;
        this.arg = arg;
    }

    // }}}

    // {{{ commit

    /**
     * Call the {@link java.util.Observable#notifyObservers
     * java.util.Observable.notifyObservers} on the {@link Observable
     * Observable}.
     */
    public void commit()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "commit()", "");
        }
        observable.notifyObservers(arg);
    }

    // }}}
    // {{{ rollback

    /**
     * This is a noop, since nothing has happened yet.
     */
    public void rollback()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "rollback()", "");
        }
    }

    // }}}
}
