// {{{ copyright

/********************************************************************
 *
 * $Id: IntTests.jpp,v 1.6 2000/11/29 14:38:27 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.transaction;


// {{{ imports

import java.util.Observable;
import java.util.Observer;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.logrmi.RemoteLogWriter;

// }}}

/**
 * Integration tests for the transaction package.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.6 $
 * @since       0.97.0
 */
public class IntTests
    extends TestCase
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (IntTests.class);

    private State s1, s2;
    private StateObserver o1, o2;

    // }}}

    // {{{ initialize logging

    static {
        try {
            RemoteLogWriter.logRemote("qflog", "trtest");
            Log.setQueueing(false);
            Log.setOutputLevel(Log.ERRDETAIL);
            Logger.setDefaultLogLevel(Log.DBGDETAIL);
        } catch (Exception ex) {
        }
    }

    // }}}

    // {{{ constructor

    /**
     * Create a new IntTests.
     *
     * @param   name    The name of the test.
     */
    public IntTests(String name)
    {
        super(name);
    }

    // }}}

    // {{{ setUp

    /**
     * Set up the test fixture.
     */
    protected void setUp()
    {
        s1 = new State(1, 2);
        o1 = new StateObserver ();
        s1.addObserver(o1);

        s2 = new State(11, 12);
        o2 = new StateObserver ();
        s2.addObserver(o2);
    }

    // }}}
    // {{{ suite

    /**
     * Get the suite of tests.
     *
     * @return  The tests.
     */
    public static Test suite()
    {
            return new TestSuite(IntTests.class);
    }

    // }}}

    //----------------------------------------------------------------------
    // Tests
    //----------------------------------------------------------------------
    // {{{ testCommit

    /**
     * Test setting some States and committing.
     */
    public void testCommit()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "testCommit()", "");
        }

        Transaction.begin();
        Transaction.instance().add(new CommitableStateChange (s1));
        s1.set(2, 3);
        assert(s1.x == 2);
        assert(s1.y == 3);
        assert(o1.x == 0);
        assert(o1.y == 0);
        Transaction.instance().add(new CommitableStateChange (s2));
        s2.set(12, 13);
        assert(s2.x == 12);
        assert(s2.y == 13);
        assert(o2.x == 0);
        assert(o2.y == 0);
        Transaction.instance().commit();
        assert(s1.x == 2);
        assert(s1.y == 3);
        assert(o1.x == 2);
        assert(o1.y == 3);
        assert(s2.x == 12);
        assert(s2.y == 13);
        assert(o2.x == 12);
        assert(o2.y == 13);
    }

    // }}}
    // {{{ testRollback

    /**
     * Test setting some States and rolling back.
     */
    public void testRollback()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "testRollback()", "");
        }

        Transaction.begin();
        Transaction.instance().add(new CommitableStateChange (s1));
        s1.set(2, 3);
        assert(s1.x == 2);
        assert(s1.y == 3);
        assert(o1.x == 0);
        assert(o1.y == 0);
        Transaction.instance().add(new CommitableStateChange (s2));
        s2.set(12, 13);
        assert(s2.x == 12);
        assert(s2.y == 13);
        assert(o2.x == 0);
        assert(o2.y == 0);
        Transaction.instance().rollback();
        assert(s1.x == 1);
        assert(s1.y == 2);
        assert(o1.x == 0);
        assert(o1.y == 0);
        assert(s2.x == 11);
        assert(s2.y == 12);
        assert(o2.x == 0);
        assert(o2.y == 0);
    }

    // }}}
    // {{{ testExceptions

    /**
     * Test exceptions during rollback.
     */
    public void testExceptions()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "testExceptions()", "");
        }

        s1 = new State (1, 4); // will throw RuntimeException
        Transaction.begin();
        Transaction.instance().add(new CommitableStateChange (s1));
        s1.set(2, 3);
        assert(s1.x == 2);
        assert(s1.y == 3);
        assert(o1.x == 0);
        assert(o1.y == 0);
        Transaction.instance().add(new CommitableStateChange (s2));
        s2.set(12, 13);
        assert(s2.x == 12);
        assert(s2.y == 13);
        assert(o2.x == 0);
        assert(o2.y == 0);
        try {
            // avoid log messages for provoked exceptions
            Logger.setLogLevel("de.qfs.lib.transaction.Transaction", 0);
            Transaction.instance().rollback();
            assert(false);
        } catch (RuntimeException ex) {
        } finally {
            Logger.removeLogLevel("de.qfs.lib.transaction.Transaction");
        }
        assert(s1.x == 1);
        assert(s1.y == 4);
        assert(o1.x == 0);
        assert(o1.y == 0);
        assert(s2.x == 11);
        assert(s2.y == 12);
        assert(o2.x == 0);
        assert(o2.y == 0);

        s1 = new State (1, 2);
        Transaction.begin();
        Transaction.instance().add(new CommitableStateChange (s1));
        s1.set(2, 3);
        assert(s1.x == 2);
        assert(s1.y == 3);
        assert(o1.x == 0);
        assert(o1.y == 0);

        s2 = new State (11, 5); // will throw FatalTransactionException
        Transaction.instance().add(new CommitableStateChange (s2));
        s2.set(12, 13);
        assert(s2.x == 12);
        assert(s2.y == 13);
        assert(o2.x == 0);
        assert(o2.y == 0);
        try {
            // avoid log messages for provoked exceptions
            Logger.setLogLevel("de.qfs.lib.transaction.Transaction", 0);
            Transaction.instance().rollback();
            assert(false);
        } catch (FatalTransactionException ex) {
        } finally {
            Logger.removeLogLevel("de.qfs.lib.transaction.Transaction");
        }
        assert(s1.x == 2);
        assert(s1.y == 3);
        assert(o1.x == 0);
        assert(o1.y == 0);
        assert(s2.x == 11);
        assert(s2.y == 5);
        assert(o2.x == 0);
        assert(o2.y == 0);
    }

    // }}}
    // {{{ testThreads

    /**
     * Test creating Transactions from different Threads.
     */
    public void testThreads()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "testThreads()", "");
        }
        Transaction.setMaxTransactions(2);

        Request r1 = new Request("r1", 3000);
        Request r2 = new Request("r2", 1000);
        Request r3 = new Request("r3", 20);

        r1.start();
        try {
            Thread.currentThread().sleep(10);
        } catch (InterruptedException ex) {
        }

        r2.start();
        try {
            Thread.currentThread().sleep(10);
        } catch (InterruptedException ex) {
        }
        r3.start();

        try {
            r1.join();
            r2.join();
            r3.join();
        } catch (InterruptedException ex) {
        }

        assert(r1.waited < 500);
        assert(r2.waited < 500);
        assert(r3.waited > 500 && r3.waited < 1500);

        Transaction.setMaxTransactions(0);
    }

    // }}}

    //----------------------------------------------------------------------
    // Helpers
    //----------------------------------------------------------------------
    // {{{ class State

    /**
     * Used to test CommitableStateChange and CommitableNotify.
     */
    private static class State
        extends Observable
        implements StateHolder
    {
        // {{{ variables

        /**
         * The Logger used for logging State methods.
         */
        private final static Logger logger = new Logger (State.class);

        /**
         * Internal state x.
         */
        public int x;

        /**
         * Internal state y.
         */
        public int y;

        // }}}

        // {{{ constructor

        public State (int x, int y)
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "State(int,int)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "x: " + x + ", " +
                           "y: " + y);
            }
            this.x = x;
            this.y = y;
        }

        // }}}

        // {{{ saveState

        public Object saveState()
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "saveState()", "");
            }
            return new int[] {x, y};
        }


        // }}}
        // {{{ commitState

        public void commitState(Object info)
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "commitState(Object)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "info: " + ((int[]) info)[0] + "," +
                           ((int[]) info)[1]);
            }
        }

        // }}}
        // {{{ restoreState

        public void restoreState(Object info)
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "restoreState(Object)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "info: " + ((int[]) info)[0] + "," +
                           ((int[]) info)[1]);
            }
            x = ((int[]) info)[0];
            y = ((int[]) info)[1];
            clearChanged();
            // special case to test exceptions
            if (y == 4) {
                throw new RuntimeException ();
            } else if (y == 5) {
                throw new FatalTransactionException ("test");
            }
        }

        // }}}

        // {{{ notifyObservers

        public void notifyObservers(Object arg)
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "notifyObservers(Object)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "arg: " + arg);
            }
            super.notifyObservers(arg);
        }

        // }}}

        // {{{ set

        public void set(int x, int y)
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "set(int,int)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "x: " + x + ", " +
                           "y: " + y);
            }
            this.x = x;
            this.y = y;

            setChanged();
            if (Transaction.insideTransaction()) {
                Transaction.instance().add(new CommitableNotify (this));
            }
        }

        // }}}
    }

    // }}}
    // {{{ class StateObserver

    /**
     * Observer for State objects.
     */
    private static class StateObserver
        implements Observer
    {
        // {{{ variables

        /**
         * The Logger used for logging StateObserver methods.
         */
        private final static Logger logger = new Logger (StateObserver.class);

        /**
         * The x value of the observed State.
         */
        public int x;

        /**
         * The y value of the observed State.
         */
        public int y;

        // }}}

        // {{{ constructor

        public StateObserver ()
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "StateObserver()", "");
            }
        }

        // }}}
        // {{{ update

        public void update(Observable observable, Object arg)
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "update(Observable,Object)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "observable: " + observable + ", " +
                           "arg: " + arg);
            }
            x = ((State) observable).x;
            y = ((State) observable).y;
        }

        // }}}
    }

    // }}}
    // {{{ class Request

    /**
     * Used to test limited numbers of Transactions.
     */
    private static class Request
        extends Thread
    {
        // {{{ variables

        /**
         * The Logger used for logging.
         */
        private final static Logger logger = new Logger (IntTests.class);

        /**
         * How long to keep a Transaction open.
         */
        private int delay;

        /**
         * How long the Thread waited for a Transaction.
         */
        public long waited;

        // }}}

        // {{{ constructor

        public Request(String name, int delay)
        {
            super(name);
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "Request(String,int)",
                           logger.level < Log.MTDDETAIL ? "" :
                           "name: " + name + ", " +
                           "delay: " + delay);
            }
            this.delay = delay;
        }


        // }}}

        // {{{ run

        public void run()
        {
            if (logger.level >= Log.MTD) {
                logger.log(Log.MTD, "run()", "");
            }
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "run()",
                           "requesting Transaction");
            }
            waited = System.currentTimeMillis();
            Transaction.begin();
            waited = System.currentTimeMillis() - waited;
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "run()",
                           "got Transaction, waited " + waited + " ms");
            }
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "run()",
                           "delaying " + delay + "ms");
            }
            try {
                sleep(delay);
            } catch (InterruptedException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log("run()", ex);
                }
            }
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "run()",
                           "releasing Transaction");
            }
            Transaction.instance().commit();
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "run()",
                           "done");
            }

        }

        // }}}
    }

    // }}}
}

