// {{{ copyright

/********************************************************************
 *
 * $Id: IndexTraversal.java,v 1.3 2001/04/21 19:55:03 gs Exp $
 *
 * Copyright (C) 2000 Gregor Schmid, Quality First Software
 * All rights reserved
 *
 *******************************************************************/

// }}}

package de.qfs.lib.tree;

// {{{ imports

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * Index based tree traversal.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.3 $
 * @since       0.98.1
 */
public class IndexTraversal
    extends AbstractTraversal
    implements MutableTraversal
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (IndexTraversal.class);

    /**
     * Whether to abort the traversal.
     */
    boolean finished;

    /**
     * Whether the current node was removed.
     */
    boolean removed;

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ IndexTraversal

    /**
     * Create a new IndexTraversal.
     */
    public IndexTraversal (TreeAdapter adapter, Object root)
    {
        super(adapter, root);
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "IndexTraversal(TreeAdapter,Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "adapter: " + adapter + ", " +
                       "root: " + root);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Traversal methods
    //----------------------------------------------------------------------
    // {{{ traverse

    /**
     * Traverse the tree.
     *
     * @param   callback The callback to call back to for each node
     *          encountered.
     *
     * @throws  ReturnFromTraversalException  If the callback method does.
     */
    public void traverse(TraversalCallback callback)
        throws ReturnFromTraversalException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "traverse(Callback)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "callback: " + callback);
        }

        removed = false;
        boolean ret = callback.nodeEntered(this);
        if (finished || removed) {
            removed = false;
            return;
        }
        if (!ret) {
            callback.nodeExited(this);
            return;
        }

        if (! adapter.isLeaf(getNode())
            && adapter.getChildCount(getNode()) > 0) {
            Level parent = currentLevel;
            levels.push(parent);
            Object previous = null;
            currentLevel = new Level (null);
            for (currentLevel.index = 0;
                 currentLevel.index < adapter.getChildCount(parent.node);
                 currentLevel.index++) {
                currentLevel.previous = currentLevel.node;
                currentLevel.node = adapter.getChildNode
                    (parent.node, currentLevel.index);
                traverse(callback);
                if (finished) {
                    return;
                }
            }
            currentLevel = (Level) levels.pop();
        }

        callback.nodeExited(this);
        removed = false;
    }

    // }}}
    // {{{ traverseFrom

    /**
     * Traverse the tree, skipping the callback for some nodes.
     *
     * @param   callback        The callback for the traversal.
     * @param   path            The path of the first node to call back to the
     *                          callback for.
     *
     * @throws  ReturnFromTraversalException
     */
    public void traverseFrom(TraversalCallback callback, Object[] path)
        throws ReturnFromTraversalException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "traverseFrom(Callback,Object[])",
                       logger.level < Log.MTDDETAIL ? "" :
                       "callback: " + callback + ", " +
                       "path: " + path);
        }

        if (path == null || path[0] == getNode()) {
            boolean mustExit = false;
            if (path.length == 1) {
                boolean ret = callback.nodeEntered(this);
                if (finished) {
                    return;
                }
                if (!ret) {
                    callback.nodeExited(this);
                    return;
                }
                mustExit = true;
            }

            boolean skip = true;
            if (path != null && path.length > 1) {
                Object[] newPath = new Object[path.length - 1];
                System.arraycopy(path, 1, newPath, 0,
                                 newPath.length);
                path = newPath;
            } else {
                skip = false;
                path = null;
            }

            if (! adapter.isLeaf(getNode())
                && adapter.getChildCount(getNode()) > 0) {
                Level parent = currentLevel;
                levels.push(parent);
                Object previous = null;
                for (int i = 0; i < adapter.getChildCount(parent.node); i++) {
                    currentLevel = new Level
                        (adapter.getChildNode(parent.node, i), previous, i);
                    previous = currentLevel.node;
                    if (skip) {
                        traverseFrom(callback, path);
                    } else {
                        traverse(callback);
                    }
                    if (finished) {
                        return;
                    }
                    if (path != null && path[0] == currentLevel.node) {
                        skip = false;
                    }
                }
            }
            if (mustExit) {
                callback.nodeExited(this);
            }
        }
    }

    // }}}

    //------------------------------------------------------------------
    // The Traversal interface
    //------------------------------------------------------------------
    // {{{ finish

    /**
     * Cleanly abort the traversal.
     */
    public void finish()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "finish()", "");
        }
        finished = true;
    }

    // }}}

    //------------------------------------------------------------------
    // The MutableTraversal interface
    //------------------------------------------------------------------
    // {{{ removeNode

    /**
     * Remove the currently visited node from the tree.
     *
     * @return  True if the node could be removed, false otherwise.
     */
    public boolean removeNode()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "removeNode()", "");
        }
        if (getParent() != null
            && (adapter instanceof MutableTreeAdapter)) {
            Object node =
                ((MutableTreeAdapter) adapter).removeNode
                (getParent(), currentLevel.index);
            if (node != null) {
                nodeRemoved();
                return true;
            }
        }
        return false;
    }

    // }}}
    // {{{ nodeRemoved

    /**
     * Notify the traversal that the currently visited node has been removed.
     */
    public void nodeRemoved()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "nodeRemoved()", "");
        }
        removed = true;
        currentLevel.node = currentLevel.previous;
        currentLevel.index--;
    }

    // }}}
    // {{{ insertNodeBefore

    /**
     * Insert a node before the current node of the traversal.
     *
     * @param   node    The node to insert.
     *
     * @return  True if the node could be inserted, false if not.
     */
    public boolean insertNodeBefore(Object node)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "insertNodeBefore(Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "node: " + node);
        }
        if (getParent() != null
            && (adapter instanceof MutableTreeAdapter)) {
            boolean ret =
                ((MutableTreeAdapter) adapter).insertNodeBefore
                (getParent(), node, getNode());
            if (ret) {
                nodeInsertedBefore(node);
                return true;
            }
        }
        return false;
    }

    // }}}
    // {{{ nodeInsertedBefore

    /**
     * Notify the traversal that a node has been inserted before the currently
     * visited node.
     *
     * @param node The inserted node which will become the node returned by
     * {@link #getPreviousSibling getPreviousSibling}.
     */
    public void nodeInsertedBefore(Object node)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "nodeInsertedBefore(Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "node: " + node);
        }
        currentLevel.previous = node;
        currentLevel.index++;
    }

    // }}}
    // {{{ insertNodeAfter

    /**
     * Insert a node after the current node of the traversal.
     *
     * @param   node    The node to insert.
     *
     * @return  True if the node could be inserted, false if not.
     */
    public boolean insertNodeAfter(Object node)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "insertNodeAfter(Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "node: " + node);
        }
        if (getParent() != null
            && (adapter instanceof MutableTreeAdapter)) {
            boolean ret =
                ((MutableTreeAdapter) adapter).insertNodeAfter
                (getParent(), node, getNode());
            if (ret) {
                nodeInsertedAfter(node);
                return true;
            }
        }
        return false;
    }

    // }}}
    // {{{ nodeInsertedAfter

    /**
     * Notify the traversal that a node has been inserted after the currently
     * visited node.
     *
     * @param node The inserted node.
     */
    public void nodeInsertedAfter(Object node)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "nodeInsertedAfter()", "");
        }
    }

    // }}}
}

