// {{{ copyright

/********************************************************************
 *
 * $Id: MutableTreeAdapter.java,v 1.2 2001/04/06 15:03:52 gs Exp $
 *
 * Copyright (C) 2000 Gregor Schmid, Quality First Software
 * All rights reserved
 *
 *******************************************************************/

// }}}

package de.qfs.lib.tree;

/**
 * This extension of the TreeAdapter interface has to be implemented in order
 * to support mutable tree hierarchies.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.2 $
 * @since       0.98.1
 */
public interface MutableTreeAdapter
    extends TreeAdapter
{
    // {{{ setParentNode

    /**
     * Set the parent of a node.
     *
     * @param   child   The node whose parent to set.
     * @param   parent  The parent to set.
     */
    public void setParentNode(Object child, Object parent);

    // }}}
    // {{{ addNode(Object,Object)

    /**
     * Add a child to the end of a node's children.
     *
     * @param   parent  The node to add the child to.
     * @param   child   The child to add.
     *
     * @return  True if the child could be added, false if not
     */
    public boolean addNode(Object parent, Object child);

    // }}}
    // {{{ removeNode(Object,Object)

    /**
     * Remove a child from the node.
     *
     * @param   parent  The node to remove the child from.
     * @param   child   The child to remove.
     *
     * @return  True if the child was removed, false if the
     *          removal failed or the child is not a child of this node.
     */
    public boolean removeNode(Object parent, Object child);

    // }}}

    //----------------------------------------------------------------------
    // Index based
    //----------------------------------------------------------------------
    // {{{ addNode(Object,int,Object)

    /**
     * Add a child at a given index to a node.
     *
     * @param   parent  The node to add the child to.
     * @param   index   The index at which to insert.
     * @param   child   The child to insert.
     *
     * @return  True if the child could be added, false if not
     */
    public boolean addNode(Object parent, int index, Object child);

    // }}}
    // {{{ removeNode(Object,int)

    /**
     * Remove a child from a node at a given index.
     *
     * @param   parent  The node to remove the child from.
     * @param   index   The index from which to remove.
     *
     * @return  The node that was removed or null if the removal failed.
     *
     * @throws   IndexOutOfBoundsException If the index is invalid.
     */
    public Object removeNode(Object parent, int index)
         throws IndexOutOfBoundsException;

    // }}}

    //----------------------------------------------------------------------
    // Link based
    //----------------------------------------------------------------------
    // {{{ insertNodeBefore(Object,Object,Object)

    /**
     * Insert a child node before another node. If the reference node is null,
     * add the child as the last child.
     *
     * @param   parent  The node to add the child to.
     * @param   child   The child to insert.
     * @param   ref     The reference node to insert the child before.
     *
     * @return  True if the child could be added, false if not
     */
    public boolean insertNodeBefore(Object parent, Object child, Object ref);

    // }}}
    // {{{ insertNodeAfter(Object,Object,Object)

    /**
     * Insert a child node after another node. If the reference node is null,
     * add the child as the first child.
     *
     * @param   parent  The node to add the child to.
     * @param   child   The child to insert.
     * @param   ref     The reference node to insert the child after.
     *
     * @return  True if the child could be added, false if not
     */
    public boolean insertNodeAfter(Object parent, Object child, Object ref);

    // }}}

    // {{{ cloneNode

    /**
     * Create a shallow copy of a node, similar to {@link
     * java.lang.Object#clone Object.clone}, but with the parent of the copy
     * set to null. <p>
     *
     * Also the copy of the node should not have any children.
     *
     * @param   node    The node to copy.
     *
     * @return  A copy of the node.
     */
    public Object cloneNode(Object node);

    // }}}
}
