// {{{ copyright

/********************************************************************
 *
 * $Id: TreeBuilder.java,v 1.1 2001/04/06 15:03:42 gs Exp $
 *
 * Copyright (C) 2000 Gregor Schmid, Quality First Software
 * All rights reserved
 *
 *******************************************************************/

// }}}

package de.qfs.lib.tree;

// {{{ imports

import java.util.Vector;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * Helper class that simplifies building a tree from some kind of data, e.g. a
 * file.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.1 $
 * @since       0.98.1
 */
public class TreeBuilder
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (TreeBuilder.class);

    /**
     * Dummy object to represent null nodes internally.
     */
    protected final static Object dummy = new Object ();

    /**
     * Flag to check for the root node.
     */
    protected boolean first = true;

    /**
     * The adapter used for tree manipulation.
     */
    protected MutableTreeAdapter adapter;

    /**
     * The node factory used to create the nodes.
     */
    protected NodeFactory factory;

    /**
     * The root object of the new tree.
     */
    protected Object root;

    /**
     * The current node during the build process-
     */
    protected Object currentNode;

    /**
     * The path to the current node.
     */
    protected Vector path = new Vector ();

    // }}}

    // {{{ constructor

    /**
     * Create a new TreeBuilder.
     *
     * @param   adapter The adapter to use for tree manipulation.
     * @param   factory The node factory to use to create the nodes.
     */
    public TreeBuilder (MutableTreeAdapter adapter, NodeFactory factory)
    {
        this.adapter = adapter;
        this.factory = factory;
    }

    // }}}

    // {{{ openNode

    /**
     * Open a new node.
     *
     * @param   data   The data for the new node.
     *
     * @return  The new node.
     */
    public Object openNode(Object data)
    {
        Object node = factory.createNode(data);
        if (first) {
            root = node;
            first = false;
        } else {
            if (currentNode != null) {
                if (node != null) {
                    adapter.addNode(currentNode, node);
                    adapter.setParentNode(node, currentNode);
                }
                path.addElement(currentNode);
            } else {
                path.addElement(dummy);
            }
        }
        currentNode = node;
        return node;
    }

    // }}}
    // {{{ closeNode

    /**
     * Close the current node.
     *
     * @return  The new current node, i.e. the parent of the closed node.
     */
    public Object closeNode()
    {
        if (currentNode != null) {
            factory.finishNode(currentNode);
        }
        int size = path.size();
        if (size > 0) {
            currentNode = path.elementAt(size -1);
            path.removeElementAt(size - 1);
            if (currentNode == dummy) {
                currentNode = null;
            }
        } else {
            currentNode = null;
        }
        return currentNode;
    }

    // }}}
    // {{{ getRoot

    /**
     * Get the root node of the new tree.
     *
     * @return  The root of the tree.
     */
    public Object getRoot()
    {
        return root;
    }

    // }}}
    // {{{ reset

    /**
     * Reset the TreeBuilder to build a new tree.
     */
    public void reset()
    {
        root = null;
        first = true;
    }

    // }}}
}


