// {{{ copyright

/********************************************************************
 *
 * $Id: ArchiveClassLoader.java,v 1.5 2000/06/06 17:00:04 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.IOException;

import java.util.Hashtable;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * An ArchiveClassLoader loads classes from a single source, a ZipFile or a
 * directory. Referenced classes will be loaded from the source, if possible,
 * or loaded via the system ClassLoader otherwise.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.5 $
 */
public class ArchiveClassLoader extends ClassLoader
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (ArchiveClassLoader.class);

    /**
     * Cache for loaded classes.
     */
    protected Hashtable classes = new Hashtable ();

    /**
     * The ZipFile holding the classes.
     */
    protected ZipFile zip;

    /**
     * The directory holding the classes.
     */
    protected File dir;

    // }}}

    // {{{ constructors

    /**
     * Create a new ArchiveClassLoader that loads classes from a ZipFile.
     *
     * @param   zip     The ZipFile to read the classes from.
     */
    public ArchiveClassLoader (ZipFile zip)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "ArchiveClassLoader(ZipFile)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "zip: " + zip);
        }
        this.zip = zip;
    }

    /**
     * Create a new ArchiveClassLoader that loads classes from a directory.
     *
     * @param   dir     The directory to read the classes from.
     */
    public ArchiveClassLoader (File dir)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "ArchiveClassLoader(File)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "dir: " + dir);
        }
        this.dir = dir;
    }

    // }}}

    // {{{ loadClass

    /**
     * Load and resolve a class.
     *
     * @param   name    The class' name.
     *
     * @return  The loaded class.
     *
     * @throws  ClassNotFoundException  If no such class exists or some
     *          IO problem occurs.
     */
    public Class loadClass(String name)
         throws ClassNotFoundException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "loadClass(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }
        return loadClass(name, true);
    }

    // }}}
    // {{{ getResourceAsStream

    /**
     * Find a resource in the classpath and open a stream for it.
     *
     * @param   name    The name of the resource.
     *
     * @return  The stream.
     */
    public InputStream getResourceAsStream(String name)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getResourceAsStream(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name);
        }
        if (dir != null) {
            String sysName = name;
            if (!(File.separatorChar == '/')) {
                sysName =
                    name.replace('/', File.separator
                                 .charAt(0));
            }
            try {
                File f = new File
                    (((File) dir).getCanonicalPath() +
                     File.separator +
                     sysName);
                if (! f.exists()) {
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG,
                                   "getResourceAsStream(String)",
                                   f.toString() + "Not Found");
                    }
                    return null;
                }
                return new FileInputStream (f);
            } catch (IOException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log("getResourceAsStream(String)", ex);
                }
                return null;
            }
        } else if (zip != null) {
            try {
                ZipEntry ze = ((ZipFile) zip).getEntry(name);
                if (ze == null) {
                    if (logger.level >= Log.DBG) {
                        logger.log(Log.DBG, "getResourceAsStream(String)",
                                   "Not Found");
                    }
                    return null;
                }
                return ((ZipFile) zip).getInputStream(ze);
            } catch (ZipException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log("getResourceAsStream(String)", ex);
                }
                return null;
            } catch (IOException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log("getResourceAsStream(String)", ex);
                }
                return null;
            }
        } else {
            return null;
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ loadClass

    /**
     * Load a class and maybe resolve it.
     *
     * @param   name    The class' name.
     * @param   resolve Whether to resolve the class.
     *
     * @return  The loaded class.
     *
     * @throws  ClassNotFoundException  If no such class exists or some
     *          IO problem occurs.
     */
    protected synchronized Class loadClass(String name, boolean resolve)
         throws ClassNotFoundException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "loadClass(String,boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "resolve: " + resolve);
        }

        // First check the cache
        Class ret = (Class) classes.get(name);
        if (ret != null) {
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "loadClass(String,boolean)",
                           "Cache hit for " + name);
            }
            return ret;
        }

        String resName = name.replace('.', '/') + ".class";
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "loadClass(String,boolean)",
                       "Converted name is " + resName);
        }

        InputStream is = getResourceAsStream(resName);
        if (is == null) {
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "loadClass(String,boolean)",
                           "Null resource for " + resName);
            }
            return makeSystemClass(name, resolve);
        }

        byte[] buf;
        try {
            buf = StreamInput.read(is);
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "loadClass(String,boolean)",
                           "class " + name + " has " + buf.length + " bytes" );
            }
        } catch (IOException ex) {
            if (logger.level >= Log.ERR) {
                logger.log("loadClass(String,boolean)", ex);
            }
            return makeSystemClass(name, resolve);
        }

        Class clazz = null;
        try {
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "loadClass(String,boolean)",
                           "defining " + name);
            }
            clazz = defineClass(name, buf, 0, buf.length);
        } catch (Throwable ex) {
            if (logger.level >= Log.ERR) {
                logger.log("loadClass(String,boolean) " + name, ex);
            }
            return makeSystemClass(name, resolve);
        }
        classes.put (name, clazz);
        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "loadClass(String,boolean)",
                       name + " defined");
        }
        if (resolve) {
            resolveClass(clazz);
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "loadClass(String,boolean)",
                           name + " resolved");
            }
        }
        return clazz;
    }

    // }}}
    // {{{ makeSystemClass

    /**
     * Get a system class and maybe resolve it.
     *
     * @param   name    The class' name.
     * @param   resolve Whether to resolve the class.
     *
     * @return  The system class found.
     *
     * @throws  ClassNotFoundException  If no such class exists or some
     *          IO problem occurs.
     */
    protected Class makeSystemClass (String name, boolean resolve)
         throws ClassNotFoundException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "makeSystemClass(String,boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "name: " + name + ", " +
                       "resolve: " + resolve);
        }
        Class clazz = findSystemClass(name);
        if (clazz == null) {
            if (logger.level >= Log.ERR) {
                logger.log(Log.ERR, "makeSystemClass(String,boolean)",
                           "Cannot fing system class " + name);
            }
            throw new ClassNotFoundException (name);
        }
        if (resolve) {
            resolveClass(clazz);
        }
        return clazz;
    }

    // }}}
}
