// {{{ copyright

/********************************************************************
 *
 * $Id: ExtensionFileFilter.java,v 1.2 2000/10/31 13:43:54 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports

import java.io.File;
import java.io.FilenameFilter;

import javax.swing.filechooser.FileFilter;

// }}}

/**
 * Generic extension based file filter for AWT and Swing file chooser dialogs
 * as well as for {@link java.io.File#list File.list}. <p>
 *
 * An ExtensionFileFilter will accept all files whose name matches one of a
 * fixed set of extenstions. Optionally directories can be accepted without
 * checking their name.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.2 $
 * @since       0.98.0
 */
public class ExtensionFileFilter
    extends FileFilter
    implements FilenameFilter
{
    // {{{ variables

    /**
     * The description for the ExtensionFileFilter.
     */
    private String desc;

    /**
     * The extensions to accept.
     */
    private String[] ext;

    /**
     * Whether directories are always accepted.
     */
    private boolean acceptDirectories;

    /**
     * Whether case is ignore when checking the file's extension.
     */
    private boolean ignoreCase;

    // }}}

    //----------------------------------------------------------------------
    // Constructors
    //----------------------------------------------------------------------
    // {{{ ExtensionFileFilter(String,String,boolean,boolean)

    /**
     * Create a new ExtensionFileFilter.
     *
     * @param   desc    The description for the ExtensionFileFilter.
     * @param   ext     The extension to accept (including the '.').
     * @param   acceptDirectories Whether directories are always accepted.
     * @param   ignoreCase Whether case is ignore when checking the file's
     *                  extension.
     */
    public ExtensionFileFilter (String desc, String ext,
                                boolean acceptDirectories, boolean ignoreCase)
    {
        this.ext = new String[] {ext};
        this.desc = desc;
        this.acceptDirectories = acceptDirectories;
        this.ignoreCase = this.ignoreCase;
    }

    // }}}
    // {{{ ExtensionFileFilter(String,String[],boolean,boolean)

    /**
     * Create a new ExtensionFileFilter.
     *
     * @param   desc    The description for the ExtensionFileFilter.
     * @param   ext     The extensions to accept (including the '.').
     * @param   acceptDirectories Whether directories are always accepted.
     * @param   ignoreCase Whether case is ignore when checking the file's
     *                  extension.
     */
    public ExtensionFileFilter (String desc, String[] multi,
                                boolean acceptDirectories, boolean ignoreCase)
    {
        this.ext = (String[]) multi.clone();
        this.desc = desc;
        this.acceptDirectories = acceptDirectories;
        this.ignoreCase = this.ignoreCase;
    }

    // }}}

    // {{{ accept(File)

    /**
     * Accept or reject a file.
     *
     * @param   file    The file to test.
     *
     * @return  True if the file's extension matches one of the
     *          ExtensionFileFilter's extensions or if it is a directory
     *          and acceptDirectories is true.
     */
    public boolean accept(File file)
    {
        if (acceptDirectories && file.isDirectory()) {
            return true;
        }
        return checkExt(file.getName());
    }

    // }}}
    // {{{ accept(File,String)

    /**
     * Accept or reject a file.
     *
     * @param   directory The drirectory of the file to test.
     * @param   name      The name of the file to test.
     *
     * @return  True if the file's extension matches one of the
     *          ExtensionFileFilter's extensions or if it is a directory
     *          and acceptDirectories is true.
     */
    public boolean accept(File directory, String name)
    {
        if (checkExt(name)) {
            return true;
        }
        return acceptDirectories &&
            new File(directory, name).isDirectory();
    }

    // }}}

    // {{{ getDescription

    /**
     * Get a description for the ExtensionFileFilter.
     *
     * @return  The ExtensionFileFilter's description.
     */
    public String getDescription()
    {
        return desc;
    }

    // }}}
    // {{{ forceExtension

    /**
     * Check whether a filename has a proper extension for this
     * ExtensionFileFilter. If not, create a new File with the first of the
     * valid extenstions added and return it. This is useful when a user
     * types a filename without an extension into a save file dialog.
     *
     * @param   file    The file to correct.
     *
     * @return  The correct file.
     */
    public File forceExtension(File file)
    {
        String name = file.getName();
        if (ignoreCase) {
            if (name.length() >= ext[0].length() &&
                name.substring(name.length() - ext[0].length())
                .equalsIgnoreCase(ext[0])) {
                return file;
            }
        } else if (name.endsWith(ext[0])) {
            return file;
        }
        return new File(file.getPath() + ext[0]);
    }

    // }}}

    // {{{ checkExt

    /**
     * Compare a filename against the list of extensions, ignoring case if
     * necessary.
     *
     * @param   name    The name to check.
     *
     * @return  True if the name matches one of the extenstions.
     */
    private boolean checkExt(String name)
    {
        for (int i = 0; i < ext.length; i++) {
            if (ignoreCase) {
                if (name.length() >= ext[i].length() &&
                    name.substring(name.length() - ext[i].length())
                    .equalsIgnoreCase(ext[i])) {
                    return true;
                }
            } else if (name.endsWith(ext[i])) {
                return true;
            }
        }
        return false;
    }

    // }}}
}

