// {{{ copyright

/********************************************************************
 *
 * $Id: InputStreamAdapter.java,v 1.2 2000/06/06 16:41:29 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports

import java.io.InputStream;
import java.io.IOException;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * This is a baseclass for InputStream adapters. It delegates all calls to
 * some other InputStream. Derived classes can overload the methods to
 * modify the InputStream's behaviour.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.2 $
 * @since       0.96.0
 */
public class InputStreamAdapter
    extends InputStream
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (InputStreamAdapter.class);

    /**
     * The InputStream used for delegation.
     */
    protected InputStream stream;

    // }}}

    // {{{ constructor

    /**
     * Create a new InputStreamAdapter.
     *
     * @param   stream  The InputStream used for delegation.
     */
    public InputStreamAdapter(InputStream stream)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "InputStreamAdapter(InputStream)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "stream: " + stream);
        }
        this.stream = stream;
    }

    // }}}

    // {{{ available

    /**
     * Returns the number of bytes that can be read (or skipped over) from
     * this input stream without blocking by the next caller of a method for
     * this input stream. The next caller might be the same thread or or
     * another thread.
     *
     * @return  The number of bytes that can be read from this input stream
     *          without blocking.
     * @throws  IOException     If an I/O error occurs.
     */
    public int available()
        throws IOException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "available()", "");
        }
        return stream.available();
    }

    // }}}
    // {{{ close

    /**
     * Closes this input stream and releases any system resources associated
     * with the stream.
     *
     * @throws  IOException     If an I/O error occurs.
     */
    public void close()
        throws IOException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "close()", "");
        }
        stream.close();
    }

    // }}}
    // {{{ mark

    /**
     * Marks the current position in this input stream. A subsequent call to
     * the reset method repositions this stream at the last marked position so
     * that subsequent reads re-read the same bytes. <p>
     *
     * The readlimit arguments tells this input stream to allow that many
     * bytes to be read before the mark position gets invalidated. <p>
     *
     * The general contract of mark is that, if the method markSupported
     * returns true, the stream somehow remembers all the bytes read after the
     * call to mark and stands ready to supply those same bytes again if and
     * whenever the method reset is called.  However, the stream is not
     * required to remember any data at all if more than readlimit bytes are
     * read from the stream before reset is called.  Marks the current
     * position in this input stream.
     *
     * @param   readlimit       The maximum limit of bytes that can be read
     *          before the mark position becomes invalid.
     */
    public void mark(int readlimit)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "mark(int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "readlimit: " + readlimit);
        }
        stream.mark(readlimit);
    }

    // }}}
    // {{{ markSupported

    /**
     * Tests if this input stream supports the mark and reset methods.
     *
     * @return  True if this true type supports the mark and reset method,
     *          false otherwise.
     */
    public boolean markSupported()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "markSupported()", "");
        }
        return stream.markSupported();
    }

    // }}}
    // {{{ read

    /**
     * Reads the next byte of data from the input stream. The value byte is
     * returned as an int in the range 0 to 255. If no byte is available
     * because the end of the stream has been reached, the value -1 is
     * returned. This method blocks until input data is available, the end of
     * the stream is detected, or an exception is thrown.
     *
     * @return  The next byte of data, or -1 if the end of the stream is
     *          reached.
     *
     * @throws  IOException     If an I/O error occurs.
     */
    public int read()
        throws IOException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "read()", "");
        }
        return stream.read();
    }

    // }}}
    // {{{ read

    /**
     * Reads some number of bytes from the input stream and stores them into
     * the buffer array b. The number of bytes actually read is returned as an
     * integer. This method blocks until input data is available, end of file
     * is detected, or an exception is thrown. <p>
     *
     * If b is null, a NullPointerException is thrown. If the length of b is
     * zero, then no bytes are read and 0 is returned; otherwise, there is an
     * attempt to read at least one byte. If no byte is available because the
     * stream is at end of file, the value -1 is returned; otherwise, at least
     * one byte is read and stored into b. <p>
     *
     * The first byte read is stored into element b[0], the next one into
     * b[1], and so on. The number of bytes read is, at most, equal to the
     * length of b. Let k be the number of bytes actually read; these bytes
     * will be stored in elements b[0] through b[k-1], leaving elements b[k]
     * through b[b.length-1] unaffected. <p>
     *
     * If the first byte cannot be read for any reason other than end of file,
     * then an IOException is thrown. In particular, an IOException is thrown
     * if the input stream has been closed.
     *
     * @param   b       The buffer into which the data is read.
     *
     * @return  The total number of bytes read into the buffer, or -1 if there
     *          is no more data because the end of the stream has been
     *          reached.
     *
     * @throws  IOException     If an I/O error occurs.
     */
    public int read(byte[] b)
        throws IOException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "read(byte[])",
                       logger.level < Log.MTDDETAIL ? "" :
                       "b: " + b);
        }
        return stream.read(b);
    }

    // }}}
    // {{{ read

    /**
     * Reads up to len bytes of data from the input stream into an array of
     * bytes. An attempt is made to read as many as len bytes, but a smaller
     * number may be read, possibly zero. The number of bytes actually read is
     * returned as an integer. <p>
     *
     * This method blocks until input data is available, end of file is
     * detected, or an exception is thrown. <p>
     *
     * If b is null, a NullPointerException is thrown. <p>
     *
     * If off is negative, or len is negative, or off+len is greater than the
     * length of the array b, then an IndexOutOfBoundsException is thrown. <p>
     *
     * If len is zero, then no bytes are read and 0 is returned; otherwise,
     * there is an attempt to read at least one byte. If no byte is available
     * because the stream is at end of file, the value -1 is returned;
     * otherwise, at least one byte is read and stored into b. <p>
     *
     * The first byte read is stored into element b[off], the next one into
     * b[off+1], and so on. The number of bytes read is, at most, equal to
     * len. Let k be the number of bytes actually read; these bytes will be
     * stored in elements b[off] through b[off+k-1], leaving elements b[off+k]
     * through b[off+len-1] unaffected. <p>
     *
     * In every case, elements b[0] through b[off] and elements b[off+len]
     * through b[b.length-1] are unaffected. <p>
     *
     * If the first byte cannot be read for any reason other than end of file,
     * then an IOException is thrown. In particular, an IOException is thrown
     * if the input stream has been closed. <p>
     *
     * The read(b, off, len) method for class InputStream simply calls the
     * method read() repeatedly. If the first such call results in an
     * IOException, that exception is returned from the call to the read(b,
     * off, len) method. If any subsequent call to read() results in a
     * IOException, the exception is caught and treated as if it were end of
     * file; the bytes read up to that point are stored into b and the number
     * of bytes read before the exception occurred is returned. Subclasses are
     * encouraged to provide a more efficient implementation of this method.
     *
     * @param   b       The buffer into which the data is read.
     * @param   off     The start offset in array b at which the data is
     *                  written.
     * @param   len     The maximum number of bytes to read.
     *
     * @return  The total number of bytes read into the buffer, or -1 if there
     *          is no more data because the end of the stream has been
     *          reached.
     *
     * @throws  IOException     If an I/O error occurs.
     */
    public int read(byte[] b, int off, int len)
        throws IOException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "read(byte[],int,int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "b: " + b + ", " +
                       "off: " + off + ", " +
                       "len: " + len);
        }
        return stream.read(b, off, len);
    }

    // }}}
    // {{{ reset

    /**
     * Repositions this stream to the position at the time the mark method was
     * last called on this input stream.
     *
     * @throws  IOException     If an I/O error occurs.
     */
    public void reset()
        throws IOException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "reset()", "");
        }
        stream.reset();
    }

    // }}}
    // {{{ skip

    /**
     * Skips over and discards n bytes of data from this input stream. The
     * skip method may, for a variety of reasons, end up skipping over some
     * smaller number of bytes, possibly 0. This may result from any of a
     * number of conditions; reaching end of file before n bytes have been
     * skipped is only one possibility. The actual number of bytes skipped is
     * returned. If n is negative, no bytes are skipped.
     *
     * @param   n       The number of bytes to be skipped.
     *
     * @return  The actual number of bytes skipped.
     *
     * @throws  IOException     If an I/O error occurs.
     */
    public long skip(long n)
        throws IOException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "skip(long)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "n: " + n);
        }
        return stream.skip(n);
    }

    // }}}
}
