// {{{ copyright

/********************************************************************
 *
 * $Id: MRUCache.jpp,v 1.1 2000/09/06 07:15:13 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports

import java.util.Hashtable;
import java.util.Vector;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;


import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;


// }}}

// {{{ doc

/**
 * This is a cache that stores only a certain number of resources. If more
 * resources are added, only the most recently used ones are kept. <p>
 *
 * The MRUCache is based on a {@link MultiMap MultiMap}, so it can store
 * different kinds of resources distinguished by a key. The size limit of the
 * cache applies to the total number of resorces stored independent of their
 * type. <p>
 *
 * When a resource is retrieved with {@link #get(Object) get}, it is
 * automatically removed from the MRUCache. <p>
 *
 * If you just want to store resources of one kind, there is a simplified
 * interface of {@link #get() get()} and {@link #put(Object) put(Object)}
 * where you don't need to provide a key. <p>
 *
 * You cannot store the same resource more than once, not even using different
 * keys. <p>
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.1 $
 * @since       0.98.0
 */

// }}}
public class MRUCache
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (MRUCache.class);

    /**
     * Unique key for simplified use.
     */
    private static Object uniqueKey = new Object ();

    /**
     * The map used for the cache.
     */
    private MultiMap cache;

    /**
     * The list of most recently used resources.
     */
    private Vector mruList;

    /**
     * Parallel vector for the keys of the resources on the mruList.
     */
    private Vector keys;

    /**
     * The maximum size of the MRU list.
     */
    private int mruSize;

    /**
     * The listeners for dropped resources.
     */
    private Vector listeners = new Vector ();

    // }}}

    // {{{ constructor

    /**
     * Create a new MRUCache.
     *
     * @param   size    The maximum number of resources to keep.
     */
    public MRUCache (int size)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "MRUCache(int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "size: " + size);
        }
        mruSize = size;
        cache = new MultiMap (size);
        mruList = new Vector (size);
        keys = new Vector (size);
    }

    // }}}

    // {{{ get

    /**
     * Get a resource from the MRUCache.
     *
     * @return  An arbitrary resource or null if none is available.
     */
    public synchronized Object get()
    {
        return get(uniqueKey);
    }

    /**
     * Get a resource from the MRUCache.
     *
     * @param   key     The key for the resource.
     *
     * @return  An arbitrary resource stored for that key or null if none is
     *          available.
     */
    public synchronized Object get(Object key)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "get(Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "key: " + key);
        }
        Object resource = cache.get(key);
        if (resource != null) {
            cache.remove(key, resource);

            int idx = mruList.indexOf(resource);

            if (idx < 0) {
                if (logger.level >= Log.WRN) {
                    logger.log(Log.WRN, "get(Object)",
                               "OOPS, object not on mruList");
                }
            } else {
                keys.removeElementAt(idx);
                mruList.removeElementAt(idx);
            }
        }
        return resource;
    }

    // }}}
    // {{{ put

    /**
     * Store a resource in the MRUCache.
     *
     * @param   resource The resource to store.
     */
    public synchronized void put(Object resource)
    {
        put(uniqueKey, resource);
    }

    /**
     * Store a resource in the MRUCache.
     *
     * @param   key      The key to store the resource under.
     * @param   resource The resource to store.
     */
    public synchronized void put(Object key, Object resource)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "put(Object,Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "key: " + key + ", " +
                       "resource: " + resource);
        }

        int idx = mruList.indexOf(resource);
        if (idx >= 0) {
            if (logger.level >= Log.WRN) {
                logger.log(Log.WRN, "put(Object,Object)",
                           "Object already in MRUCache: " + resource);
            }
            Object oldkey = keys.elementAt(idx);
            cache.remove(oldkey, resource);
            keys.removeElementAt(idx);
            mruList.removeElementAt(idx);
        }

        if (logger.level >= Log.DBG) {
            logger.log(Log.DBG, "put(Object,Object)",
                       "storing " + resource + "(hash: " +
                       resource.hashCode() +
                       ") in reverseMap");
        }

        if (mruList.size() >= mruSize) {
            Object dropped = mruList.elementAt(mruList.size() - 1);
            mruList.removeElementAt(mruList.size() - 1);
            Object oldkey = keys.elementAt(keys.size() - 1);
            keys.removeElementAt(keys.size() - 1);
            cache.remove(oldkey, dropped);
            fireResourceDropped(oldkey, dropped);
        }
        mruList.insertElementAt(resource, 0);
        keys.insertElementAt(key, 0);

        cache.put(key, resource);
    }

    // }}}

    // {{{ getMaximumSize

    /**
     * Query the maximum size of the MRUCache.
     *
     * @return  The maximum number of resources stored by this cache.
     */
    public synchronized int getMaximumSize()
    {
        return mruSize;
    }

    // }}}
    // {{{ setMaximumSize

    /**
     * Set the maximum size of the MRUCache.
     *
     * @param size  The maximum number of resources allowed for this cache.
     */
    public synchronized void setMaximumSize(int size)
    {
        mruSize = size;

        while (mruList.size() > mruSize) {
            Object dropped = mruList.elementAt(mruList.size() - 1);
            mruList.removeElementAt(mruList.size() - 1);
            Object oldkey = keys.elementAt(keys.size() - 1);
            keys.removeElementAt(keys.size() - 1);
            cache.remove(oldkey, dropped);
            fireResourceDropped(oldkey, dropped);
        }
    }

    // }}}
    // {{{ getCurrentSize

    /**
     * Query the current size of the MRUCache.
     *
     * @return  The current number of resources stored by this cache.
     */
    public synchronized int getCurrentSize()
    {
        return mruList.size();
    }

    // }}}

    // {{{ addMRUCacheListener

    /**
     * Add a listener for dropped resources to the MRUCache.
     *
     * @param   listener        The listener to add.
     */
    public synchronized void addMRUCacheListener(MRUCacheListener listener)
    {
        // avoid duplicates
        listeners.removeElement(listener);
        listeners.insertElementAt(listener, 0);
    }

    // }}}
    // {{{ removeMRUCacheListener

    /**
     * Remove a listener for dropped resources from the MRUCache.
     *
     * @param   listener        The listener to remove.
     */
    public synchronized void removeMRUCacheListener(MRUCacheListener listener)
    {
        listeners.removeElement(listener);
    }

    // }}}
    // {{{ fireResourceDropped

    /**
     * Notify all listeners about a resource dropped from the cache.
     *
     * @param   key     The key the resource was stored under.
     * @param   resource        The dropped resource.
     */
    protected void fireResourceDropped(Object key, Object resource)
    {
        MRUCacheEvent event = new MRUCacheEvent (this, key, resource);

        Vector tmp;
        synchronized(this) {
            tmp = (Vector) listeners.clone();
        }
        for (int i = 0; i < tmp.size(); i++) {
            ((MRUCacheListener) tmp.elementAt(i)).resourceDropped(event);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Tests
    //----------------------------------------------------------------------

    // {{{ class UnitTest

    /**
     * Test cases for the MRUCache class.
     */
    public static class UnitTest extends TestCase
    {
        // {{{ variables

        /**
         * The MRUCache to test.
         */
        private MRUCache cache;

        /**
         * Simple test keys and values.
         */
        private Object key1, key2, value1, value2, value3, value4;

        // }}}

        // {{{ constructor

        /**
         * Create a new UnitTest.
         *
         * @param       name    The name of the test.
         */
        public UnitTest(String name)
        {
                super(name);
        }

        // }}}

        // {{{ setUp

        /**
         * Set up the test fixture.
         */
        protected void setUp()
        {
            //  Log.setOutputLevel(10);
            //  Log.setPreQueueLevel(10);
            //  Log.setQueueing(false);
            //  Logger.setDefaultLogLevel(10);
            cache = new MRUCache (3);
            key1 = new Integer (1);
            key2 = new Integer (2);
            value1 = new Integer (11);
            value2 = new Integer (12);
            value3 = new Integer (13);
            value4 = new Integer (14);
        }

        // }}}
        // {{{ suite

        /**
         * Get the suite of tests.
         *
         * @return      The tests.
         */
        public static Test suite()
        {
            return new TestSuite(UnitTest.class);
        }

        // }}}

        // {{{ testMulti

        /**
         * Test the extended multi map functionality of the MRUCache.
         */
        public void testMulti()
        {
            cache.put(key1, value1);
            cache.put(key2, value2);
            assert(cache.getCurrentSize() == 2);
            assert(cache.get(key1).equals(value1));
            assert(cache.get(key2).equals(value2));
            assert(cache.getCurrentSize() == 0);

            cache.put(key1, value1);
            cache.put(key2, value2);
            cache.put(key1, value3);
            assert(cache.getCurrentSize() == 3);
            Object got = cache.get(key1);
            assert(got.equals(value1) || got.equals(value3));
            assert(cache.getCurrentSize() == 2);
        }

        // }}}
        // {{{ testListener

        /**
         * Test adding and removing an MRUCacheListener.
         */
        public void testListener()
        {
            CacheListener listener = new CacheListener ();
            cache.addMRUCacheListener(listener);

            cache.put(key1, value1);
            cache.put(key2, value2);
            cache.put(key1, value3);
            assert(cache.getCurrentSize() == 3);
            assert(listener.events.size() == 0);

            cache.put(key1, value4);
            assert(cache.getCurrentSize() == 3);
            assert(listener.events.size() == 1);
            MRUCacheEvent e = (MRUCacheEvent) listener.events.elementAt(0);
            assert(e.getKey() == key1);
            assert(e.getResource() == value1);

            assert(cache.getMaximumSize() == 3);
            cache.setMaximumSize(4);
            assert(cache.getMaximumSize() == 4);
            cache.put(key1, value1);
            assert(cache.getCurrentSize() == 4);
            cache.setMaximumSize(2);
            assert(cache.getMaximumSize() == 2);
            assert(cache.getCurrentSize() == 2);
            assert(listener.events.size() == 3);
            e = (MRUCacheEvent) listener.events.elementAt(1);
            assert(e.getKey() == key2);
            assert(e.getResource() == value2);
            e = (MRUCacheEvent) listener.events.elementAt(2);
            assert(e.getKey() == key1);
            assert(e.getResource() == value3);
        }

        // }}}

        // {{{ class CacheListener

        private class CacheListener
            implements MRUCacheListener
        {
            /**
             * Buffer for events.
             */
            public Vector events = new Vector ();

            // {{{ constructor

            /**
             * Create a new CacheListener.
             */
            public CacheListener ()
            {
            }

            // }}}

            // {{{ resourceDropped

            /**
             * Notify the listener about dropped resource.
             *
             * @param   e       The event containing the details.
             */
            public void resourceDropped(MRUCacheEvent e)
            {
                events.addElement(e);
            }

            // }}}
        }

        // }}}
    }

    // }}}

}
