// {{{ copyright

/********************************************************************
 *
 * $Id: MapResourceBundle.java,v 1.12 2000/11/29 14:38:16 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports

import java.io.InputStream;
import java.io.IOException;

import java.net.URL;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Locale;
import java.util.MissingResourceException;
import java.util.Properties;
import java.util.PropertyResourceBundle;
import java.util.ResourceBundle;

import javax.swing.Icon;
import javax.swing.ImageIcon;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

// {{{ doc

/**
 * The MapResourceBundle class extends the standard ResourceBundle to provide
 * a more convenient overall handling of resources and property files. <p>
 *
 * Instead of using the parent mechanism, all key/value pairs are flattened
 * into one {@link java.util.Hashtable Hashtable}, with values added later via
 * {@link #addResources addResources} overriding the previous
 * setting for an identical key. <p>
 *
 * Additional queries have been added that return a default value for a
 * non-existing key instead of throwing a {@link
 * java.util.MissingResourceException MissingResourceException}: {@link
 * #getString getString}, {@link #getInteger getInteger} and {@link
 * #getBoolean getBoolean}. <p>
 *
 * There is special support for {@link javax.swing.Icon Icon}s through {@link
 * #getIcon getIcon}, {@link #setIcon setIcon} and {@link #setResolvingClass
 * setResolvingClass}. It adds one level of indirection. Instead of fetching
 * Icons directly, hardwiring their locations into the program, a property
 * that maps to a String containing the Icon's location is used. The
 * MapResourceBundle takes care of fetching and caching the icon itself. <p>
 *
 * With {@link #fetchProperties fetchProperties} resources can be added from a
 * property file that is bundled with a jar archive, something that doesn't
 * always work with {@link java.util.ResourceBundle#getBundle(String)
 * ResourceBundle.getBundle}, especially in Applets.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.12 $
 */

// }}}
public class MapResourceBundle extends ResourceBundle
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (MapResourceBundle.class);

    /**
     * The properties.
     */
    private Hashtable map = new Hashtable ();

    /**
     * Cache for Icons.
     */
    private Hashtable icons = new Hashtable ();

    /**
     * Resolving class for icons, needed for jars.
     */
    private Class resolver = null;

    // }}}

    // {{{ constructors

    /**
     * Create a new MapResourceBundle.
     */
    public MapResourceBundle()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "MapResourceBundle()", "");
        }
    }

    /**
     * Create a new MapResourceBundle from a ResourceBundle, copying its
     * key/value pairs.
     *
     * @param   rb      The resources to start with.
     */
    public MapResourceBundle(ResourceBundle rb)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "MapResourceBundle(ResourceBundle)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "rb: " + rb);
        }
        addResources(rb);
    }

    // }}}

    // {{{ addResources(ResourceBundle)

    /**
     * Add a ResourceBundle's resources to the MapResourceBundle, overriding
     * existing resources with identical keys.
     *
     * @param   bundle  The ResourceBundle to add.
     */
    public void addResources(ResourceBundle bundle)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "addResources(ResourceBundle)", "");
        }
        for (Enumeration e = bundle.getKeys(); e.hasMoreElements();) {
            String key = (String) e.nextElement();
            map.put(key, bundle.getObject(key));
        }
    }

    // }}}
    // {{{ addResources(Properties)

    /**
     * Add a some Properties to the MapResourceBundle, overriding existing
     * resources with identical keys.
     *
     * @param   properties      The Properties to add.
     */
    public void addResources(Properties properties)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "addResources(Properties)", "");
        }
        for (Enumeration e = properties.keys(); e.hasMoreElements();) {
            String key = (String) e.nextElement();
            map.put(key, properties.get(key));
        }
    }

    // }}}
    // {{{ fetchProperties(String, Locale, Class)

    /**
     * This is a replacement for {@link
     * java.util.ResourceBundle#getBundle(String,Locale)
     * ResourceBundle.getBundle} that can handle resources from jar files.
     * This version searches for <code>.properties</code> files only. The
     * properties are then added to the MapResourceBundle.
     *
     * @param   basename        Basename for resource file.
     * @param   locale          The locale to use.
     * @param   resolve         A class from the same jar file as the resource.
     *
     * @throws  MissingResourceException If the resource cannot be found.
     */
    public void fetchProperties(String basename, Locale locale, Class resolve)
         throws MissingResourceException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "fetchProperties(String,Locale,Class)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "basename: " + basename + ", " +
                       "locale: " + locale + ", " +
                       "resolve: " + resolve);
        }

        boolean found = false;

        // first the base file
        InputStream stream =
            resolve.getResourceAsStream(basename + ".properties");
        if (stream != null) {
            try {
                addResources (new PropertyResourceBundle (stream));
                found = true;
            } catch (IOException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log("fetchProperties(String,Locale,Class)", ex);
                }
            }
        }

        // now add the language
        stream =
            resolve.getResourceAsStream(basename + "_" +
                                        locale.getLanguage() + ".properties");
        if (stream != null) {
            try {
                addResources (new PropertyResourceBundle (stream));
                found = true;
            } catch (IOException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log("fetchProperties(String,Locale,Class)", ex);
                }
            }
        }

        // and now the full locale (no variants)
        stream =
            resolve.getResourceAsStream(basename + "_" +
                                        locale.getLanguage() + "_" +
                                        locale.getCountry() + ".properties");
        if (stream != null) {
            try {
                addResources (new PropertyResourceBundle (stream));
                found = true;
            } catch (IOException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log("fetchProperties(String,Locale,Class)", ex);
                }
            }
        }

        if (found == false) {
            throw new MissingResourceException
                ("Can't find resource for " + basename +
                 " locale " + locale,
                 basename, locale.toString());
        }

    }


    // }}}
    // {{{ fetchProperties(String, Class)

    /**
     * This is a replacement for {@link
     * java.util.ResourceBundle#getBundle(String) ResourceBundle.getBundle}
     * that can handle resources from jar files.  This version searches for
     * <code>.properties</code> files only. The properties are then added to
     * the MapResourceBundle.
     *
     * @param   basename        Basename for resource file.
     * @param   resolve         A class from the same jar file as the resource.
     *
     * @throws  MissingResourceException If the resource cannot be found.
     */
    public void fetchProperties(String basename, Class resolve)
         throws MissingResourceException
    {
        fetchProperties(basename, Locale.getDefault(), resolve);
    }

    // }}}

    //----------------------------------------------------------------------
    // Necessary overrides
    //----------------------------------------------------------------------
    // {{{ getKeys

    /**
     * Return an enumeration of the keys.
     *
     * @return  The enumeration.
     */
    public Enumeration getKeys()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getKeys()", "");
        }
        return map.keys();
    }

    // }}}
    // {{{ handleGetObject


    /**
     * Get an object from the MapResourceBundle.
     *
     * @param   key     The key to look for.
     *
     * @return  The Object associated with the key.
     *
     * @throws  MissingResourceException        If the key is not found.
     */
    public Object handleGetObject(String key)
         throws MissingResourceException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "handleGetObject(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "key: " + key);
        }
        if (map.containsKey(key)) {
            return map.get(key);
        }
        throw new MissingResourceException ("Missing resource: " + key,
                                            MapResourceBundle.class.getName(),
                                            key);
    }

    // }}}

    //----------------------------------------------------------------------
    // Additional queries
    //----------------------------------------------------------------------
    // {{{ getString

    /**
     * Get a String from the MapResourceBundle. If it ain't there, return
     * the supplied default value.
     *
     * @param   key     The key to look for.
     * @param   def     The default value in case key doesn't exist.
     *
     * @return  The String associated with the key.
     */
    public String getString(String key, String def)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "handleGetObject(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "key: " + key);
        }
        if (map.containsKey(key)) {
            return (String) map.get(key);
        }
        if (logger.level >= Log.MSG) {
            logger.log(Log.MSG, "getString(String,String)",
                       "Warning - unknown resource: " + key);
        }
        return def;
    }

    // }}}
    // {{{ getInteger

    /**
     * Get an int value from the MapResourceBundle. If it ain't there, or if
     * it is not a valid integer, return the supplied default value.
     *
     * @param   key     The key to look for.
     * @param   def     The default value in case key doesn't exist.
     *
     * @return  The value associated with the key.
     *
     * @deprecated      Replaced in versio 0.98.0 with {@link #getInt getInt}
     *                  for the sake of conformance with {@link
     *                  de.qfs.lib.option.OptionSet OptionSet} and {@link
     *                  ArgsParser ArgsParser}.
     */
    public int getInteger(String key, int def)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getInteger(String,int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "key: " + key + ", " +
                       "def: " + def);
        }
        if (map.containsKey(key)) {
            try {
                return Integer.parseInt((String) map.get(key));
            } catch (NumberFormatException ex) {
                logger.log(Log.WRN, "getInteger(String,int)", ex);
                return def;
            }
        }
        if (logger.level >= Log.MSG) {
            logger.log(Log.MSG, "getInteger(String,int)",
                       "Warning - unknown resource: " + key);
        }
        return def;
    }

    // }}}
    // {{{ getInteger

    /**
     * Get an int value from the MapResourceBundle. If it ain't there, or if
     * it is not a valid integer, return the supplied default value.
     *
     * @param   key     The key to look for.
     * @param   def     The default value in case key doesn't exist.
     *
     * @return  The value associated with the key.
     *
     * @since   0.98.0
     */
    public int getInt(String key, int def)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getInteger(String,int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "key: " + key + ", " +
                       "def: " + def);
        }
        if (map.containsKey(key)) {
            try {
                return Integer.parseInt((String) map.get(key));
            } catch (NumberFormatException ex) {
                logger.log(Log.WRN, "getInteger(String,int)", ex);
                return def;
            }
        }
        if (logger.level >= Log.MSG) {
            logger.log(Log.MSG, "getInteger(String,int)",
                       "Warning - unknown resource: " + key);
        }
        return def;
    }

    // }}}
    // {{{ getBoolean

    /**
     * Get a boolean from the MapResourceBundle. If it ain't there, return
     * the supplied default value.
     *
     * @param   key     The key to look for.
     * @param   def     The default value in case key doesn't exist.
     *
     * @return  The boolean associated with the key. Anything that is not
     *          the string "TRUE" (independent of case) is considered false.
     */
    public boolean getBoolean(String key, boolean def)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getBoolean(String,boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "key: " + key + ", " +
                       "def: " + def);
        }
        if (map.containsKey(key)) {
            return new Boolean ((String) map.get(key)).booleanValue();
        }
        if (logger.level >= Log.MSG) {
            logger.log(Log.MSG, "getBoolean(String,boolean)",
                       "Warning - unknown resource: " + key);
        }
        return def;
    }

    // }}}

    // {{{ setResolvingClass

    /**
     * Set the Class to use for getting Icon resources. This is useful with
     * applets. Resources will be located in the jar file that clazz came from.
     *
     * @param   clazz   The Class to use with getResource.
     */
    public void setResolvingClass(Class clazz)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setResolvingClass(Class)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "clazz: " + clazz);
        }

        resolver = clazz;
    }

    // }}}
    // {{{ getIcon

    /**
     * Get an icon for a resource name. The resource name should map to the
     * icon's resource path.
     *
     * @param   key     The resource name.
     * @param   def     The default value in case key doesn't exist.
     *
     * @return  The icon, if found, or the default value.
     */
    public Icon getIcon(String key, Icon def)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getIcon(String,Icon)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "key: " + key + ", " +
                       "def: " + def);
        }
        if (icons.containsKey(key)) {
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "getIcon(String,Icon)",
                           "Cache hit");
            }
            return (Icon) icons.get(key);
        }
        if (map.containsKey(key)) {
            String iconName = (String) map.get(key);
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "getIcon(String,Icon)",
                           "Icon name: " + iconName);
            }
            URL res;
            if (resolver != null) {
                res = resolver.getResource(iconName);
            } else {
                res = ClassLoader.getSystemResource(iconName);
            }
            if (logger.level >= Log.DBG) {
                logger.log(Log.DBG, "getIcon(String,Icon)",
                           "resource: " + res);
            }
            if (res == null) {
                if (logger.level >= Log.MSG) {
                    logger.log(Log.MSG, "getIcon(String,Icon)",
                               "Cannot get URL for " + iconName);
                }
                return def;
            }
            ImageIcon icon = new ImageIcon (res);
            icons.put(key, icon);
            return icon;
        }
        if (logger.level >= Log.MSG) {
            logger.log(Log.MSG, "getIcon(String,Icon)",
                       "Warning - unknown resource: " + key);
        }
        return def;
    }

    // }}}
    // {{{ setIcon

    /**
     * Set an icon for a resource name. This can be useful with with some
     * browsers, e.g. netscape, that don't implement getResource correctly.
     *
     * @param   key     The resource name.
     * @param   icon    The icon to set.
     */
    public void setIcon(String key, Icon icon)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setIcon(String,Icon)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "key: " + key + ", " +
                       "icon: " + icon);
        }

        icons.put(key, icon);
    }

    // }}}
}
