// {{{ copyright

/********************************************************************
 *
 * $Id: MessageCollector.jpp,v 1.6 2001/02/04 12:19:55 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports


import java.lang.ThreadLocal;



import java.util.ArrayList;
import java.util.List;


import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

// }}}

/**
 * Very simple mechanism for collecting warning and error messages. <p>
 *
 * This class is intended for use in a parser or similar application where
 * many warnings and errors may arise during one run. Each message passed to
 * the MessageCollector is assigned a level of severity, {@link #WARNING
 * WARNING} for warnings, {@link #ERROR ERROR} for errors and {@link
 * #FATAL_ERROR FATAL_ERROR} for fatal errors. After processing is finished,
 * the number of warnings, errors and fatal errors can be queried and the
 * collected messages are available via {@link #getMessages getMessages}. <p>
 *
 * Things get even more interesting when nested calls need to be handled,
 * e.g. when a parser needs to include files. To that end an {@link #instance
 * instance} mechanism is implemented that associates a MessageCollector with
 * the current thread. For JDK 1.2 and above, {@link java.lang.ThreadLocal
 * ThreadLocal} is used for this, but for JDK 1.1 be sure to {@link #release
 * release} MessageCollector instances before a thread terminates, or they
 * will become uncollectable garbage.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.6 $
 */
public class MessageCollector
{
    // {{{ constants

    /**
     * Level for neither warnings nor errors.
     */
    public final static int NO_ERROR = 0;

    /**
     * Level for warnings.
     */
    public final static int WARNING = 1;

    /**
     * Level for errors.
     */
    public final static int ERROR = 2;

    /**
     * Level for fatal errors.
     */
    public final static int FATAL_ERROR = 3;

    // }}}
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (MessageCollector.class);


    /**
     * Thread local MessageCollector instance.
     *
     * @since   0.98.0
     */
    private static ThreadLocal collectors = new ThreadLocal ()
    {
        protected Object initialValue() {return new MessageCollector ();}
    };

    /**
     * The collected messages.
     */
    private List messages = new ArrayList ();

    /**
     * The number of warnings.
     */
    private int warnings;

    /**
     * The number of errors.
     */
    private int errors;

    /**
     * The number of fatal errors.
     */
    private int fatalErrors;

    // }}}

    //----------------------------------------------------------------------
    // Constructor
    //----------------------------------------------------------------------
    // {{{ MessageCollector

    /**
     * Create a new MessageCollector.
     */
    public MessageCollector ()
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // Static methods
    //----------------------------------------------------------------------
    // {{{ instance

    /**
     * Get the MessageCollector for the current thread, creating a new one if
     * necessary.
     *
     * @return  The MessageCollector for the current thread.
     *
     * @since   0.98.0
     */
    public static synchronized MessageCollector instance()
    {

        return (MessageCollector) collectors.get();

    }

    // }}}
    // {{{ release

    /**
     * Destroy the MessageCollector for the current thread. With JDK 1.2 and
     * above, MessageCollectors will be garbage collected after a thread
     * terminates, but with JDK 1.1 it is necessary to call this method to
     * avoid memory leaks.
     *
     * @since   0.98.0
     */
    public static synchronized void release()
    {

        ((MessageCollector) collectors.get()).reset();

    }

    // }}}

    //----------------------------------------------------------------------
    // Instance methods
    //----------------------------------------------------------------------
    // {{{ addMessage

    /**
     * Add a message to the MessageCollector.
     *
     * @param   level   The level of the message.
     * @param   message The message.
     */
    public synchronized void addMessage(int level, String message)
    {
        messages.add(message);
        switch(level) {
        case WARNING:
            warnings++;
            break;
        case ERROR:
            errors++;
            break;
        case FATAL_ERROR:
            fatalErrors++;
            break;
        }
    }

    // }}}
    // {{{ getMessages

    /**
     * Get the collected messages.
     *
     * @return  The collected messages.
     */
    public synchronized String[] getMessages()
    {
        return (String[]) messages.toArray(new String[0]);
    }

    // }}}

    // {{{ getWarningCount

    /**
     * Get the number of warnings that occured.
     *
     * @return  The number of warnings.
     */
    public int getWarningCount()
    {
        return warnings;
    }

    // }}}
    // {{{ getErrorCount

    /**
     * Get the number of errors that occured.
     *
     * @return  The number of errors.
     */
    public int getErrorCount()
    {
        return errors;
    }

    // }}}
    // {{{ getFatalErrorCount

    /**
     * Get the number of fatal errors that occured.
     *
     * @return  The number of fatal errors.
     */
    public int getFatalErrorCount()
    {
        return fatalErrors;
    }

    // }}}

    // {{{ getWorstCase

    /**
     * Get the highest level of the collected messages.
     *
     * @return  The highest level so far.
     */
    public synchronized int getWorstCase()
    {
        if (fatalErrors > 0) {
            return FATAL_ERROR;
        }
        if (errors > 0) {
            return ERROR;
        }
        if (warnings > 0) {
            return WARNING;
        }
        return NO_ERROR;
    }

    // }}}

    // {{{ reset

    /**
     * Reset the MessageCollector by clearing the collected messages and
     * resetting the counts.
     */
    public synchronized void reset()
    {
        warnings = 0;
        errors = 0;
        fatalErrors = 0;
        messages.clear();
    }

    // }}}
}
