// {{{ copyright

/********************************************************************
 *
 * $Id: MultiMap.jpp,v 1.5 2001/04/21 19:55:04 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.lib.util;

// {{{ imports

import java.io.Serializable;


import java.util.AbstractCollection;
import java.util.AbstractSet;
import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.TreeMap;


import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;


import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;


// }}}

/**
 * This is a special implementation of the Map interface that can store more
 * than one value per key. When a value is retrieved with {@link #get get}, an
 * arbitrary one is returned if more than one is available. <p>
 *
 * The {@link #remove(Object) remove(key)} removes all values stored for
 * that key. To remove only a single value, use the {@link
 * #remove(Object,Object) remove(key,value)} method. <p>
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.5 $
 * @since       0.98.0
 */
public class MultiMap
    implements Map, Cloneable, Serializable
{
    // {{{ variables

    /**
     * The UID used for Serialization.
     */
    static final long serialVersionUID = 5827101685390117257L;

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (MultiMap.class);

    /**
     * Type selector for a key iterator.
     */
    protected final static int KEYS = 0;

    /**
     * Type selector for a value iterator.
     */
    protected final static int VALUES = 1;

    /**
     * Type selector for an entry iterator.
     */
    protected final static int ENTRIES = 2;

    /**
     * The actual map used for storing keys and values.
     * @serial
     */
    private Map delegate;

    /**
     * Number of values stored in the MultiMap.
     * @serial
     */
    protected int valCount;

    /**
     * The Set returned by {@link #keySet keySet}.
     */
    protected transient Set keySet;

    /**
     * The Collection returned by {@link #values values}.
     */
    protected transient Collection values;

    /**
     * The Set returned by {@link #entrySet entrySet}.
     */
    protected transient Set entrySet;

    /**
     * Since HashMap's modification counter is private the MultiMap needs its
     * own.
     */
    protected transient int modCounter = 0;

    // }}}

    // {{{ constructors

    /**
     * Create a new MultiMap.
     */
    public MultiMap()
    {
        delegate = new HashMap ();
    }

    /**
     * Create a new MultiMap.
     *
     * @param   initialCapacity   The initial capacity of the MultiMap.
     */
    public MultiMap(int initialCapacity)
    {
        delegate = new HashMap (initialCapacity);
    }

    /**
     * Create a new MultiMap and initialize it with all the mappings from
     * the passed map.
     *
     * @param   map     The map that initializes this MultiMap.
     */
    public MultiMap (Map map)
    {
        delegate = new HashMap (map.size());
        putAll(map);
    }

    // }}}

    //----------------------------------------------------------------------
    // Queries
    //----------------------------------------------------------------------
    // {{{ isEmpty

    /**
     * Returns <tt>true</tt> if this map contains no key-value mappings.
     *
     * @return <tt>true</tt> if this map contains no key-value mappings.
     */
    public boolean isEmpty()
    {
        return valCount == 0;
    }

    // }}}
    // {{{ size

    /**
     * Returns the number of key-value mappings in this map.  If the
     * map contains more than <tt>Integer.MAX_VALUE</tt> elements, returns
     * <tt>Integer.MAX_VALUE</tt>.
     *
     * @return the number of key-value mappings in this map.
     */
    public int size()
    {
        return valCount;
    }

    // }}}
    // {{{ get

    /**
     * Returns the value to which this map maps the specified key. If more
     * than one value is found, an arbitrary one is chosen. Returns
     * <tt>null</tt> if the map contains no mapping for this key.  A return
     * value of <tt>null</tt> does not <i>necessarily</i> indicate that the
     * map contains no mapping for the key; it's also possible that the map
     * explicitly maps the key to <tt>null</tt>.  The <tt>containsKey</tt>
     * operation may be used to distinguish these two cases.
     *
     * @param key key whose associated value is to be returned.
     * @return the value to which this map maps the specified key, or
     *	       <tt>null</tt> if the map contains no mapping for this key.
     *
     * @see #containsKey(Object)
     */
    public Object get(Object key)
    {
        Object result = delegate.get(key);
        if (result instanceof ValueSet) {
            if (((ValueSet) result).size() == 0) {
                return null;
            }
            // get an arbitrary object from the set
            return ((ValueSet) result).iterator().next();
        }
        return result;
    }

    // }}}
    // {{{ containsKey

    /**
     * Returns <tt>true</tt> if this map contains a mapping for the specified
     * key.
     *
     * @param key key whose presence in this map is to be tested.
     * @return <tt>true</tt> if this map contains a mapping for the specified
     * key.
     *
     */
    public boolean containsKey(Object key)
    {
        return delegate.containsKey(key);
    }

    // }}}
    // {{{ containsValue

    /**
     * Returns <tt>true</tt> if this map maps one or more keys to the
     * specified value.  More formally, returns <tt>true</tt> if and only if
     * this map contains at least one mapping to a value <tt>v</tt> such that
     * <tt>(value==null ? v==null : value.equals(v))</tt>.  This operation
     * does require time linear in the map size.
     *
     * @param value value whose presence in this map is to be tested.
     * @return <tt>true</tt> if this map maps one or more keys to the
     *         specified value.
     */
    public boolean containsValue(Object value)
    {
        for (Iterator i = values().iterator(); i.hasNext(); ) {
            Object cmp = i.next();
            if (cmp instanceof ValueSet) {
                if (((ValueSet) cmp).contains(value)) {
                    return true;
                }
            } else {
                if (areEqual(cmp, value)) {
                    return true;
                }
            }
        }
        return false;
    }

    // }}}

    //----------------------------------------------------------------------
    // Modification
    //----------------------------------------------------------------------
    // {{{ put

    /**
     * Associates the specified value with the specified key in this map. If
     * the map previously contained a mapping for this key, the old value is
     * <strong>NOT</strong> replaced.
     *
     * @param key key with which the specified value is to be associated.
     * @param value value to be associated with the specified key.
     * @return Since no value is ever replaced, this method always returns
     * null.
     */
    public Object put(Object key, Object value)
    {
        if (! delegate.containsKey(key)) {
            delegate.put(key, value);
            valCount++;
            modCounter++;
            return null;
        }

        Object current = delegate.get(key);
         if (current instanceof ValueSet) {
            if (((ValueSet) current).add(value)) {
                valCount++;
                modCounter++;
            }
         } else if (! areEqual(current, value)) {
            ValueSet set = createValueSet();
            set.add(current);
            set.add(value);
            valCount++;
            modCounter++;
            delegate.put(key, set);
        }
        return null;
    }

    // }}}
    // {{{ putAll

    /**
     * Copies all of the mappings from the specified map to this map.
     *
     * @param map Mappings to be stored in this map.
     */
    public void putAll(Map map)
    {
        for (Iterator i = map.entrySet().iterator(); i.hasNext(); ) {
            Map.Entry entry = (Map.Entry) i.next();
            put(entry.getKey(), entry.getValue());
        }
    }

    // }}}
    // {{{ remove(Object)

    /**
     * Removes all mappings for this key from this map if present.
     *
     * @param key key whose mapping is to be removed from the map.
     * @return previous value associated with specified key, or <tt>null</tt>
     *	       if there was no mapping for key. If there was more than one
     *	       mapping, an arbitrary one is choden for the return value. A
     *	       <tt>null</tt> return can also indicate that the map previously
     *	       associated <tt>null</tt> with the specified key.
     */
    public Object remove(Object key)
    {
        if (! delegate.containsKey(key)) {
            return null;
        }
        Object current = delegate.get(key);
        if (current instanceof ValueSet) {
            delegate.remove(key);
            if (((ValueSet) current).size() > 0) {
                valCount -= ((ValueSet) current).size();
                modCounter++;
                Object result = ((ValueSet) current).iterator().next();
                return result;
            } else {
                return null;
            }
        }
        valCount--;
        modCounter++;
        return delegate.remove(key);
    }

    // }}}
    // {{{ clear

    /**
     * Removes all mappings from this map.
     */
    public void clear()
    {
        delegate.clear();
        valCount = 0;
        modCounter++;
    }

    // }}}

    //----------------------------------------------------------------------
    // Keys, values, entries
    //----------------------------------------------------------------------
    // {{{ keySet

    /**
     * Returns a set view of the keys contained in this map.  The set is
     * backed by the map, so changes to the map are reflected in the set, and
     * vice-versa.  If the map is modified while an iteration over the set is
     * in progress, the results of the iteration are undefined.  The set
     * supports element removal, which removes the corresponding mapping from
     * the map, via the <tt>Iterator.remove</tt>, <tt>Set.remove</tt>,
     * <tt>removeAll</tt> <tt>retainAll</tt>, and <tt>clear</tt> operations.
     * It does not support the add or <tt>addAll</tt> operations.
     *
     * @return a set view of the keys contained in this map.
     */
    public Set keySet()
    {
        if (keySet == null) {
            keySet = new AbstractSet () {
		public Iterator iterator() {
		    return new MapIterator(KEYS);
		}
		public int size() {
		    return delegate.size();
		}
                public boolean contains(Object o) {
                    return delegate.containsKey(o);
                }
		public boolean remove(Object o) {
		    return MultiMap.this.remove(o) != null;
		}
		public void clear() {
		    MultiMap.this.clear();
		}
            };
        }
        return keySet;
    }

    // }}}
    // {{{ values

    /**
     * Returns a collection view of the values contained in this map.  The
     * collection is backed by the map, so changes to the map are reflected in
     * the collection, and vice-versa.  If the map is modified while an
     * iteration over the collection is in progress, the results of the
     * iteration are undefined.  The collection supports element removal,
     * which removes the corresponding mapping from the map, via the
     * <tt>Iterator.remove</tt>, <tt>Collection.remove</tt>,
     * <tt>removeAll</tt>, <tt>retainAll</tt> and <tt>clear</tt> operations.
     * It does not support the add or <tt>addAll</tt> operations.
     *
     * @return A collection view of the values contained in this map.
     */
    public Collection values()
    {
        if (values == null) {
            values = new AbstractCollection() {
                public Iterator iterator() {
                    return new MapIterator(VALUES);
                }
                public int size() {
                    return valCount;
                }
                public boolean contains(Object o) {
                    return containsValue(o);
                }
                public void clear() {
                    MultiMap.this.clear();
                }
            };
        }
        return values;
    }

    // }}}
    // {{{ entrySet

    /**
     * Returns a set view of the mappings contained in this map.  Each element
     * in the returned set is a <tt>Map.Entry</tt>.  The set is backed by the
     * map, so changes to the map are reflected in the set, and vice-versa.
     * If the map is modified while an iteration over the set is in progress,
     * the results of the iteration are undefined.  The set supports element
     * removal, which removes the corresponding mapping from the map, via the
     * <tt>Iterator.remove</tt>, <tt>Set.remove</tt>, <tt>removeAll</tt>,
     * <tt>retainAll</tt> and <tt>clear</tt> operations.  It does not support
     * the <tt>add</tt> or <tt>addAll</tt> operations.
     *
     * @return a set view of the mappings contained in this map.
     */
    public Set entrySet()
    {
        if (entrySet == null) {
            entrySet = new AbstractSet () {
                public Iterator iterator() {
                    return new MapIterator(ENTRIES);
                }

                public boolean contains(Object o) {
                    if (!(o instanceof Map.Entry))
                        return false;
                    Map.Entry entry = (Map.Entry)o;
                    Object key = entry.getKey();
                    if (! delegate.containsKey(key)) {
                        return false;
                    }
                    Object current = delegate.get(key);
                    if (current instanceof ValueSet) {
                        return ((ValueSet) current).contains(entry.getValue());
                    }
                    return areEqual(current, entry.getValue());
                }

		public boolean remove(Object o) {
                    if (!(o instanceof Map.Entry))
                        return false;
                    Map.Entry entry = (Map.Entry)o;
                    return MultiMap.this._remove
                        (entry.getKey(), entry.getValue());
                }

                public int size() {
                    return valCount;
                }

                public void clear() {
                    MultiMap.this.clear();
                }
            };
        }
        return entrySet;
    }

    // }}}

    //----------------------------------------------------------------------
    // Comparison and hashing
    //----------------------------------------------------------------------
    // {{{ equals

    /**
     * Test whether an object is equal to the MultiMap. As required by the Map
     * interface, the object is considered euqal, if it is a Map whose
     * entrySet is equal to the MultiMap's entrySet.
     *
     * @param   o       The object to compare to.
     *
     * @return  True if the object is considered equal to the Map.
     */
    public boolean equals(Object o)
    {
        if (! (o instanceof Map)) {
            return false;
        }
        return entrySet().equals(((Map) o).entrySet());
    }

    // }}}
    // {{{ hashCode

    /**
     * Get the hashCode of the MultiMap. As required by the Map interface,
     * this is the sum of the hashCodes of the MultiMap's entrySet.
     *
     * @return  The MultiMap's hashCode.
     */
    public int hashCode()
    {
        return entrySet().hashCode();
    }

    // }}}

    //----------------------------------------------------------------------
    // Additional public methods
    //----------------------------------------------------------------------
    // {{{ clone

    /**
     * Returns a shallow copy of this MultiMap: the keys and values themselves
     * are not cloned. If a derived class uses a delegate that is not a
     * HashMap, it must override {@link #cloneDelegate cloneDelegate},
     * otherwise this method will fail with a ClassCastException.
     *
     * @return a shallow copy of this map.
     *
     * @throws ClassCastException If the delegate is not a HashMap and the
     *          derived class doesn't override {@link #cloneDelegate
     *          cloneDelegate}.
     */
    public Object clone()
        throws ClassCastException
    {
	try {
	    MultiMap map = (MultiMap) super.clone();
	    map.delegate = cloneDelegate();
	    map.keySet = null;
            map.values = null;
	    map.entrySet = null;
	    map.modCounter = 0;
	    return map;
	} catch (CloneNotSupportedException e) {
	    // this shouldn't happen, since we are Cloneable
	    throw new InternalError();
	}
    }

    // }}}
    // {{{ remove(Object,Object)

    /**
     * In contrast to {@link #remove(Object) remove(Object)} this method
     * removes only the one binding for key and value. <p>
     *
     * Since this method is not part of the standard Map interface, it cannot
     * directly be synchronized like all the others with the help of a
     * Collections.SynchronizedMap. Nevertheless you can use a SynchronizedMap
     * wrapper around the MultiMap for synchronized standard Map access and
     * explicitely synchronize on that SynchronizedMap when calling this
     * method.
     *
     * @param   key     The key of the binding to remove.
     * @param   value   The value of the binding to remove.
     *
     * @return  True if there really was a binding removed, false otherwise.
     */
    public boolean remove(Object key, Object value)
    {
        if (! delegate.containsKey(key)) {
            return false;
        }
        Object result = delegate.get(key);
        if (result instanceof ValueSet) {
            if (((ValueSet) result).size() == 0) {
                return false;
            }
            if (((ValueSet) result).remove(value)) {
                valCount--;
                modCounter++;
                return true;
            }
            return false;
        } else {
            delegate.remove(key);
            valCount--;
            modCounter++;
            return true;
        }
    }

    // }}}
    // {{{ mapsToMany

    /**
     * Query whether a key maps to more than one values. <p>
     *
     * Since this method is not part of the standard Map interface, it cannot
     * directly be synchronized like all the others with the help of a
     * Collections.SynchronizedMap. Nevertheless you can use a SynchronizedMap
     * wrapper around the MultiMap for synchronized standard Map access and
     * explicitely synchronize on that SynchronizedMap when calling this
     * method. <p>
     *
     * @param   key     The key to look for.
     *
     * @return  True if there is more than one value bound for the key, false
     *          if the key is either unknown or has only single value.
     */
    public boolean mapsToMany(Object key)
    {
        Object result = delegate.get(key);
        return (result instanceof ValueSet) && ((ValueSet) result).size() > 1;
    }

    // }}}
    // {{{ getAll

    /**
     * Get a Set of all Objects to which this map maps the key. Returns
     * <tt>null</tt> if the map contains no mapping for this key. The result
     * is a clone of the actual Set used and may be modified without changing
     * the MultiMap. <p>
     *
     * Since this method is not part of the standard Map interface, it cannot
     * directly be synchronized like all the others with the help of a
     * Collections.SynchronizedMap. Nevertheless you can use a SynchronizedMap
     * wrapper around the MultiMap for synchronized standard Map access and
     * explicitely synchronize on that SynchronizedMap when calling this
     * method. <p>
     *
     * <strong>Note:</strong> Though a set backed by the map would have been
     * nice, the implications of modification to the set are not quite clear
     * and difficult to implement. Maybe in a later version.
     *
     * @param key key whose associated values are to be returned.
     * @return the set of values to which this map maps the specified key, or
     *	       <tt>null</tt> if the map contains no mapping for this key.
     */
    public Set getAll(Object key)
    {
        Object result = delegate.get(key);
        if (result instanceof ValueSet) {
            if (((ValueSet) result).size() == 0) {
                return null;
            }
            // Copy the set but don't use clone so the ValueSet class is not
            // exposed.
            return new HashSet ((ValueSet) result);
        }
        HashSet set = new HashSet ();
        set.add(result);
        return set;
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ createValueSet

    /**
     * This method creates a new ValueSet to store multiple values for the
     * same key. The default is to use a NestedSet which extends HashSet. By
     * overriding this method, a derived class can change the implementation
     * of the ValueSet used by the MultiMap.
     *
     * @return  A new ValueSet.
     */
    protected ValueSet createValueSet()
    {
        return new NestedSet ();
    }

    // }}}
    // {{{ cloneDelegate

    /**
     * This method is used by the {@link #clone clone} method to create a copy
     * of the delegate. If a derived class uses a delegate that is not a
     * HashMap it needs to override this method, otherwise {@link #clone
     * clone} will result in a ClassCastException.
     *
     * @return  A copy of the delegate.
     *
     * @throws ClassCastException If the delegate is not a HashMap and the
     *          derived class doesn't override {@link #cloneDelegate
     *          cloneDelegate}.
     */
    protected Map cloneDelegate()
        throws ClassCastException
    {
        return (Map) ((HashMap) delegate).clone();
    }

    // }}}
    // {{{ areEqual

    /**
     * Compare two objects.
     *
     * @param   o1      The first object.
     * @param   o2      The second object.
     *
     * @return  True if both objects are null or one equals the other.
     */
    protected final boolean areEqual(Object o1, Object o2)
    {
        return (o1 == null) ? (o2 == null) : o1.equals(o2);
    }

    // }}}
    // {{{ _remove(Object,Object)

    /**
     * This is the actual implementation of the {@link #remove(Object,Object)
     * remove(Object,Object)} method.
     *
     * @param   key     The key of the binding to remove.
     * @param   value   The value of the binding to remove.
     *
     * @return  True if there really was a binding removed, false otherwise.
     */
    protected boolean _remove(Object key, Object value)
    {
        if (! delegate.containsKey(key)) {
            return false;
        }
        Object current = delegate.get(key);
        if (current instanceof ValueSet) {
            boolean result = ((ValueSet) current).remove(value);
            if (result) {
                valCount--;
                modCounter++;
            }
            if (((ValueSet) current).size() == 0) {
                delegate.remove(key);
            }
            return result;
        }
        if (areEqual(current, value)) {
            delegate.remove(key);
            valCount--;
            modCounter++;
            return true;
        }
        return false;
    }

    // }}}

    //----------------------------------------------------------------------
    // Nested classes
    //----------------------------------------------------------------------
    // {{{ class MapIterator

    /**
     * Iterator for the key, value and entry collections of the MultiMap.
     */
    protected class MapIterator
        implements Iterator
    {
        // {{{ variables

        /**
         * The type of the iterator, {@link #KEYS KEYS}, {@link #VALUES
         * VALUES} or {@link #ENTRIES ENTRIES}.
         */
        protected int type;

        /**
         * An iterator over the entries of the delegate.
         */
        protected Iterator entries;

        /**
         * An iteratore over the elements of a ValueSet.
         */
        protected Iterator inner;

        /**
         * The ValueSet that is currently iterated over.
         */
        protected ValueSet current;

        /**
         * The modCounter that the iterator expects.
         */
        protected int expectedModCounter;

        /**
         * The Entry used in the last call to {@link #next next}.
         */
        protected Map.Entry lastEntry;

        /**
         * The lastValue returned from {@link #next next}.
         */
        protected Object lastValue;

        // }}}

        // {{{ constructor

        /**
         * Create a new MapIterator.
         *
         * @param       type    The type of the iterator.
         */
        public MapIterator (int type)
        {
            this.type = type;
            entries = delegate.entrySet().iterator();
            inner = null;
            expectedModCounter = modCounter;
        }

        // }}}

        // {{{ hasNext

        /**
         * Returns <tt>true</tt> if the iteration has more elements. (In other
         * words, returns <tt>true</tt> if <tt>next</tt> would return an
         * element rather than throwing an exception.)
         *
         * @return <tt>true</tt> if the iterator has more elements.
         */
	public boolean hasNext()
        {
            if (inner != null && inner.hasNext()) {
                return true;
            }
            return entries.hasNext();
        }

        // }}}
        // {{{ next

        /**
         * Returns the next element in the interation.
         *
         * @return the next element in the interation.
         *
         * @throws NoSuchElementException if the iteration has no more
         *      elements.
         *
         * @throws ConcurrentModificationException if the MultiMap has been
         *      modified (except through this iterator).
         */
	public Object next()
        {
            if (expectedModCounter != modCounter) {
                throw new ConcurrentModificationException ();
            }
            // This while loop is just a paranoia to handle empty ValueSets
            // in the delegate. It SHOULD be unnecessary.
            while (true) {
                if (inner != null) {
                    try {
                        lastValue = inner.next();
                        if (type == VALUES) {
                            return lastValue;
                        } else {
                            return new MapEntry
                                (lastEntry.getKey(), lastValue);
                        }
                    } catch (NoSuchElementException ex) {
                        inner = null;
                        current = null;
                    }
                }

                lastEntry = (Map.Entry) entries.next();
                if (type == KEYS) {
                    lastValue = lastEntry.getKey();
                    return lastValue;
                }
                if (lastEntry.getValue() instanceof ValueSet) {
                    // Move to the next inner set and repeat the loop
                    current = (ValueSet) lastEntry.getValue();
                    inner = current.iterator();
                } else {
                    lastValue = lastEntry.getValue();
                    if (type == VALUES) {
                        return lastValue;
                    } else {
                        return new MapEntry
                            (lastEntry.getKey(), lastValue);
                    }
                }
            }
	}

        // }}}
        // {{{ remove

        /**
         *
         * Removes from the underlying collection the last element returned by
         * the iterator (optional operation).  This method can be called only
         * once per call to <tt>next</tt>.  The behavior of an iterator is
         * unspecified if the underlying collection is modified while the
         * iteration is in progress in any way other than by calling this
         * method.
         *
         * @exception IllegalStateException if the <tt>next</tt> method has
         *		  not yet been called, or the <tt>remove</tt> method
         *		  has already been called after the last call to the
         *		  <tt>next</tt> method.
         *
         * @throws ConcurrentModificationException if the MultiMap has been
         *      modified (except through this iterator).
         */
	public void remove() {
            if (lastValue == null) {
		throw new IllegalStateException();
            }
            if (expectedModCounter != modCounter) {
                throw new ConcurrentModificationException ();
            }
            if (inner != null) {
                inner.remove();
                lastValue = null;
                if (current.size() == 0) {
                    entries.remove();
                }
                valCount--;
                modCounter++;
                expectedModCounter++;
            } else {
                if (type == KEYS
                    && lastEntry.getValue() instanceof ValueSet) {
                    valCount -= ((ValueSet) lastEntry.getValue()).size();
                } else {
                    valCount--;
                }
                entries.remove();
                lastValue = null;
                modCounter++;
                expectedModCounter++;
            }
        }

        // }}}
    }

    // }}}
    // {{{ class MapEntry

    /**
     * An implementation of the Map.Entry interface.
     */
    protected class MapEntry
        implements Map.Entry
    {
        // {{{ variables

        /**
         * The key of the entry.
         */
        protected Object key;

        /**
         * The value of the entry.
         */
        protected Object value;

        // }}}

        // {{{ constructor

        /**
         * Create a new MapEntry.
         *
         * @param       key     The key of the entry.
         * @param       value   The value of the entry.
         */
        public MapEntry(Object key, Object value)
        {
            this.key = key;
            this.value = value;
        }

        // }}}

        // {{{ getKey

        /**
         * Get the key of the entry.
         *
         * @return      The key of the entry.
         */
        public Object getKey()
        {
            return key;
        }

        // }}}
        // {{{ getValue

        /**
         * Get the value of the entry.
         *
         * @return      The value of the entry.
         */
        public Object getValue()
        {
            return value;
        }

        // }}}
        // {{{ setValue

        /**
         * Set the value of this entry.
         *
         * @param       value   The new value.
         *
         * @return      The previous value of this entry.
         */
        public Object setValue(Object value)
        {
            Object ret = this.value;
            this.value = value;
            if (! areEqual(ret, value)) {
                MultiMap.this.remove(key, ret);
                MultiMap.this.put(key, value);
            }
            return ret;
        }

        // }}}
        // {{{ equals

        /**
         * Test whether an object is equal to this entry.
         *
         * @param       o       The object to test.
         *
         * @return      True if the object implements Map.Entry and has
         * a key and value equal to this entry.
         */
        public boolean equals(Object o)
        {
            if (! (o instanceof Map.Entry)) {
                return false;
            }
            return areEqual(key, ((Map.Entry) o).getKey())
                && areEqual(value, ((Map.Entry) o).getValue());
        }

        // }}}
        // {{{ hashCode

        /**
         * Get a hash code for the entry.
         *
         * @return      A hash code for the entry.
         */
        public int hashCode()
        {
            return (key == null ? 0 : key.hashCode()) ^
                (value == null ? 0 : value.hashCode());
        }

        // }}}

        // {{{ toString

        /**
         * Get a String representatio for the entry.
         *
         * @return      A String representatio for the entry.
         */
        public String toString()
        {
            return super.toString() +
                "[" + key + "=" + value + "]";
        }

        // }}}
    }

    // }}}
    // {{{ interface ValueSet

    /**
     * This interface is a marker for the actual Set used to store multiple
     * values for the same key in a MultiMap. It is needed by a MultiMap to
     * distinguish between a ValueSet and some other Set stored by the user.
     */
    protected interface ValueSet
        extends Set
    {
    }

    // }}}
    // {{{ class NestedSet

    /**
     * This is the class used to store multiple values for the same key in a
     * MultiMap.
     */
    private class NestedSet
        extends HashSet
        implements ValueSet
    {
        // {{{ constructor

        /**
         * Create a new NestedSet.
         */
        public NestedSet ()
        {
        }

        // }}}
    }

    // }}}

    //----------------------------------------------------------------------
    // Tests
    //----------------------------------------------------------------------

    // {{{ class UnitTest

    /**
     * Test cases for the MultiMap class.
     */
    public static class UnitTest extends TestCase
    {
        // {{{ variables

        /**
         * The MultiMap to test.
         */
        private MultiMap map;

        /**
         * Simple test keys and values.
         */
        private Object key1, key2, value1, value2;

        // }}}

        // {{{ constructor

        /**
         * Create a new UnitTest.
         *
         * @param       name    The name of the test.
         */
        public UnitTest(String name)
        {
                super(name);
        }

        // }}}

        // {{{ setUp

        /**
         * Set up the test fixture.
         */
        protected void setUp()
        {
            //  Log.setOutputLevel(10);
            //  Log.setPreQueueLevel(10);
            //  Log.setQueueing(false);
            //  Logger.setDefaultLogLevel(10);
            map = new MultiMap ();
            key1 = new Object ();
            key2 = new Object ();
            value1 = new Object ();
            value2 = new Object ();
        }

        // }}}
        // {{{ suite

        /**
         * Get the suite of tests.
         *
         * @return      The tests.
         */
        public static Test suite()
        {
            return new TestSuite(UnitTest.class);
        }

        // }}}

        // {{{ testMulti

        /**
         * Test the multi functionality of the MultiMap.
         */
        public void testMulti()
        {
            map.put(key1, value1);
            map.put(key2, value2);
            assert(map.size() == 2);
            assert(map.get(key1).equals(value1));
            assert(map.get(key2).equals(value2));

            map.put(key1, value2);
            assert(map.size() == 3);
            Object got = map.get(key1);
            assert(got.equals(value1) || got.equals(value2));

            map.remove(key1, value1);
            assert(map.size() == 2);
            assert(map.get(key1).equals(value2));

            map.put(key1, value1);
            assert(map.size() == 3);
            map.remove(key1);
            assert(map.size() == 1);
        }

        // }}}
        // {{{ testSets

        /**
         * Test key, value and entry sets.
         */
        public void testSets()
        {
            map.put(key1, value1);
            map.put(key2, value1);
            map.put(null, value1);
            map.put(key1, value2);
            map.put(key2, value2);
            map.put(null, value2);

            assert(map.size() == 6);

            Set keys = map.keySet();
            assert(keys.size() == 3);
            assert(keys.contains(key1));
            assert(keys.contains(key2));
            assert(keys.contains(null));
            assert(! keys.contains(value1));
            assert(! keys.contains(value2));

            keys.remove(key1);
            assert(map.size() == 4);
            assert(! keys.contains(key1));
            assert(! map.containsKey(key1));
            map.put(key1, value1);
            assert(keys.contains(key1));
            map.put(key1, value2);
            assert(map.size() == 6);

            Collection values = map.values();
            assert(values.size() == 6);
            assert(values.contains(value1));
            assert(values.contains(value2));
            assert(! values.contains(key1));
            assert(! values.contains(key2));
            assert(! values.contains(null));

            assert(values.remove(value1));
            assert(map.size() == 5);
            assert(values.contains(value1));
            assert(values.remove(value1));
            assert(map.size() == 4);
            assert(values.contains(value1));
            assert(values.remove(value1));
            assert(map.size() == 3);
            assert(values.size() == 3);
            assert(! values.contains(value1));

            map.put(key1, value1);
            assert(values.contains(value1));
            assert(values.size() == 4);
            map.put(key2, value1);
            map.put(null, value1);


            Set entries = map.entrySet();
            assert(entries.size() == 6);
            assert(entries.contains(map.new MapEntry (key1, value1)));
            assert(entries.contains(map.new MapEntry (key2, value1)));
            assert(entries.contains(map.new MapEntry (null, value1)));
            assert(entries.contains(map.new MapEntry (key1, value2)));
            assert(entries.contains(map.new MapEntry (key2, value2)));
            assert(entries.contains(map.new MapEntry (null, value2)));

            entries.remove(map.new MapEntry (key1, value2));
            assert(map.size() == 5);
            assert(entries.size() == 5);
            assert(! entries.contains(map.new MapEntry (key1, value2)));

            entries.remove(map.new MapEntry (key1, value1));
            assert(map.size() == 4);
            assert(entries.size() == 4);
            assert(! entries.contains(map.new MapEntry (key1, value2)));

        }

        // }}}
        // {{{ testIterators

        /**
         * Test iterators.
         */
        public void testIterators()
        {
            map.put(key1, value1);
            map.put(key2, value1);
            map.put(null, value1);
            map.put(key1, value2);
            map.put(key2, value2);
            map.put(null, value2);

            assert(map.size() == 6);
        }

        // }}}
        // {{{ testChange

        /**
         * Test for concurrent changes.
         */
        public void testChange()
        {
            map.put(key1, value1);
            assert(map.get(key1).equals(value1));
            assert(map.size() == 1);

            Iterator i = map.values().iterator();
            // mustn't change
            map.put(key1, value1);
            assert(map.size() == 1);
            i.next();

            i = map.values().iterator();
            // must change
            map.put(key1, value2);
            assert(map.size() == 2);
            try {
                i.next();
                assert(false);
            } catch (ConcurrentModificationException ex) {
            }

            i = map.values().iterator();
            Iterator i2 = map.entrySet().iterator();
            // must change
            // this is a nasty one: replace the entry's value with a value
            // that is already stored under the same key, which must have the
            // same effect as i2.remove()
            Map.Entry entry = (Map.Entry) i2.next();
            if (entry.getValue().equals(value1)) {
                entry.setValue(value2);
            } else {
                entry.setValue(value1);
            }
            assert(map.size() == 1);
            try {
                i.next();
                assert(false);
            } catch (ConcurrentModificationException ex) {
            }

            i = map.values().iterator();
            i2 = map.entrySet().iterator();
            // must not change
            entry = (Map.Entry) i2.next();
            entry.setValue(entry.getValue());
            assert(map.size() == 1);
            i.next();
        }

        // }}}
    }

    // }}}

}
